//////////////////////////////////////////////////////////////////////////////
// dcapi.h
//
// Header to the functions to manage the Datatec Rentas Vitalicias API
//////////////////////////////////////////////////////////////////////////////

#include "../arrm.h"
#include "../wsock.h"

#define USE_DCON_MON_MESS

#if USE_LINUX || USE_SOLARIS
#define LONG long
#define DWORD unsigned int
#define Sys_Error       errno
#else
#define Sys_Error       GetLastError()
#endif

#include "../dcon.h" // this sets the alignment to 8

#include "../flddescr.h"
#include "../message.h"
#include "../field.h"
#include "../table.h"
#include "../global.h"
#include "log.h"
#include "support.h"
#include "version.h"


#ifndef _DRVAPI_H_
#define _DRVAPI_H_

#define DAPI_INITIALIZED_AS_DS   2
#define DAPI_INITIALIZED         1
#define DAPI_NOT_INITIALIZED     0

#define DAPI_ERROR_MESSAGE_LEN   200

#include <string>
using namespace std;

#define MX_ENVIRONMENT  "MX"


// CDcApi : See dcapi.cpp for implementation.
class CDcApi
{
  // Variables

  char   dest_host[60];
  short  dest_port;
  
  char  *user_name;
  char  *password;
  char  *loc_machine;
  
  string last_error_msg;
  int    uses_process_events;

  CSeqKey       cur_CSeqKey;
  Mutex_Type    mutex_lock;
  Mutex_Type    master_mutex;

  CFitFields    m_fit_fields;
  
public:
  static int  initialize_flag;
  static bool atexitdapi;
  static int  m_verbose_mode;
  
// Constructor
public:

  CDcApi();

// Implementation
protected:
	virtual ~CDcApi();

// Dispatch and event IDs
public:

  int                     Login ( char * user_name, char * password );
  int                   Connect ( char * dest_host, int dest_port, char * conf_data, int conf_code );
  int                    Logout ( void );
  int                    LogoutForce( void );
  int              GetNextField ( char *field_name, char *data_copy );
  int                  GetField ( char *field_name, char *data_copy );
  char 	             *GetSField ( char *field_name, char *data_copy );
  int              GetFieldBSTR ( char *field_name, char *data_copy );
  int              GetFieldDATE ( char *field_name, char *date_copy );
  long             GetFieldLONG ( char *field_name );
  int               InitMessage ( char *message_type );
  int   LastAcknowledgedMessage (char* date, unsigned long message_num);
  int   LastAcknowledgedMessage ( char *date, unsigned int message_num );
  int               SendMessage ( char *message_type );
  int             SendMessageTo ( char *address_data );
  //void        MarketInsertOrder ( CMsgOrder *new_msg_order );
  int              SetFieldBSTR ( char *message_type,  char *field_name, char *new_value );
  int              SetFieldDATE ( char *message_type,  char *field_name, char *new_value );
  int              SetFieldLONG ( char *message_type,  char *field_name, long  new_value );
  int              SetFieldUINT ( char *message_type,  char *field_name, unsigned int  new_value );
  int              SetFieldULONG( char *message_type,  char *field_name, unsigned long new_value);
  int              TakeMessage  ( char *message_type, unsigned long message_num);
  int              TakeMessage  ( char *message_type, unsigned int  message_num );
  int      FinishMessageProcess ( char *message_type, unsigned long message_num );
  int      FinishMessageProcess ( char *message_type, unsigned int  message_num);

  int         AddLoginCondition ( char *message_type, char *field_name, char *cond_operation, char *eq_value );
  void      CallMessageReceived ( char *message_type, unsigned long message_num);
  void      CallMessageReceived ( char *message_type, unsigned int message_num );
  void         CallOnConnection ( int result_code );
  void      CallOnDisconnection ( int reason_code, const char * reason_text );
  void          CallReportError ( int error_code, const char * error_text, int  next_action_code, int  next_action_data );
  void        CallStatusChanged ( int status_code, char * status_text );
  void         CallDataReceived ( const char *source_id, char *message_name, char *message_data );

  int           TakeMasterMutex ( void );
  void       ReleaseMasterMutex ( void );
  void          ProcessMyEvents ( void );

  // this methods should be redefined by the API user
#ifdef WIN32
  virtual void OnMessageReceived ( char *message_type, unsigned long message_num ){};
#endif
  virtual void OnMessageReceived(char* message_type, unsigned int message_num) {};
  virtual void      OnConnection ( int result_code ){};
  virtual void   OnDisconnection ( int reason_code, const char *reason_text ){};
  virtual void           OnError ( int error_code,  char *error_text, int next_action_code, int next_action_data ){};
  virtual void   OnStatusChanged ( int status_code, char *status_text ){};
  // virtual void OnDataReceived  (char * source_id, char * message_name, char * message_data){};

  int   Initialize    ( void );

public:
  int              FitConditions             ( int message_t, CMessage *CMessage_Ptr );
  int              GetCompilationType        ( char *plaftm);
  int              IsDefined                 ( char *field_name);
  int              GetVersion                ( char *vr_dapi);

protected:
  int             Locate_Record       ( char *message_type, int *message_type_code = NULL );
  int             Read_Conf_File      ( char *conf_file );
  int             Read_Tmp_Conf_File  ( char *tmp_conf_file);
  int             Process_Tmp_Config_File(void);
  int             MakeLoginReportError ( int bandera_login );
  static UNIX_PTR Process_Messages    ( void *ptr );
  static UNIX_PTR Process_DS_Messages ( void *ptr );
  void              DAPI_Clean_Tables ( void );
  char                *Get_Error_Text ( char *format, ... );
};

#define MESSAGE_RECORD_SIZE             2800
#define MESSAGE_TRANS_RECORD_SIZE       3300
#define MESSAGE_TRANS_PROCESS_SIZE       100
#define MESSAGE_TRANS_SAVE_LENGTH      (MESSAGE_TRANS_RECORD_SIZE - MESSAGE_TRANS_PROCESS_SIZE)

#define DATAMANAGER_LIST_CODE            8
#define INSTIT_TABLE_CODE                21
#define BRANCHES_TABLE_CODE              22
#define PRIMARY_USERS_TABLE_CODE         23
#define USER_TABLE_CODE                  24
#define CITIES_TABLE_CODE                25
#define COUNTRIES_TABLE_CODE             26
#define MARKETS_TABLE_CODE               27
#define MONEYS_TABLE_CODE                28
#define INSTPERM_TABLE_CODE              54
#define VOICE_TERM_GROUP_TABLE_CODE      31
#define SUPERGROUP_TABLE_CODE            30
#define EMISORES_TABLE_CODE              35
#define COMITENTES_TABLE_CODE            66
#define FIDUCIARIOS_TABLE_CODE           32026 
#define EMPLOYEE_TABLE_CODE              32053
#define ESQUEMA_TABLE_CODE               32055 
#define SUBCOMITENTE_TABLE_CODE          32141

#undef GLOBAL
#ifdef DME_GLOBAL_HERE 
#  define GLOBAL
#  define INIT(v) =(v)
#  define INIT_C(x)    (x)
#  define INIT_CS(x,y) (x,y)
#else
#  define GLOBAL extern
#  define INIT(v)
#  define INIT_C(x)
#  define INIT_CS(x,y)
#endif

GLOBAL Thread_Result glb_thread_dcon_result;
GLOBAL Thread_Result glb_thread_read_result;
GLOBAL Thread_Result glb_thread_send_result;
GLOBAL Thread_Result glb_thread_log_result;
GLOBAL Thread_Type   glb_thread_dcon;
GLOBAL Thread_Type   glb_thread_read;
GLOBAL Thread_Type   glb_thread_send;
GLOBAL Thread_Type   glb_thread_log;
GLOBAL Thread_Type   glb_master_thread;

GLOBAL  char *Glb_compilation_date  INIT(__DATE__); //The date is a string literal of the form Mmm dd yyyy.
GLOBAL  char *Glb_compilation_hour  INIT(__TIME__); //The time is a string literal of the form hh:mm:ss. 
GLOBAL  char *DAPI_delimiters       INIT(" \t\n\r");

#define USER_CODE_LEN     4
#define BRANCH_CODE_LEN   8

#define DAPI_NEXT_ACTION_CONTINUE    0
#define DAPI_NEXT_ACTION_RECONNECT   1
#define DAPI_NEXT_ACTION_MAKE_LOGON  2

#define      DAPI_ERR_INTERNAL_ERROR                 1
#define      DAPI_ERR_CONFIG_NOT_FOUND               2
#define      DAPI_ERR_CONFIG_BAD_DATA                3
#define      DAPI_ERR_NOT_INITIALIZED_YET            4
#define      DAPI_ERR_REJECTED_CONNECTION            5

#define      DAPI_ERR_RC_NOT_AUTHORIZED_USER         6
#define      DAPI_ERR_RC_USERNAME_DONOT_EXIST        7
#define      DAPI_ERR_RC_ILLEGAL_PASSWORD            8
#define      DAPI_ERR_RC_NO_DMACHINES_TRY_LATER      9
#define      DAPI_ERR_RC_DUPLICATED_MACHINE_CODE     10
#define      DAPI_ERR_RC_MACHINE_NOT_REGISTERED      11
#define      DAPI_ERR_RC_ILLEGAL_BRANCH_PASSWORD     12
#define      DAPI_ERR_STATUS_NOT_EXPECTED            13
#define      DAPI_ERR_IMPROPER_OPERATION             14
#define      DAPI_ERR_ILLEGAL_RECORD_TYPE_SPECIFIED  15
#define      DAPI_ERR_ILLEGAL_FIELD_NAME_SPECIFIED   16
#define      DAPI_ERR_ALREADY_INITIALIZED            17

#define      DAPI_ERR_RECORD_NOT_FOUND               18
#define      DAPI_ERR_CONDITION_EVALUATION           19
#define      DAPI_ERR_COULDNT_USE_LOG_FILES          20
#define      DAPI_ERR_NO_DS_DIRECTIVE                21
#define      DAPI_ERR_ILLEGAL_ADDRESS_TYPE_SPECIFIED 22
#define      DAPI_ERR_NO_RECORD_TAKEN                23

#define      DAPI_ERR_NO_FIELD_TAKEN                 24
#define      DAPI_ERR_READING_TMP_CONFIG             25  // error de lectura
#define      DAPI_ERR_EMPTY_TMP_CONFIG               26
#define      DAPI_ERR_INVALID_DATE                   27  //
#define      DAPI_ERR_CANNOT_DELETE_TMP_CONFIG       28
#define      DAPI_ERR_DATE_NOT_EXIST_DS              29

GLOBAL char *DAPI_ERR_INTERNAL_ERROR_TXT       
                                      INIT ("Internal system error %d");
GLOBAL char *DAPI_ERR_CONFIG_NOT_FOUND_TXT     
                                      INIT ("Config file %s does not exist");
GLOBAL char *DAPI_ERR_CONFIG_BAD_DATA_TXT      
                                      INIT ("Config file has incorrect line %d");
GLOBAL char *DAPI_ERR_REJECTED_CONNECTION_TXT  
                                      INIT ("The connection has been denied");
GLOBAL char *DAPI_ERR_RC_NOT_AUTHORIZED_USER_TXT
                                      INIT ("User login is not permitted");
GLOBAL char *DAPI_ERR_RC_USERNAME_DONOT_EXIST_TXT
                                      INIT ("The username does not exist");
GLOBAL char *DAPI_ERR_RC_ILLEGAL_PASSWORD_TXT
                                      INIT ("Username correct, but password is incorrect");
GLOBAL char *DAPI_ERR_RC_NO_DMACHINES_TRY_LATER_TXT
                                      INIT ("No administration machine is connected to authorize access to the system");
GLOBAL char *DAPI_ERR_RC_DUPLICATED_MACHINE_CODE_TXT
                                      INIT ("Another user is currently connected with this code");
GLOBAL char *DAPI_ERR_RC_MACHINE_NOT_REGISTERED_TXT
                                      INIT ("The computer is not registered");
GLOBAL char *DAPI_ERR_RC_ILLEGAL_BRANCH_PASSWORD_TXT
                                      INIT ("Incorrect branch password");
GLOBAL char *DAPI_ERR_STATUS_NOT_EXPECTED_TXT
                                      INIT ("The connection changed to status %d, detail (%s)");
GLOBAL char *DAPI_ERR_ILLEGAL_RECORD_TYPE_SPECIFIED_TXT
                                      INIT ("Illegal record type specified (%s)");
GLOBAL char *DAPI_ERR_RECORD_NOT_FOUND_TXT
                                      INIT ("Record (%s) number (%d) not found");
GLOBAL char *DAPI_ERR_CONDITION_EVALUATION_TXT
                                      INIT ("Error evaluating login conditions pos (%d)");
GLOBAL char *DAPI_ERR_ILLEGAL_FIELD_NAME_SPECIFIED_TXT
                                      INIT ("Illegal name field specified");
GLOBAL char *DAPI_ERR_ALREADY_INITIALIZED_TXT
                                      INIT ("The API was already initiated state %s");
GLOBAL char *DAPI_ERR_COULDNT_USE_LOG_FILES_TXT
                                      INIT ("The following log file couldn't be opened %s");
GLOBAL char *DAPI_ERR_NO_DS_DIRECTIVE_TXT
                                      INIT ("There is no \"data_server\" directive in the config file");
GLOBAL char *DAPI_ERR_ILLEGAL_ADDRESS_TYPE_SPECIFIED_TXT
                                      INIT ("Illegal address type specified (%s)");
GLOBAL char *DAPI_ERR_NO_RECORD_TAKEN_TXT
                                      INIT ("There is no record previously taken");
GLOBAL char *DAPI_ERR_NO_FIELD_TAKEN_TXT
                                      INIT ("There is no field to take from referenced table");                                      
GLOBAL char *DAPI_ERR_READING_TMP_CONFIG_TXT
                                      INIT("Reading temporal config file %s");
GLOBAL char *DAPI_ERR_EMPTY_TMP_CONFIG_TXT
                                      INIT("Empty temporal config file %s");
GLOBAL char *DAPI_ERR_TMP_CONFIG_BAD_DATA_TXT
                                      INIT("Config file (dapi_tmp_previos_date_log_request.txt) has incorrect line %d");
GLOBAL char *DAPI_ERR_INVALID_DATE_TXT
                                      INIT("Invalid date (%s) in the tmp config file (%s)");
GLOBAL char *DAPI_ERR_CANNOT_DELETE_TMP_CONFIG_TXT
                                      INIT("Can not delete tmp config file %s");
GLOBAL char *DAPI_ERR_DATE_NOT_EXIST_DS_TXT
                                      INIT("Messages asked of previous date (%s) not received. Shutdown the aplication and try again.");

#define      DAPI_NEXT_ACTION_EXIT          999

#define      ARRAY_MSG_NAMES_DIMENSION        3

// Selected Idiom:
// 0 -> Dapi Names (by default)
// 1 -> DFN Names
// 2 -> English Names
#define Glb_Old_Names       0
#define Glb_DFN_Names       1
#define Glb_English_Names   2

GLOBAL int Glb_Language     INIT (Glb_Old_Names); 


//***************************************************//
//     Nombres de mensajes definidos por el Dapi     // 
//***************************************************//
GLOBAL char *DAPI_Guibos_msg_type                 INIT ("Guibos");
GLOBAL char *DAPI_Guibos_Wh_msg_type              INIT ("GuibosWh");
GLOBAL char *DAPI_Ingreso_equipo_msg_type         INIT ("Login");
GLOBAL char *DAPI_Pedido_log_msg_type             INIT ("Log_Req");
GLOBAL char *DAPI_Inicio_log_msg_type             INIT ("Log_Begin");
GLOBAL char *DAPI_Fin_log_msg_type                INIT ("Log_End");
GLOBAL char *DAPI_Pedido_siguiente_log_msg_type   INIT ("Next_Log_Req");
GLOBAL char *DAPI_Finalpending_msg_type           INIT ("End_Pending");
GLOBAL char *DAPI_Tabla_global_msg_type           INIT ("Global_Table");
GLOBAL char *DAPI_Pedir_tablas_msg_type           INIT ("Table_Requirement");
GLOBAL char *DAPI_Fin_pedir_tablas_msg_type       INIT ("End_Table_Requirement");
GLOBAL char *DAPI_Lista_datamanagers_msg_type     INIT ("DataManagers_List");
GLOBAL char *DAPI_Trans_Moneda_msg_type           INIT ("Mantenimiento_Transacciones");
GLOBAL char *DAPI_Orden_msg_type                  INIT ("Orden_Transaccion");
GLOBAL char* DAPI_Chat_msg_type                   INIT ("Chat");
GLOBAL char* DAPI_TransComment_msg_type           INIT ("Comentario_Transaccion");
GLOBAL char *DAPI_Orden_RFQ_msg_type              INIT ("Orden_Transaccion_RFQ");
GLOBAL char *DAPI_TranmHoy_msg_type               INIT ("Precios_Montos_Transados");
GLOBAL char *DAPI_HistmHoy_msg_type               INIT ("Precios_Por_Hora");
GLOBAL char *DAPI_MntTransRentaFija_msg_type      INIT ("Mantenimiento_Transaccion_Renta_Fija");
GLOBAL char *DAPI_PayInstruc_msg_type             INIT ("Instruccion_De_Pago");
GLOBAL char *DAPI_EndTransaction_msg_type         INIT ("Fin_Transaccion");
GLOBAL char *DAPI_LogOrderTable_msg_type          INIT ("Historia_Posturas");

GLOBAL char *DAPI_Order_global_table_type         INIT ("Posturas_Vigentes");
GLOBAL char *DAPI_VoiceSummary_global_table_type  INIT ("Resumenes_Vigentes");
GLOBAL char *DAPI_VoiceGroup_global_table_type    INIT ("Grupos_Vigentes");
GLOBAL char *DAPI_FixedIncomeMarkets_global_table_type
                                                  INIT ("Mercados_Renta_Fija_Vigentes");
GLOBAL char *DAPI_HistoryTransDayTerm_type        INIT ("Hechos_Vigentes");
GLOBAL char *DAPI_Rates_Table_type                INIT ("Indices_Vigentes");

GLOBAL char *DAPI_HistmTermHoy_msg_type           INIT ("Precios_Por_Hora_Por_Plazo");
// En el ejemplo visual basic est como Precios_Por_Hora_Am_Pm, no como est definido en el dapi Precios_Am_Pm
GLOBAL char *DAPI_HistmAmPmHoy_msg_type           INIT ("Precios_Por_Hora_Am_Pm");  
GLOBAL char *DAPI_AvgPurchasePrice_msg_type       INIT ("Precios_Promedio_Compra");

//***************************************************//
//      Nombres de mensajes definidos por el DFN     // 
//***************************************************//
GLOBAL char *DAPI_Guibos_dfn_msg_type                 INIT ("mensaje_transaccion_guibos_dapi");
GLOBAL char *DAPI_Guibos_Wh_dfn_msg_type              INIT ("mensaje_transaccion_guiboswh_dapi");
GLOBAL char *DAPI_Ingreso_equipo_dfn_msg_type         INIT ("mensaje_ingreso_equipo");
GLOBAL char *DAPI_Pedido_log_dfn_msg_type             INIT ("mensaje_pedir_log");
GLOBAL char *DAPI_Inicio_log_dfn_msg_type             INIT ("mensaje_inicio_log");
GLOBAL char *DAPI_Fin_log_dfn_msg_type                INIT ("mensaje_fin_log");
GLOBAL char *DAPI_Pedido_siguiente_log_dfn_msg_type   INIT ("mensaje_siguiente_log");
GLOBAL char *DAPI_Finalpending_dfn_msg_type           INIT ("mensaje_procesar_ordenes_pendientes");
GLOBAL char *DAPI_Tabla_global_dfn_msg_type           INIT ("mensaje_cambio_tabla_global");
GLOBAL char *DAPI_Pedir_tablas_dfn_msg_type           INIT ("mensaje_pedir_tablas");
GLOBAL char *DAPI_Fin_pedir_tablas_dfn_msg_type       INIT ("mensaje_fin_pedir_tablas");
GLOBAL char *DAPI_Lista_datamanagers_dfn_msg_type     INIT ("mensaje_lista_datamanagers");
GLOBAL char *DAPI_Trans_Moneda_dfn_msg_type           INIT ("mensaje_mantenimiento_transacciones");
GLOBAL char *DAPI_Orden_dfn_msg_type                  INIT ("mensaje_tabla_moneda");
GLOBAL char *DAPI_TranmHoy_dfn_msg_type               INIT ("mensaje_precios_montos_transados_monedas");
GLOBAL char *DAPI_HistmHoy_dfn_msg_type               INIT ("mensaje_precios_por_hora_monedas");
GLOBAL char *DAPI_MntTransRentaFija_dfn_msg_type      INIT ("mensaje_mantenimiento_transaccion_renta_fija");
GLOBAL char *DAPI_HistmTermHoy_dfn_msg_type           INIT ("mensaje_precios_por_hora_por_plazo_monedas");
GLOBAL char *DAPI_TranmAmPmHoy_dfn_msg_type           INIT ("mensaje_precios_am_pm_monedas");
GLOBAL char *DAPI_PayInstruc_dfn_msg_type             INIT ("mensaje_instruccion_pago");
GLOBAL char *DAPI_EndTransaction_dfn_msg_type         INIT ("mensaje_fin_transaccion");
GLOBAL char *DAPI_LogOrderTable_dfn_msg_type          INIT ("mensaje_log_posturas");
GLOBAL char *DAPI_AvgPurchasePrice_dfn_msg_type       INIT ("mensaje_precios_promedio_compra");
GLOBAL char *DAPI_Orden_RFQ_dfn_msg_type              INIT ("mensaje_tabla_moneda_rfq");
GLOBAL char* DAPI_Chat_dfn_msg_type                   INIT ("mensaje_charla");
GLOBAL char* DAPI_Trans_Comment_dfn_msg_type          INIT ("mensaje_comentario_transaccion");

GLOBAL char *DAPI_Order_global_dfn_table_type         INIT ("Posturas_Vigentes");
GLOBAL char *DAPI_VoiceSummary_global_dfn_table_type  INIT ("Resumenes_Vigentes");
GLOBAL char *DAPI_VoiceGroup_global_dfn_table_type    INIT ("Grupos_Vigentes");
GLOBAL char *DAPI_FixedIncomeMarkets_global_dfn_table_type
                                                      INIT ("Mercados_Renta_Fija_Vigentes");
GLOBAL char *DAPI_HistoryTransDayTerm_dfn_type        INIT ("Hechos_Vigentes");
GLOBAL char *DAPI_Rates_Table_dfn_type                INIT ("Indices_Vigentes");

//***************************************************//
//      Nombres de mensajes definidos en Ingls      // 
//***************************************************//
GLOBAL char *DAPI_Guibos_english_msg_type                 INIT ("trade_guibos_dapi_message");
GLOBAL char *DAPI_Guibos_Wh_english_msg_type              INIT ("trade_guiboswh_dapi_message");
GLOBAL char *DAPI_Ingreso_equipo_english_msg_type         INIT ("primaryuser_login_message");
GLOBAL char *DAPI_Pedido_log_english_msg_type             INIT ("log_request_message");
GLOBAL char *DAPI_Inicio_log_english_msg_type             INIT ("log_start_message");
GLOBAL char *DAPI_Fin_log_english_msg_type                INIT ("log_end_message");
GLOBAL char *DAPI_Pedido_siguiente_log_english_msg_type   INIT ("next_log_request_message");
GLOBAL char *DAPI_Finalpending_english_msg_type           INIT ("process_pending_order_message");
GLOBAL char *DAPI_Tabla_global_english_msg_type           INIT ("global_table_change_message");
GLOBAL char *DAPI_Pedir_tablas_english_msg_type           INIT ("table_request_message");
GLOBAL char *DAPI_Fin_pedir_tablas_english_msg_type       INIT ("table_request_end_message");
GLOBAL char *DAPI_Lista_datamanagers_english_msg_type     INIT ("datamanager_list_message");
GLOBAL char *DAPI_Trans_Moneda_english_msg_type           INIT ("currency_trade_message");
GLOBAL char *DAPI_Orden_english_msg_type                  INIT ("currency_table_message");
GLOBAL char *DAPI_TranmHoy_english_msg_type               INIT ("currency_price_amount_traded_message");
GLOBAL char *DAPI_HistmHoy_english_msg_type               INIT ("currency_price_by_hour_message"); 
GLOBAL char *DAPI_MntTransRentaFija_english_msg_type      INIT ("fixed_income_trade_maintenance_message"); 
GLOBAL char *DAPI_HistmTermHoy_english_msg_type           INIT ("currency_price_by_hour_by_term_message"); 
GLOBAL char *DAPI_HistmAmPmHoy_english_msg_type           INIT ("currency_price_am_pm_message"); 
GLOBAL char *DAPI_PayInstruc_english_msg_type             INIT ("payment_instruction_message"); 
GLOBAL char *DAPI_EndTransaction_english_msg_type         INIT ("end_trade_message"); 
GLOBAL char *DAPI_LogOrderTable_english_msg_type          INIT ("order_log_message"); 
GLOBAL char *DAPI_AvgPurchasePrice_english_msg_type       INIT ("average_purchase_price_message"); 
GLOBAL char *DAPI_Orden_RFQ_english_msg_type              INIT ("currency_rfq_table_message");
GLOBAL char* DAPI_Chat_english_msg_type                   INIT ("chat_message");
GLOBAL char* DAPI_Trans_Comment_english_msg_type          INIT ("transaction_comment_message");

GLOBAL char *DAPI_Order_global_english_table_type             INIT ("Posturas_Vigentes");
GLOBAL char *DAPI_VoiceSummary_global_english_table_type      INIT ("Resumenes_Vigentes");
GLOBAL char *DAPI_VoiceGroup_global_english_table_type        INIT ("Grupos_Vigentes");
GLOBAL char *DAPI_FixedIncomeMarkets_global_english_table_type
                                                              INIT ("Mercados_Renta_Fija_Vigentes");
GLOBAL char *DAPI_HistoryTransDayTerm_english_type            INIT ("Hechos_Vigentes");
GLOBAL char *DAPI_Rates_Table_english_type                    INIT ("Indices_Vigentes");

//****************************************************//

#define DAPI_GLOBAL_TABLE_SIZE           6
#define ARRAY_MSG_NAMES_SIZE             26

#ifdef DME_GLOBAL_HERE
int   DAPI_Messages_Codes[] = { MSS_GUIBOS, 
                                MSS_GUIBOSWH, 
                                MSS_INGRESO_EQUIPO, 
                                MSS_PEDIDO_LOG, 
                                MSS_INICIO_LOG, 
                                MSS_FIN_LOG, 
                                MSS_PEDIDO_SIGUIENTE_LOG, 
                                MSS_FINALPENDING, 
                                MSS_TABLA_GLOBAL, 
                                MSS_PEDIR_TABLAS, 
                                MSS_FIN_PEDIR_TABLAS, 
                                MSS_LISTA_DATAMANAGERS, 
                                MSS_ORDEN,
                                MSS_TRANS_MONEDA,
                                MSS_PRCMNT_TRANS_HOY,
                                MSS_PRCMNT_HORA_HOY,
                                MSS_PAY_INSTRUCTION,
                                MSS_END_TRANSACTION,
                                MSS_LOG_ORDER_TABLE,
                                MSS_PRCMNT_HORA_PLAZO_HOY,
                                MSS_PRCMNT_HORA_AM_PM_HOY,
                                MSS_PRCMNT_TRANS_RENTA_FIJA,
                                MSS_AVGPURCHASEPRICE,
                                MSS_ORDEN_RFQ,
                                MSS_CHAT,
                                MSS_COMENTARIO_TRANSACCION};

char *DAPI_Messages_Names[ARRAY_MSG_NAMES_SIZE][ARRAY_MSG_NAMES_DIMENSION] = {
  {DAPI_Guibos_msg_type,               DAPI_Guibos_dfn_msg_type,               DAPI_Guibos_english_msg_type},
  {DAPI_Guibos_Wh_msg_type,            DAPI_Guibos_Wh_dfn_msg_type,            DAPI_Guibos_Wh_english_msg_type},
  {DAPI_Ingreso_equipo_msg_type,       DAPI_Ingreso_equipo_dfn_msg_type,       DAPI_Ingreso_equipo_english_msg_type},  
  {DAPI_Pedido_log_msg_type,           DAPI_Pedido_log_dfn_msg_type,           DAPI_Pedido_log_english_msg_type},
  {DAPI_Inicio_log_msg_type,           DAPI_Inicio_log_dfn_msg_type,           DAPI_Inicio_log_english_msg_type},
  {DAPI_Fin_log_msg_type,              DAPI_Fin_log_dfn_msg_type,              DAPI_Fin_log_english_msg_type},
  {DAPI_Pedido_siguiente_log_msg_type, DAPI_Pedido_siguiente_log_dfn_msg_type, DAPI_Pedido_siguiente_log_english_msg_type},
  {DAPI_Finalpending_msg_type,         DAPI_Finalpending_dfn_msg_type,         DAPI_Finalpending_english_msg_type},
  {DAPI_Tabla_global_msg_type,         DAPI_Tabla_global_dfn_msg_type,         DAPI_Tabla_global_english_msg_type},
  {DAPI_Pedir_tablas_msg_type,         DAPI_Pedir_tablas_dfn_msg_type,         DAPI_Pedir_tablas_english_msg_type},
  {DAPI_Fin_pedir_tablas_msg_type,     DAPI_Fin_pedir_tablas_dfn_msg_type,     DAPI_Fin_pedir_tablas_english_msg_type},
  {DAPI_Lista_datamanagers_msg_type,   DAPI_Lista_datamanagers_dfn_msg_type,   DAPI_Lista_datamanagers_english_msg_type},
  {DAPI_Orden_msg_type,                DAPI_Orden_dfn_msg_type,                DAPI_Orden_english_msg_type},
  {DAPI_Trans_Moneda_msg_type,         DAPI_Trans_Moneda_dfn_msg_type,         DAPI_Trans_Moneda_english_msg_type},
  {DAPI_TranmHoy_msg_type,             DAPI_TranmHoy_dfn_msg_type,             DAPI_TranmHoy_english_msg_type},
  {DAPI_HistmHoy_msg_type,             DAPI_HistmHoy_dfn_msg_type,             DAPI_HistmHoy_english_msg_type},
  {DAPI_PayInstruc_msg_type,           DAPI_PayInstruc_dfn_msg_type,           DAPI_PayInstruc_english_msg_type},
  {DAPI_EndTransaction_msg_type,       DAPI_EndTransaction_dfn_msg_type,       DAPI_EndTransaction_english_msg_type},
  {DAPI_LogOrderTable_msg_type,        DAPI_LogOrderTable_dfn_msg_type,        DAPI_LogOrderTable_english_msg_type},
  {DAPI_HistmTermHoy_msg_type,         DAPI_HistmTermHoy_dfn_msg_type,         DAPI_HistmTermHoy_english_msg_type},
  {DAPI_HistmAmPmHoy_msg_type,         DAPI_TranmAmPmHoy_dfn_msg_type,         DAPI_HistmAmPmHoy_english_msg_type},
  {DAPI_MntTransRentaFija_msg_type,    DAPI_MntTransRentaFija_dfn_msg_type,    DAPI_MntTransRentaFija_english_msg_type},
  {DAPI_AvgPurchasePrice_msg_type,     DAPI_AvgPurchasePrice_dfn_msg_type,     DAPI_AvgPurchasePrice_english_msg_type},
  {DAPI_Orden_RFQ_msg_type,            DAPI_Orden_RFQ_dfn_msg_type,            DAPI_Orden_RFQ_english_msg_type},
  {DAPI_Chat_msg_type,                 DAPI_Chat_dfn_msg_type,                 DAPI_Chat_english_msg_type},
  {DAPI_TransComment_msg_type,         DAPI_Trans_Comment_dfn_msg_type,        DAPI_Trans_Comment_english_msg_type},
}; 
// if you add anything  ^^^ here, change the value of >>> ARRAY_MSG_NAMES_SIZE <<<<

int   DAPI_Global_Table_Types[] = { TABLE_CODE_ID, 
                                    VOICE_SUMMARY_CODE_ID,
                                    VOICE_GROUP_CODE_ID,    
                                    FIXEDINCOMEMARKETS_TABLE_ID,  
                                    HISTORYTRANSDAYTERM_CODE_ID,  
                                    RATES_TABLE_ID    // add here and add one to DAPI_GLOBAL_TABLE_SIZE now in 6
                                    };                // when this array is changed
                                                             
char *DAPI_Global_Table_Names[][ARRAY_MSG_NAMES_DIMENSION] = { 
  {DAPI_Order_global_table_type,       DAPI_Order_global_dfn_table_type,       DAPI_Order_global_english_table_type},
  {DAPI_VoiceSummary_global_table_type,DAPI_VoiceSummary_global_dfn_table_type,DAPI_VoiceSummary_global_english_table_type},
  {DAPI_VoiceGroup_global_table_type,  DAPI_VoiceGroup_global_dfn_table_type,  DAPI_VoiceGroup_global_english_table_type}, 
  {DAPI_FixedIncomeMarkets_global_table_type, 
                                       DAPI_FixedIncomeMarkets_global_dfn_table_type, DAPI_FixedIncomeMarkets_global_english_table_type},
  {DAPI_HistoryTransDayTerm_type,      DAPI_HistoryTransDayTerm_dfn_type,      DAPI_HistoryTransDayTerm_english_type},  
  {DAPI_Rates_Table_type,              DAPI_Rates_Table_dfn_type,              DAPI_Rates_Table_english_type}};

#else
extern int   DAPI_Messages_Codes[ARRAY_MSG_NAMES_SIZE];
extern char *DAPI_Messages_Names[ARRAY_MSG_NAMES_SIZE][ARRAY_MSG_NAMES_DIMENSION];

extern int   DAPI_Global_Table_Types[DAPI_GLOBAL_TABLE_SIZE];
extern char *DAPI_Global_Table_Names[DAPI_GLOBAL_TABLE_SIZE][ARRAY_MSG_NAMES_DIMENSION]; 
#endif

//#if defined (_DEBUG) || defined (DEBUG)
#define SET_FLAG(x,y) x=(x<<4)+y
//#else
//#define SET_FLAG(x,y)
//#endif

GLOBAL char                 Gbl_Oper_File_Version   [CONF_FILE_VERSION];
GLOBAL char                 Gbl_User_Name       [10];
GLOBAL char                 Glb_User_Key        [16];
GLOBAL char                 Glb_User_Code[4];

GLOBAL char                 * Glb_log_name;
GLOBAL char                 * Glb_ds_log_name;
GLOBAL char                 * Glb_ds_put_name;
GLOBAL CGuibosMsg             Glb_cur_guibos_msg          INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGuibosWhMsg           Glb_cur_guiboswh_msg        INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgTransMon           Glb_cur_transmon_msg        INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPrcMntTrans        Glb_cur_htramhoy_msg        INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPrcMntHourTerm     Glb_cur_histmtermhoy_msg    INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPrcMntHourAmPm     Glb_cur_histmampmhoy_msg    INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPrcMntHour         Glb_cur_histmhoy_msg        INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgAvgPurchasePrice   Glb_cur_avgpurchaseprice_msg
                                                          INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPrcMntTransRentaFija Glb_cur_mnt_trans_renta_fija_msg
                                                          INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgPayInstruc         Glb_cur_payinst_msg         INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgEndTransaction     Glb_cur_endtransaction_msg  INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMsgOrder              Glb_cur_logordertable_msg   INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMessage             * Glb_cur_message_ptr         INIT_C (&Glb_cur_guibos_msg);

GLOBAL CGTInstit              Glb_cur_instit_msg          INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTCity                Glb_cur_city_msg            INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTCountry             Glb_cur_country_msg         INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTPu                  Glb_cur_primaryuser_msg     INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTMoney               Glb_cur_money_msg           INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTMarkets             Glb_cur_markets_msg         INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTInsPerm             Glb_cur_insperm_msg         INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTVoiceGroup          Glb_cur_voicegroup          INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTFixedIncomeMarkets  Glb_cur_fixedincomemarkets  INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTHistoryTransDayTerm Glb_cur_historytransdayterm INIT_C ( Glb_cur_msg_buffer);
GLOBAL CGTRates               Glb_cur_rates               INIT_C ( Glb_cur_msg_buffer);
GLOBAL CMessage             * Glb_taken_message_ptr       INIT_C (&Glb_cur_guibos_msg);
GLOBAL CGTSuperGroup          Glb_cur_supergroup_msg      INIT_C ( Glb_cur_msg_buffer);

GLOBAL int                    Glb_taken_message_type;
GLOBAL int                    Glb_taken_global_table_type;
GLOBAL int                    Glb_Log_Flag;

GLOBAL int                    Glb_run_log_testing;            // uses a log file to reproduce the received messages
GLOBAL int                    Glb_run_log_output;             // uses a log file and generate another file without compression
GLOBAL char                 * Glb_logging_test_name;          // name of the file
GLOBAL char                 * Glb_logging_test_output_name;   // output name of the testing log file
GLOBAL char                 * Glb_logging_start;              // start date-time of messages
GLOBAL char                 * Glb_logging_end;                // end   date-time of messages

GLOBAL CMsgLogReq             Glb_cur_log_req_msg  INIT_C (Glb_cur_msg_buffer);
GLOBAL CMsgNextLogReq         Glb_cur_next_log_req INIT_C (Glb_cur_msg_buffer);
GLOBAL CMsgLogBegin           Glb_cur_log_begin    INIT_C (Glb_cur_msg_buffer);
GLOBAL CMsgLogEnd             Glb_cur_log_end      INIT_C (Glb_cur_msg_buffer);

GLOBAL CGuibosMsg             Glb_send_guibos_msg;
GLOBAL CGuibosWhMsg           Glb_send_guiboswh_msg;
GLOBAL CMsgOrder              Glb_send_order_msg;
#ifdef DAPI
GLOBAL CMsgChat               Glb_send_chat_msg;
GLOBAL CMsgTransComment       Glb_send_table_comment_msg;
#endif
GLOBAL CMsgTransMon           Glb_send_transmon_msg;
GLOBAL CMsgPrcMntTrans        Glb_send_htramhoy_msg;
GLOBAL CMsgPrcMntHour         Glb_send_histmhoy_msg;
GLOBAL CMsgPrcMntHourTerm     Glb_send_htramtermhoy_msg;
GLOBAL CMsgPrcMntHourAmPm     Glb_send_histmampmhoy_msg;
GLOBAL CMsgPrcMntTransRentaFija 
                              Glb_send_mnt_trans_renta_fija_msg;
GLOBAL CMsgAvgPurchasePrice   Glb_send_average_puchase_price_msg;
                                                        
GLOBAL CMsgVrsCtrl            Glb_cur_vrsctrl_msg   INIT_C (Glb_cur_msg_buffer);
GLOBAL CMsgVrsCtrl            Glb_send_vrsctrl_msg;

GLOBAL unsigned int           Glb_last_sent_message_num;
GLOBAL char                   Glb_last_sent_date [8];

GLOBAL unsigned int           Glb_req_message_num;
GLOBAL char                   Glb_req_date [8];

GLOBAL int                    Glb_log_last_processed_message  INIT ('L');

GLOBAL char                   Glb_Prev_Date_Log_Req[10]        INIT_AN();

// por confirmar si el valor por 0 milisegundos.
GLOBAL unsigned long long      Glb_Send_Msg_Delay_Millisec    INIT (0); 

// Messages and global tables available to the interface.
GLOBAL CTable <CGuibosMsg,      CMapKeyOrder <CIdxKey,  CGuibosMsg>         ,CTableList> Glb_Guibos_Message_Table;
GLOBAL CTable <CGuibosWhMsg,    CMapKeyOrder <CIdxKey,  CGuibosWhMsg>       ,CTableList> Glb_Guibos_Wh_Message_Table;
GLOBAL CTable <CMsgOrder,       CMapKeyOrder <CIdxKey,  CMsgOrder>          ,CTableList> Glb_Order_Message_Table;
#ifdef DAPI
GLOBAL CTable <CMsgChat,        CMapKeyOrder <CIdxKey,  CMsgChat>           ,CTableList> Glb_Chat_Message_Table;
GLOBAL CTable <CMsgTransComment,CMapKeyOrder <CIdxKey,  CMsgTransComment>   ,CTableList> Glb_Transac_Comment_Message_Table;
#endif
GLOBAL CTable <CMsgTransMon,    CMapKeyOrder <CIdxKey,  CMsgTransMon>       ,CTableList> Glb_Trans_Moneda_Message_Table;
GLOBAL CTable <CMsgPrcMntTrans, CMapKeyOrder <CIdxKey,  CMsgPrcMntTrans>    ,CTableList> Glb_PrcMntTrans_Message_Table;
GLOBAL CTable <CMsgPrcMntHourTerm,  
                                CMapKeyOrder <CIdxKey,  CMsgPrcMntHourTerm> ,CTableList> Glb_PrcMntHourTerm_Message_Table;
GLOBAL CTable <CMsgPrcMntHourAmPm,  
                                CMapKeyOrder <CIdxKey,  CMsgPrcMntHourAmPm> ,CTableList> Glb_PrcMntHourAmPm_Message_Table;
GLOBAL CTable <CMsgAvgPurchasePrice,
                                CMapKeyOrder <CIdxKey,  CMsgAvgPurchasePrice>
                                                                            ,CTableList> Glb_Average_Puchase_Price_Message_Table;
GLOBAL CTable <CMsgPrcMntHour,  CMapKeyOrder <CIdxKey,  CMsgPrcMntHour>     ,CTableList> Glb_PrcMntHour_Message_Table;
GLOBAL CTable <CMsgPrcMntTransRentaFija,  
                                CMapKeyOrder <CIdxKey,  CMsgPrcMntTransRentaFija>     
                                                                            ,CTableList> Glb_PrcMnt_Trans_Renta_Fija_Table;
GLOBAL CTable <CMsgPayInstruc,  CMapKeyOrder <CIdxKey,  CMsgPayInstruc>     ,CTableList> Glb_PayInstruct_Message_Table;
GLOBAL CTable <CMsgEndTransaction,  
                                CMapKeyOrder <CIdxKey,  CMsgEndTransaction> ,CTableList> Glb_EndTransaction_Message_Table;
GLOBAL CTable <CMsgOrder,       CMapKeyOrder <CIdxKey,  CMsgOrder>          ,CTableList> Glb_LogOrderTable_Message_Table;
GLOBAL CTable <CGTOrder,        CMapKeyOrder <CIdxKey,  CGTOrder>           ,CTableList> Glb_Order_Table;
GLOBAL CTable <CGTOrder,        CMapKeyOrder <CIdxKey,  CGTOrder>           ,CTableList> Glb_Published_Order_Table;
GLOBAL CTable <CGTVoiceSummary, CMapKeyOrder <CIdxKey,  CGTVoiceSummary>    ,CTableList> Glb_VoiceSummary_Table;
GLOBAL CTable <CGTVoiceGroup,   CMapKeyOrder <CIdxKey,  CGTVoiceGroup>      ,CTableList> Glb_VoiceGroup_Table;
GLOBAL CTable <CGTFixedIncomeMarkets,  
                                CMapKeyOrder <CIdxKey,  CGTFixedIncomeMarkets>     
                                                                            ,CTableList> Glb_FixedIncomeMarkets_Table;
GLOBAL CTable <CGTHistoryTransDayTerm,  
                                CMapKeyOrder <CIdxKey,  CGTHistoryTransDayTerm>     
                                                                            ,CTableList> Glb_HistoryTransDayTerm_Table;
GLOBAL CTable <CGTRates,  
                                CMapKeyOrder <CIdxKey,  CGTRates>     
                                                                            ,CTableList> Glb_Rates_Table;

// Global tables maintained internally.
// Persistent Tables - Primary Tables
GLOBAL CTable <CGTBranches,         CMap2KeyOrder <CBranchesKey,        CEnvCrc32Key, CGTBranches>        ,CTableList> Glb_Branches_Table;
GLOBAL CTable <CGTInstit,           CMap2KeyOrder <CInstitKey,          CEnvCrc32Key, CGTInstit>          ,CTableList> Glb_Instit_Table;
GLOBAL CTable <CGTUser,             CMap2KeyOrder <CUserKey,            CEnvCrc32Key, CGTUser>            ,CTableList> Glb_User_Table;
GLOBAL CTable <CGTCity,             CMap2KeyOrder <CCityKey,            CEnvCrc32Key, CGTCity>            ,CTableList> Glb_City_Table;
GLOBAL CTable <CGTCountry,          CMap2KeyOrder <CCountryKey,         CEnvCrc32Key, CGTCountry>         ,CTableList> Glb_Country_Table;
GLOBAL CTable <CGTPu,               CMap2KeyOrder <CPuKey,              CEnvCrc32Key, CGTPu>              ,CTableList> Glb_Primaryuser_Table;
GLOBAL CTable <CGTMoney,            CMap2KeyOrder <CMoneyKey,           CEnvCrc32Key, CGTMoney>           ,CTableList> Glb_Money_Table;
GLOBAL CTable <CGTInsPerm,          CMap2KeyOrder <CPInsPermKey,        CEnvCrc32Key, CGTInsPerm>         ,CTableList> Glb_InsPerm_Table;
GLOBAL CTable <CGTMarkets,          CMap2KeyOrder <CMarketsKey,         CEnvCrc32Key, CGTMarkets>         ,CTableList> Glb_Markets_Table;
GLOBAL CTable <CGTSuperGroup,       CMap2KeyOrder <CSuperGroupKey,      CEnvCrc32Key, CGTSuperGroup>      ,CTableList> Glb_IntSuperGroup_Table;
GLOBAL CTable <CGTVoiceGroup,       CMap2KeyOrder <CVoiceGroupKey,      CEnvCrc32Key, CGTVoiceGroup>      ,CTableList> Glb_IntVoiceGroup_Table;
GLOBAL CTable <CGTEmisores,         CMap2KeyOrder <CEmisoresKey,        CEnvCrc32Key, CGTEmisores>        ,CTableList> Glb_Emisores_Table;
GLOBAL CTable <CGTComitentes,       CMap2KeyOrder <CComitentesKey,      CEnvCrc32Key, CGTComitentes>      ,CTableList> Glb_Comitentes_Table;
GLOBAL CTable <CGTFiduciarios,      CMap2KeyOrder <CFiduciariosKey,     CEnvCrc32Key, CGTFiduciarios>     ,CTableList> Glb_Fiduciarios_Table;
GLOBAL CTable <CGTEmployee,         CMap2KeyOrder <CEmployeeKey,        CEnvCrc32Key, CGTEmployee>        ,CTableList> Glb_Employee_Table;
GLOBAL CTable <CGTEsquema,          CMap2KeyOrder <CEsquemaKey,         CEnvCrc32Key, CGTEsquema>         ,CTableList> Glb_Esquema_Table;
//GLOBAL CTable <CGTSubComitente,     CMap3KeyOrder <CSubComitenteKey,
//                                                   CSubComitenteSubKey, CEnvCrc32Key, CGTSubComitente>    ,CTableList> Glb_SubComitente_Table;
GLOBAL CTable <CGTSubComitente,     CMap2KeyOrder <CSubComitenteSubKey, CEnvCrc32Key, CGTSubComitente>    ,CTableList> Glb_SubComitente_Table;


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
GLOBAL CTable <CGTFixedIncomeMarkets,  
                                CMapKeyOrder <CFixedIncomeMarketsKey,      
                                                            CGTFixedIncomeMarkets>     
                                                                            ,CTableList> Glb_IntFixedIncomeMarkets_Table;
GLOBAL CTable <CGTHistoryTransDayTerm,  
                                CMapKeyOrder <CIdxKey,      CGTHistoryTransDayTerm>     
                                                                            ,CTableList> Glb_IntHistoryTransDayTerm_Table;
GLOBAL CTable <CGTRates,  
                                CMapKeyOrder <CIdxKey,      CGTRates>     
                                                                            ,CTableList> Glb_IntRates_Table;




GLOBAL CTableList               glb_table_list;

GLOBAL char *DAPI_CONFIG_FILE_NAME INIT("dapi.cfg");
GLOBAL char *DAPI_LOG_FILE_NAME    INIT("dapi.log");
GLOBAL char *LOG_FILE_EXT          INIT(".log");
GLOBAL int                  Glb_suscribed_message[1000];

GLOBAL char *DAPI_TMP_CONFIG_FILE_NAME INIT("dapi_tmp_previous_date_log_request.txt");

typedef struct _INFO_TABLE_CODE
{
  int table_code;
  int table_type;
}INFO_TABLE_CODE;

// Vector que guarda los codigos de tablas a suscribir
GLOBAL INFO_TABLE_CODE  Glb_access_tables[MAX_PATH]; 

// this are support functions located in the support module
// the prototype is required because none of this functions are specified in the DME
// header files
void          COMMON_Send_Log_Req              (void);
void          COMMON_Send_Log_Mss              (int value);
void          tolowerstr                       (char *msg);
int           COMMON_Log_Req                   (int message_code);
void          COMMON_Send_Final_Log_Request    (void);
void          COMMON_Send_Next_Log_Req         (void);
void          COMMON_Send_Table_Reqs           (void);
unsigned int  COMMON_Get_New_Unique_Seed       (void);
int           COMMON_get_unique_code           (char *unique_code_ptr);
void          COMMON_conv_base36_alfan         (unsigned int key_num, char* unique_code_ptr);
void          COMMON_Put_Message_In_Send_Queue (CMessage *msg_to_send, 
                                                int flag_recipients, 
                                                CDestinationsList *msg_dsts,
                                                int         msg_priority = 0,
                                                const char *log_msg_text = NULL);

void     COMMON_Market_Insert_Order            (CMsgOrder *new_msg_order );
CMarket  *COMMON_Take_Market                   (int market_id);
void     LOG_Put_Err_Msg                       (const char *text_format, ...);
char     *COMMON_Get_Message_Name              (int message_code);
int      COMMON_Take_Message_Kind              (char *message_type, int *global_table_type = NULL);
int      COMMON_Get_Field                      (char *field_name,  CField **field);
int      COMMON_Get_Next_Field                 (char **field_name, CField **field);
int      COMMON_Get_SField                     (char *field_name,  CField **field);
void     COMMON_Process_Datamanagers_List      (void);
void     COMMON_Process_Global_Table_Msg       (void);
void     COMMON_Send_Global_Tables_Inscription (void);
void     COMMON_Send_Pend_Msgs                 (void);
int      COMMON_Get_RField                     (char *field_name, CField **field);
int      COMMON_Is_Processed_Record            (CMessage *CMessage_Ptr, char *extra_data);
void     COMMON_Fill_Version_General_Info      (char *Api_desc, char *Appi_Version);
void     COMMON_Global_Codes_Add               (int type, int code);
void     COMMON_Global_Codes_Clean             (void);

// SE COMENTA FUNCION NO LLAMADA EN EL CODIGO
//void    COMMON_Send_Global_Tables_Inscription_Persistentes (void);

void    SendVersionInformation                 (void);
void    Get_Market_Access_Since_Table_User     (CGTUser *Cur_CGTUser);
void    Set_Subscription_MX_Primary_Tables     (void);

#ifdef GLOBALS_HERE
CDcApi *Cur_Dapi;
#else
extern CDcApi *Cur_Dapi;
#endif

#undef GLOBAL


#endif // _DRVAPI_H_


//////////////////////////////////////////////////////////////////////////////
