///////////////////////////////////////////////////////////////////////////////////
// dscom.h
//
// module the receive the packets and send the packets data
// manages the WDF-DS header messages and multiple addressees for a packet
///////////////////////////////////////////////////////////////////////////////////
#ifndef __DSCOM_H__
#define __DSCOM_H__

# define NO_DEST               -1
# define NORMAL_DEST           0
# define DGI_DEST              1
# define DS_DEST               2
# define DMDI_DEST             4

// Message priority values, taken from dfn.
#define PRD_ING_RET_NORMALES    4  // Ingreso/retiro normales
#define PRD_MAXIMA              5  // La maxima prioridad en mensajes
#define PRD_PEDIDO_LOG          5  // Pedido de log
#define PRD_PEDIDO_TABLAS       5  // Pedido de tablas
#define PRD_INICIO_LOG_ALTA     5  // Inicio de log de alta
#define PRD_PICTURE_PRECIO      8
#define PRD_CONTROL             9  // mensajes de toma de control
#define PRD_PERDIDA_CONEXION   10  // Mensajes de prdida de conexion
#define PRD_TRANSACCION        20  // Aceptacion/Transaccion/No transaccion/Anulacion
#define PRD_MST_NO_PERSIST     15  // Maestros no persistentes
#define PRD_TABLAS_BVG         18  // Tablas globales della BVG
#define PRD_TABLAS             20  // Tablas globales
#define PRD_CONFIGURACION_ME   20  // Configuracin de puertos en ME
#define PRD_MAESTROS           30  // Mensajes de actualizacin de maestros
#define PRD_INFORMACION        35  // Mensajes de informacin de usuarios
#define PRD_CHARLA             40  // Mensajes de charla
#define PRD_PRODUCTOS          40  // Todos los mensajes relacionados con productos
#define PRD_MAESTROS_ME        50  // Actualizacin de maestros en DATASERVER
#define PRD_FIN_LOG_ALTA       75  // Fin de log de alta
#define PRD_LOG_MERCADOS       80  // Mensajes de retiro de posturas por perdida de conexion

#define MAX_DEST_ELEMENTS      300
#define NUM_DEST_LIST_TYPES    5

extern int Glb_main_connected;

class CDestinationsList
{
  int   m_target;
public:
  char  m_dest_list         [MAX_DEST_ELEMENTS][DFN_PU_CODE_LEN];       
  char  m_branch_list       [MAX_DEST_ELEMENTS][DFN_BRANCH_CODE_LEN];
  char  m_exclude_dest_list [MAX_DEST_ELEMENTS][DFN_PU_CODE_LEN];       
  char  m_instit_list       [MAX_DEST_ELEMENTS][DFN_INSTIT_CODE_LEN];
  char  m_gen_dest_list     [GENERIC_DESTINATATION_LEN];
  char  m_group_list        [MAX_DEST_ELEMENTS][DFN_GROUP_CODE_LEN];
  int   m_gen_dest_len;
  int   m_dest_elements;
  int   m_branch_elements;
  int   m_exclude_dest_elements;
  int   m_instit_elements;
  int   m_group_elements;
        char *m_adressees   [NUM_DEST_LIST_TYPES];
        int   m_numbers     [NUM_DEST_LIST_TYPES];
static  int   m_sizes       [NUM_DEST_LIST_TYPES];
static  const char *m_ids   [NUM_DEST_LIST_TYPES];

  CDestinationsList  (int target = NORMAL_DEST);
  void CleanList  (void);
  // CDestinationsList       (const char *first_dest, int num_dest);

  // Set_Fixed_Dest          (const char *first_dest, int num_dest, int size = DFN_USER_CODE_LEN);

  int  Get_Target         (void)               { return m_target;}
  void Set_Target         (int target)         { m_target = target;}
  void Set_CDestination_List (CDestinationsList *add_d);
  void Add_CDestination_List (CDestinationsList *add_d);
  void Add_Group_In_List  (const char *cur_dst_group, int num_elements);

  void Set_in_PU_List     (const char *cur_dst);
  int  Add_in_PU_List     (const char *cur_dst);
  void Set_in_Exclude_List(const char *cur_dst);
  int  Add_in_Exclude_List(const char *cur_dst);
  void Set_in_Branch_List (const char *cur_dst);
  int  Add_in_Branch_List (const char *cur_dst);
  void Set_in_Instit_List (const char *cur_dst);
  int  Add_in_Instit_List (const char *cur_dst);
  void Set_in_Gent_To_List(const char *cur_dst);
  int  Add_in_Gen_To_List (const char *cur_dst, int cur_dst_len=0);
  void Add_in_Group_List  (const char *cur_dst);

  const char **Get_Ids      (void);
  char **Get_Adressees      (void);
  int   *Get_Sizes          (void);
  int   *Get_Numbers        (void);
  char  *Get_Gen_Adress     (void);
  int    Get_Gen_Adress_Len (void);
};

void COM_Send_Pkt   (char               * packet,
                     int                  packet_len, 
                     CDestinationsList  * c_dest,
                     int                  priority,
                     const char         * text_message);

void COM_Send_Pkt2   (char              * packet,
                     int                  packet_len, 
                     CDestinationsList  * c_dest,
                     int                  priority,
                     const char         * text_message);

void COM_Get_Target_DS_List (int target, CDestinationsList **destinations_list_ptr);

#endif

///////////////////////////////////////////////////////////////////////////////////
