/////////////////////////////////////////////////////////////////////////////
// message.h
//
// CMessage definition.
/////////////////////////////////////////////////////////////////////////////

#ifndef __MESSAGE_H__
#define __MESSAGE_H__

#include "flddescr.h"
#include "arrm.h"
#include <string.h>
void FIELD_Report_No_Definition (int id, int msg);

#if USE_SOLARIS || USE_LINUX
#undef __int64
#define __int64 long long int
#endif

#ifdef DMEAPP
extern void *DME_Get_TableList    (int tablelist_id = -1, int market_id = -1); // CTableList *
extern int   DME_Get_Market_Sort_Price_Key_Args (const char *market_str, 
                                                 int        *is_voice_market,
                                                 int        *price_key_sorting_method);

extern const char  Glb_Zero_Str[];
#endif

#ifdef DAPI
extern const char  Glb_Zero_Str[];
#endif

#ifdef GDIAPP
extern const char  Glb_Zero_Str[];
extern     void  *COMMON_Get_TableList (int tablelist_id = -1, int market_id = -1);
#endif

#ifdef DGIAPP
extern const char  Glb_Zero_Str[];
#endif

//#define MSG_TYPE_USING_FIELDS    1

#define M_TRADES_TO_DO     1
#define M_ORDERS_TO_TRADE  2
#define M_ORDERS_LINKED    3
#define M_TERM_LINKED      4

#define M_TRADES_INITIAL   0

#define NOT_USE_TRANSACTION ((char *)-1)

#define M_INITIAL_FIELDS   15

#define DO_PERSISTENT_LOADING_CHANGES        1
#define AVOID_PERSISTENT_LOADING_CHANGES     0

#define PERSISTENT_RECORD_NOT_CHANGED        0
#define PERSISTENT_RECORD_UPDATED            1
#define PERSISTENT_RECORD_SHOULD_BE_DELETED  2

#define REF_INDEXED   2
#define REF_NOT_FOUND 1

class CTableBase;
class CField;
class CIdxKey;
class COrderKey;
class COrderKeyLoginName;
class CInstrumentFeedKey;
class CInstrumentKey;
class CETCPageKey;
class CPriceOrderKey;
class CNamecheckKey;
class CNmChkPlzKey;
class CPlcCreditKey;
class CCupoGKey;
class CCupoLKey;
class CCupoXPLKey;
class CEnvCrc32Key;
class CCobranzasKey;
class CFactRiesgoKey;
class CPendCobranzasKey;
class CCompartCuposKey;
class CCompartCuposMasterKey;
class CConfCuposKey;
class CLimInstitKey;
class CBuyAverPriceKey;
class CConfirmPendKey;
class CConfirmPendCodKey;
class CConfirmPendMarketKey;
class CConfigCreditKey;
class CPolicyCreditKey;
class CPosturaKey;
class CAceptacionKey;
class CTermKey;
class CTermAltKey;
class CTermMaturityKey;
class CSummaryKey;
class CGenDataKey;
class CFixPriceKey;

#ifdef GDIAPP
class CSummaryRFKey;
class CSummaryFIKey;
class CMXInstrumentKey;
class CSummaryCOKey;
class CHistoryFIKey;
#endif

class CStatisticsKey;
#if DGIAPP || GDIAPP
class CBranchesBv1Key;
class CBranchesRFQKey;
#endif
class CBranchesKey;
class CBranchesRfqsKey;
class CBranchesNumKey;
class CSpotKey;
class CUserKey;
#ifdef DAPI
class CEmisoresKey;
class CComitentesKey;
class CFiduciariosKey;
class CEmployeeKey;
class CAvgPurchasePriceKey;
class CEsquemaKey;
class CSubComitenteKey;
class CSubComitenteSubKey;
#endif
class CMarketUserKey;
class CUserBv1Key;
class CBranchUserKey;
class CPriceG7MarketKey;
class CPriceVoicemarketKey;
class CPriceG7MarketKeyLink;
class CPriceVoicemarketKeyLink;
class CPriceG7MarketKeyInvLink;
class CPriceVoicemarketKeyInvLink;
class CDatatecTradeKey;
class CCurrencyKey;
class CCashRateKey;
class CInstNumKey; 
class CPInsDME_BCKey;     
class CGTCreditAdmCodeKey;
class CSuperGroupKey;
class CVoiceTermKey;
class CVoiceSummaryKey;
class CMarketsKey;
class CSpotMarketsKey;
#ifdef DMEAPP
class CBridgeKey;
class CCounterBridgeKey;
class CBridgeSymKey;
class CIndicativesKey;
class CPrTermKey;
#endif
class CCcyPairKey;
class CInstitKey;
class CConnDmServerKey;
class CHolidayKey;
class CRFQRequestKey;
#if DAPI || DMEAPP
class CPInsPermKey;
#endif

#if DAPI || GDIAPP || DMEAPP
class CVoiceGroupKey;
#endif

#ifdef DAPI
class CSeqKey;
class CPuKey;
class CCityKey;
class CCountryKey;
class CMoneyKey;
class CFixedIncomeMarketsKey;
#endif

class CMapService;

// defines the base message class
class CMessage
{
protected:
    int           m_type;
    int           m_msg_displacement;

    CTableBase  * m_table_ptr;
    int           m_pos_in_table;
    CField      * m_crc32_data_CField; // field which covers main data for crc32 computing

public:
    // Buffer with the message data. It includes three main parts:
    // - cabeza
    // - datos_globales     
    // - datos_adicionales
    // When writting to global tables it is used only datos_globales.
    // The data is accessed using the field equivalence generated by
    // dictgen application.
    char         *m_data; 
    int           m_data_is_allocated;
    int           m_data_len;
    int           m_transaction;
    int           m_generic_use;
    unsigned int m_unic_number;
    int           m_use_transaction;
    int           m_message_id;
    DLONG         m_env_crc32_to_use;
    int           m_is_checked;
    int           m_confirm_status;
    unsigned int m_assigned_idx;   // used for crc32 adjustment protocol
    static int    m_message_pos_num [MAX_MESSAGES_TYPES];
    static int    m_message_pos     [MAX_MESSAGES_TYPES][MAX_NUMBER_FIELDS_EQUIVALENCE];
    DLONG         m_env_crc32;

    CvArray <CMessage *> m_Trades_To_Do;     // if this order is taked these are the trades to do
    CvArray <CMessage *> m_Orders_To_Trade;  // these are the order for each trade to do
    CvArray <CMessage *> m_Orders_Linked;    // these are the orders that if taken want to trade with this
    CMessage * m_Term_Linked;               // these are the orders that if taken want to trade with this

    char                m_field_init[MAX_REFERENCED_FIELDS_IN_A_MESSAGE];
    CvArray <CField  *>  m_field;

    CMessage (char *msg_data = NULL, int inc_mess_len = MESSAGE_TRANS_RECORD_SIZE);

    virtual ~CMessage();

    inline void  Set_Owner_Table_Data (CTableBase *table_ptr, int pos_in_table)
    {  
      m_table_ptr    = table_ptr;
      m_pos_in_table = pos_in_table;
    }
    void  Set_Data_Buff (char *new_buff);

    inline void  Zero_Data_Buffer(void)
    { 
      memset((char *)(m_data + MSG_NUMBER_FIELD_SIZE), 0, ((size_t)m_data_len + MESSAGE_EXTRA_ALLOCATED_SIZE - MSG_NUMBER_FIELD_SIZE)); 
      CMessage::Init();
    }
    inline void  Clean_Data_Buffer(void)
    { 
      memset((char*)(m_data + MSG_NUMBER_FIELD_SIZE), 0, ((size_t)m_data_len + MESSAGE_EXTRA_ALLOCATED_SIZE - MSG_NUMBER_FIELD_SIZE));
    }
    virtual int Do_We_Maintain_It (void);
    int   GetFieldPos (int field_id);
    void  Construct_Global_Table_Header();

    int   Init_Transaction   (char *buff_ptr, char action);
    void  Cancel_Transaction (char *buff_ptr);
    void  Commit_Transaction (char *buff_ptr);

    CMessage *Do_Begin  (int t_kind);
    CMessage *Do_Next   (int t_kind);
    void      Do_Insert (int t_kind, CMessage *to_insert);
    void      Do_Erase  (int t_kind, CMessage *to_erase);
    int       Do_Find   (int t_kind, CMessage *to_find);
    void      Do_Clean  (int t_kind);
    void      Do_End    (int t_kind);
    void      Do_Clean  (void);

    int   Get_Rec_Length(void) 
    { 
      return m_data_len + (   m_Trades_To_Do.max_number 
                            + m_Orders_To_Trade.max_number 
                            + m_Orders_Linked.max_number 
                            + m_field.max_number          ) * (int)sizeof (CMessage *); 
    }
    // for testing purpose
    void  Backup      (char *buff_ptr);
    void  Restore     (char *buff_ptr);

    int       Init_Field            (int field_id, const char *field_name = NULL);
    CField *  Init_Field_Advancing  (int field_id, const char *field_name = NULL);
    void      Init_Header_Msg       (void);
#ifdef DAPI
    CField *Find_Field  (const char *field_name);
    int     Get_Code    (const char *name);
    CField *Next_Field  (char **field_name);
#endif

    virtual void Init_Key     (CIdxKey                *idx_ptr);
    virtual void Init_Key     (CEnvCrc32Key           *key_ptr);
    virtual void Init_Key     (CPriceOrderKey         *key_ptr) {};
    virtual void Init_Key     (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char) {};
    virtual void Init_Key     (COrderKey              *key_ptr) {};
    virtual void Init_Key     (COrderKeyLoginName     *key_ptr) {};
    virtual void Init_Key     (CInstrumentFeedKey     *key_ptr) {};
    virtual void Init_Key     (CNamecheckKey          *key_ptr) {};
    virtual void Init_Key     (CNmChkPlzKey           *key_ptr) {};
    virtual void Init_Key     (CPosturaKey            *key_ptr) {};
    virtual void Init_Key     (CAceptacionKey         *key_ptr) {};
    virtual void Init_Key     (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char) {};
    virtual void Init_Key     (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char) {};
    virtual void Init_Key     (CUserKey               *key_ptr, int is_bid) {};
    virtual void Init_Key     (CSummaryKey            *key_ptr) {};
#ifdef GDIAPP    
    virtual void Init_Key     (CSummaryRFKey          *key_ptr) {};
    virtual void Init_Key     (CSummaryFIKey          *key_ptr) {};
    virtual void Init_Key     (CSummaryCOKey          *key_ptr) {};
    virtual void Init_Key     (CMXInstrumentKey       *key_ptr, int sort_type) {};
    virtual void Init_Key     (CInstrumentKey         *key_ptr) {};    
    virtual void Init_Key     (CHistoryFIKey          *key_ptr) {};    
#endif

    virtual void Init_Key     (CTermKey               *key_ptr) {};
    virtual void Init_Key     (CTermAltKey            *key_ptr) {};
    virtual void Init_Key     (CTermMaturityKey       *key_ptr) {};
    virtual void Init_Key     (CBranchesRfqsKey       *key_ptr) {};
    virtual void Init_Key     (CBranchesKey           *key_ptr) {};
    virtual void Init_Key     (CBranchesNumKey        *key_ptr) {};
    virtual void Init_Key     (CUserKey               *key_ptr) {};
#ifdef DAPI
    virtual void Init_Key     (CEmisoresKey           *key_ptr) {};
    virtual void Init_Key     (CComitentesKey         *key_ptr) {};
    virtual void Init_Key     (CFiduciariosKey        *key_ptr) {};
    virtual void Init_Key     (CEmployeeKey           *key_ptr) {};
    virtual void Init_Key     (CAvgPurchasePriceKey   *key_ptr) {};
    virtual void Init_Key     (CEsquemaKey            *key_ptr) {};
    virtual void Init_Key     (CSubComitenteKey       *key_ptr) {};
    virtual void Init_Key     (CSubComitenteSubKey    *key_ptr) {};
#endif
    virtual void Init_Key     (CMarketUserKey         *key_ptr) {};
    virtual void Init_Key     (CSummaryKey            *key_ptr, int fill_char) {};
    virtual void Init_Key     (CSuperGroupKey         *key_ptr) {};
    virtual void Init_Key     (CVoiceTermKey          *key_ptr) {};
    virtual void Init_Key     (CVoiceSummaryKey       *key_ptr) {};
    virtual void Init_Key     (CPriceG7MarketKey      *key_ptr) {};
    virtual void Init_Key     (CPriceVoicemarketKey   *key_ptr) {};
#ifdef DMEAPP
    virtual void Init_Key     (CPriceG7MarketKeyLink    *key_ptr) {};
    virtual void Init_Key     (CPriceVoicemarketKeyLink *key_ptr) {};
    virtual void Init_Key     (CPriceG7MarketKeyInvLink    *key_ptr) {};
    virtual void Init_Key     (CPriceVoicemarketKeyInvLink *key_ptr) {};
    virtual void Init_Key     (CBranchUserKey         *key_ptr) {};
    virtual void Init_Key     (CBridgeKey             *key_ptr) {};
    virtual void Init_Key     (CCounterBridgeKey      *key_ptr) {};
    virtual void Init_Key     (CBridgeSymKey          *key_ptr) {};
    virtual void Init_Key     (CIndicativesKey        *key_ptr) {};
    virtual void Init_Key     (CPrTermKey             *key_ptr) {};
#endif

#ifdef DGIAPP
    virtual void Init_Key     (CBranchUserKey         *key_ptr) {};
#endif

    virtual void Init_Key     (CCurrencyKey           *key_ptr) {};
    virtual void Init_Key     (CCashRateKey           *key_ptr) {};
    virtual void Init_Key     (CPInsDME_BCKey         *key_ptr) {};
    virtual void Init_Key     (CMarketsKey            *key_ptr) {};
    virtual void Init_Key     (CSpotMarketsKey        *key_ptr) {};

    virtual void Init_Key     (CInstNumKey            *key_ptr) {};
    virtual void Init_Key     (CConfigCreditKey       *key_ptr) {};
    virtual void Init_Key     (CPolicyCreditKey       *key_ptr) {};
    virtual void Init_Key     (CCupoGKey              *key_ptr) {};
    
    virtual void Init_Key     (CCupoLKey              *key_ptr) {};
    virtual void Init_Key     (CCupoXPLKey            *key_ptr) {};
    virtual void Init_Key     (CCobranzasKey          *key_ptr) {};
    virtual void Init_Key     (CPendCobranzasKey      *key_ptr) {};
    virtual void Init_Key     (CFactRiesgoKey         *key_ptr) {};
    virtual void Init_Key     (CCompartCuposKey       *key_ptr) {};
    virtual void Init_Key     (CCompartCuposMasterKey *key_ptr) {};
    virtual void Init_Key     (CConfCuposKey          *key_ptr) {};
    virtual void Init_Key     (CLimInstitKey          *key_ptr) {};
    virtual void Init_Key     (CBuyAverPriceKey       *key_ptr) {};
    virtual void Init_Key     (CGTCreditAdmCodeKey    *key_ptr) {};
    virtual void Init_Key     (CConfirmPendKey        *key_ptr) {};
    virtual void Init_Key     (CConfirmPendCodKey     *key_ptr) {};
    virtual void Init_Key     (CConfirmPendMarketKey  *key_ptr) {};
    virtual void Init_Key     (CGenDataKey            *key_ptr) {};
    virtual void Init_Key     (CFixPriceKey           *key_ptr) {};
    
    
#if DGIAPP || GDIAPP
    virtual void Init_Key     (CBranchesBv1Key        *key_ptr) {};
    virtual void Init_Key     (CBranchesRFQKey        *key_ptr) {};
#endif
    virtual void Init_Key     (CUserBv1Key            *key_ptr) {};
    virtual void Init_Key     (CCcyPairKey            *key_ptr) {};
    virtual void Init_Key     (CConnDmServerKey       *key_ptr) {};
    virtual void Init_Key     (CHolidayKey            *key_ptr) {};
    virtual void Init_Key     (CRFQRequestKey         *key_ptr) {};

    virtual void Init_Lnk_Key (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char) {};
    virtual void Init_Lnk_Key (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char) {};
    virtual void Init_Lnk_Key (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char) {};

    virtual void Init_Key     (CInstitKey             *key_ptr) {};
#if DAPI || DMEAPP
    virtual void Init_Key     (CPInsPermKey           *key_ptr) {};
#endif

#ifdef DAPI
  virtual void Init_Key     (CSeqKey                *key_ptr) {};
  virtual void Init_Key     (CPuKey                 *key_ptr) {};
  virtual void Init_Key     (CCityKey               *key_ptr) {};
  virtual void Init_Key     (CCountryKey            *key_ptr) {};
  virtual void Init_Key     (CMoneyKey              *key_ptr) {};
  virtual void Init_Key     (CFixedIncomeMarketsKey *key_ptr) {};
  
#else
    virtual void Init_Key     (CStatisticsKey         *key_ptr) {};
    virtual void Init_Key     (CSpotKey               *key_ptr) {};
#endif

#if DAPI || GDIAPP || DMEAPP
    virtual void Init_Key     (CVoiceGroupKey        *key_ptr) {};
#endif

    virtual void Add_Services_Map(CMapService *services_map) {};

    char *Get_Field_Init_Pointer  (int field_id);
    int   operator=(CMessage &X);

    int   Update_Message_Info     (const char *new_msg_content);
    void  Set_Current_Displacement(int displacement)
    { 
      m_msg_displacement = displacement;
    }
    virtual int Init (void)
    {  
      memset(m_field_init, 0, sizeof (m_field_init));
      return 0; 
    }
#if DMEAPP || DAPI
    int Init_Crc32 (void);
#endif
    inline  CField &Get_Field(int field_id)   
    { 
      int p_pos;
      p_pos = GetFieldPos(field_id);
      if (!m_field(p_pos))
      {
        FIELD_Report_No_Definition (field_id, m_message_id);
        Init_Field (field_id);
      }
      return *m_field[p_pos];
    }
    inline  CField &operator [](int field_id) 
    { 
      int p_pos;
      p_pos = GetFieldPos(field_id);
      if (!m_field(p_pos))
      {
        FIELD_Report_No_Definition (field_id, m_message_id);
        Init_Field (field_id);
      }
      return *m_field[p_pos];
    }
    virtual void          Take_Unic_Number(unsigned int u_n) 
    { 
      m_unic_number = u_n;
    }

#if DMEAPP || DAPI    
    virtual int           Check_Pers_Data (int from_pers_file_load) 
                          {
                            return PERSISTENT_RECORD_NOT_CHANGED;
                          }
#endif
                          
#ifdef DMEAPP
    virtual const char  * Get_Market_Ptr  (void );
    
    virtual void        * Get_TableList   (void) {return DME_Get_TableList (); }
#endif

#ifdef GDIAPP
    virtual void        * Get_TableList   (void) {return COMMON_Get_TableList (); }
#endif
  friend class CField;
};

#if USE_SOLARIS && DMDIAPP
#define USING_CMESSAGE_INIT_KEY using CMessage::Init_Key;
#else
#define USING_CMESSAGE_INIT_KEY
#endif

//////////////////////////////////////////////////////////////////////////////
// Normal messages. Order message, Login message ...
//////////////////////////////////////////////////////////////////////////////
class CMsgOrder : public CMessage
{
public:
    CMsgOrder(char *msg_data = 0x0000);
    ~CMsgOrder() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey          *idx_ptr);
#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
#endif
#if !DMDIAPP
    virtual void Init_Key (COrderKey                *key_ptr);
#endif
#if (!DAPI && !DMDIAPP)
    virtual void Init_Key (CInstrumentFeedKey       *key_ptr);
#if DGIAPP || GDIAPP
    virtual void Init_Key (CETCPageKey              *key_ptr);
    virtual void Init_Key (CInstrumentKey           *key_ptr);
#endif

#ifdef GDIAPP
    virtual void Init_Key (CSummaryRFKey            *key_ptr);
    virtual void Init_Key (CSummaryFIKey            *key_ptr);    
    virtual void Init_Key (CSummaryCOKey            *key_ptr);  
    virtual void Init_Key (CMXInstrumentKey         *key_ptr, int sort_type);    
#endif
    virtual void Init_Key (COrderKeyLoginName       *key_ptr);
    virtual void Init_Key (CPosturaKey              *key_ptr);
    virtual void Init_Key (CAceptacionKey           *key_ptr);
    //virtual void Init_Key (CBranchesRfqsKey         *key_ptr);
    virtual void Init_Key (CBranchesKey             *key_ptr);
    virtual void Init_Key (CUserKey                 *key_ptr);
    virtual void Init_Key (CMarketUserKey           *key_ptr);
    virtual void Init_Key (CPriceOrderKey           *key_ptr);
    virtual void Init_Key (CPriceG7MarketKeyLink        *key_ptr);
    virtual void Init_Key (CPriceVoicemarketKeyLink     *key_ptr);
    virtual void Init_Key (CPriceG7MarketKeyInvLink     *key_ptr);
    virtual void Init_Key (CPriceVoicemarketKeyInvLink  *key_ptr);
    virtual void Init_Key (CPriceOrderKey           *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey        *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey     *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CDatatecTradeKey         *key_ptr);
    virtual void Init_Key (CTermKey                 *key_ptr);
    virtual void Init_Key (CSummaryKey              *key_ptr);
    virtual void Init_Key (CVoiceSummaryKey         *key_ptr);
    virtual void Init_Key (CPriceG7MarketKey        *key_ptr);
    virtual void Init_Key (CPriceVoicemarketKey     *key_ptr);
    virtual void Init_Key (CCurrencyKey             *key_ptr);

    virtual void Init_Lnk_Key (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Lnk_Key (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Lnk_Key (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char);

    virtual void        Add_Services_Map      (CMapService *services_map);
    virtual const char *Get_Owner_Branch_Ptr      (int *is_broker = NULL);
    virtual const char *Get_Aggressor_Branch_Ptr  (int *is_broker = NULL);
    virtual const char *Get_Owner_User_Ptr        (void);
    virtual const char *Get_Aggressor_User_Ptr    (void);
#ifdef DMEAPP
    virtual int         Get_Order_Market_Respect_Amounts (int market_respect_amounts);
#endif
#else
#ifdef DAPI
    virtual void Init_Key (CSeqKey          *key_ptr);
#endif
#endif
public:
};

class CMsgRFQUpdate : public CMessage
{
public:
    CMsgRFQUpdate(char *msg_data = 0x0000);
    ~CMsgRFQUpdate() {}
public:
};
#ifdef DAPI
class CMsgChat : public CMessage
{
public:
    CMsgChat(char* msg_data = 0x0000);
    ~CMsgChat() {}

public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key(CIdxKey* idx_ptr);
};
class CMsgTransComment : public CMessage
{
public:
  CMsgTransComment(char* msg_data = 0x0000);
  ~CMsgTransComment() {}
public:
  USING_CMESSAGE_INIT_KEY
    virtual void Init_Key(CIdxKey* idx_ptr);
};
#endif

class CMsgLogin : public CMessage
{
public:
    CMsgLogin(char *msg_data = 0x0000);
    ~CMsgLogin() {}

public:
};

#ifdef DAPI
class CMsgLogReq : public CMessage
{
public:
    CMsgLogReq(char *msg_data = 0x0000);
    ~CMsgLogReq() {}

public:
};

class CMsgLogBegin : public CMessage
{
public:
    CMsgLogBegin(char *msg_data = 0x0000);
    ~CMsgLogBegin() {}

public:
};
class CMsgLogEnd : public CMessage
{
public:
    CMsgLogEnd(char *msg_data = 0x0000);
    ~CMsgLogEnd() {}

public:
};

class CMsgNextLogReq : public CMessage
{
public:
    CMsgNextLogReq(char *msg_data = 0x0000);
    ~CMsgNextLogReq() {}

public:
};

class CGuibosMsg : public CMessage
{
public:
    CGuibosMsg(char *msg_data = 0x0000);
    ~CGuibosMsg() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CGuibosWhMsg : public CMessage
{
public:
    CGuibosWhMsg(char *msg_data = 0x0000);
    ~CGuibosWhMsg() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};


class CMsgPrcMntTrans :public CMessage
{
public:
    CMsgPrcMntTrans(char *msg_data = 0x0000);
    ~CMsgPrcMntTrans() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CMsgPayInstruc :public CMessage
{
public:
    CMsgPayInstruc(char *msg_data = 0x0000);
    ~CMsgPayInstruc() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CMsgPrcMntHour :public CMessage
{
public:
    CMsgPrcMntHour(char *msg_data = 0x0000);
    ~CMsgPrcMntHour() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CMsgPrcMntHourTerm :public CMessage
{
public:
    CMsgPrcMntHourTerm(char *msg_data = 0x0000);
    ~CMsgPrcMntHourTerm() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CMsgPrcMntHourAmPm :public CMessage
{
public:
    CMsgPrcMntHourAmPm(char *msg_data = 0x0000);
    ~CMsgPrcMntHourAmPm() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

class CMsgAvgPurchasePrice :public CMessage
{
public:
    CMsgAvgPurchasePrice(char *msg_data = 0x0000);
    ~CMsgAvgPurchasePrice() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey                *idx_ptr);
    virtual void Init_Key (CAvgPurchasePriceKey   *key_ptr);
};

class CMsgPrcMntTransRentaFija :public CMessage
{
public:
    CMsgPrcMntTransRentaFija(char *msg_data = 0x0000);
    ~CMsgPrcMntTransRentaFija() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
    virtual void Init_Key (CSeqKey        *key_ptr);
};

#endif

#if (DAPI || DMEAPP || DGIAPP || GDIAPP)

//CMsgEqOffLine
class CMsgEqOffLine : public CMessage
{
public:
    CMsgEqOffLine(char *msg_data = 0x0000);
    ~CMsgEqOffLine() {}
public:
};
#endif

#ifdef DAPI
class CMsgVrsCtrl : public CMessage
{
public:
    CMsgVrsCtrl(char *msg_data = 0x0000);
    ~CMsgVrsCtrl() {}
public:
    USING_CMESSAGE_INIT_KEY
};
#endif

#if DAPI || DGIAPP
class CMsgTransMon : public CMessage
{
public:
    CMsgTransMon(char *msg_data = 0x0000);
    ~CMsgTransMon() {}
public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
#ifdef DAPI
    virtual void Init_Key (CSeqKey        *key_ptr);
#endif
};

#endif

#if DMEAPP || DAPI

class CGTConnUser: public CMessage
{
public:

    CGTConnUser(char *msg_data = 0x0000);
    ~CGTConnUser() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CBranchUserKey *key_ptr);
public:
};

#endif

class CMsgEndTransaction : public CMessage
{
public:
    CMsgEndTransaction(char *msg_data = 0x0000);
    ~CMsgEndTransaction() {}

public:
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey        *idx_ptr);
#ifdef DAPI
    virtual void Init_Key (CSeqKey        *key_ptr);
#endif
};

class CMsgDMList : public CMessage
{
public:
    CMsgDMList(char *msg_data = 0x0000);
    ~CMsgDMList() {}

public:
};

class CMsgDMTList : public CMessage
{
public:
    CMsgDMTList(char *msg_data = 0x0000);
    ~CMsgDMTList() {}

public:
};

class CMsgETCResponse : public CMessage
{
public:
    CMsgETCResponse(char *msg_data = 0x0000);
    ~CMsgETCResponse() {}

public:
};

class CMsgAllCrpoliNotification: public CMessage
{
public:
    CMsgAllCrpoliNotification(char *msg_data = 0x0000);
    ~CMsgAllCrpoliNotification() {}

public:
};

class CMsgCommandHistory: public CMessage
{
public:
    CMsgCommandHistory(char *msg_data = 0x0000);
    ~CMsgCommandHistory() {}

public:
};

//////////////////////////////////////////////////////////////////////////////
// Global tables.
//////////////////////////////////////////////////////////////////////////////
class CGTHeader: public CMessage
{
public:
    CGTHeader(char *msg_data = 0x0000);
    ~CGTHeader() {}
public:
};

class CMsgSubcripReq: public CMessage
{
public:
  CMsgSubcripReq(char *msg_data = 0x0000);
  ~CMsgSubcripReq() {}
};

class CMsgTablesReq : public CMessage
{
public:
    CMsgTablesReq(char *msg_data = 0x0000);
    ~CMsgTablesReq() {}

public:
};

class CGTBranches: public CMessage
{
public:
    CGTBranches(char *msg_data = 0x0000);
    ~CGTBranches() {}
    USING_CMESSAGE_INIT_KEY
#if DGIAPP || GDIAPP
    virtual void Init_Key  (CBranchesBv1Key *key_ptr);
    virtual void Init_Key  (CBranchesRFQKey *key_ptr);
#endif
    virtual void Init_Key  (CBranchesKey    *key_ptr);
#ifdef DAPI        
    virtual int  Init      (void);    
#endif    
public:
};


class CGTBranchesRfqs: public CMessage
{
public:
    CGTBranchesRfqs(char *msg_data = 0x0000);
    ~CGTBranchesRfqs() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key  (CBranchesRfqsKey    *key_ptr);
public:
};

class CGTUser: public CMessage
{
public:

    CGTUser(char *msg_data = 0x0000);
    ~CGTUser() {}
    USING_CMESSAGE_INIT_KEY
#ifdef DAPI        
    virtual int  Init     (void);   
#endif    
    virtual void Init_Key (CUserBv1Key *key_ptr);
    virtual void Init_Key (CUserKey *key_ptr);
    virtual void Init_Key (CBranchesKey *key_ptr);
    virtual void Init_Key (CBranchUserKey *key_ptr);
public:
};

#ifdef DAPI
class CGTEmisores: public CMessage
{
public:

    CGTEmisores(char *msg_data = 0x0000);
    ~CGTEmisores() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CEmisoresKey *key_ptr);
    virtual int  Init     (void);   
public:
};
class CGTComitentes: public CMessage
{
public:

    CGTComitentes(char *msg_data = 0x0000);
    ~CGTComitentes() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CComitentesKey *key_ptr);
    virtual int  Init     (void);   
public:
};
class CGTSubComitente: public CMessage
{
public:

    CGTSubComitente(char *msg_data = 0x0000);
    ~CGTSubComitente() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSubComitenteKey    *key_ptr);
    virtual void Init_Key (CSubComitenteSubKey *key_ptr);
    virtual int  Init     (void);   
public:
};
class CGTFiduciarios: public CMessage
{
public:

    CGTFiduciarios(char *msg_data = 0x0000);
    ~CGTFiduciarios() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CFiduciariosKey *key_ptr);
    virtual int  Init     (void);   
public:
};

class CGTEmployee: public CMessage
{
public:

    CGTEmployee(char *msg_data = 0x0000);
    ~CGTEmployee() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CEmployeeKey *key_ptr);
    virtual int  Init     (void);   
public:
};

class CGTEsquema: public CMessage
{
public:

    CGTEsquema(char *msg_data = 0x0000);
    ~CGTEsquema() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CEsquemaKey *key_ptr);
    virtual int  Init     (void);   
public:
};

#endif

class CGTMarketsOld: public CMessage
{
public:
    CGTMarketsOld(char *msg_data = 0x0000);
    ~CGTMarketsOld() {}
public:
};


class CGTMarkets: public CMessage
{
public:
    CGTMarkets(char *msg_data = 0x0000);
    ~CGTMarkets() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMarketsKey *key_ptr);
    virtual void Init_Key (CSpotMarketsKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);
#endif    
#if DGIAPP || GDIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
#endif
public:
};

class CGTInstit: public CMessage
{
public:

    CGTInstit(char *msg_data = 0x0000);
    ~CGTInstit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CInstitKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);    
#endif    
public:
};

#if DAPI || DMEAPP
class CGTInsPerm: public CMessage
{
public:
    CGTInsPerm(char *msg_data = 0x0000);
    ~CGTInsPerm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPInsPermKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);   
#endif    
public:
};

#endif

#if DAPI || GDIAPP || DMEAPP
class CGTVoiceGroup: public CMessage
{
public:
    CGTVoiceGroup(char *msg_data = 0x0000);
    ~CGTVoiceGroup() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey *idx_ptr);
    virtual void Init_Key (CVoiceGroupKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);
#endif    
public:
};
#endif


#ifdef DAPI

class CGTHistoryTransDayTerm: public CMessage
{
public:
    CGTHistoryTransDayTerm(char *msg_data = 0x0000);
    ~CGTHistoryTransDayTerm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey *idx_ptr);

public:
};

class CGTPu: public CMessage
{
public:
    CGTPu(char *msg_data = 0x0000);
    ~CGTPu() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPuKey     *key_ptr);
    virtual int  Init     (void);   

public:
};

class CGTCity: public CMessage
{
public:
    CGTCity(char *msg_data = 0x0000);
    ~CGTCity() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCityKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void); 
#endif    
public:
};

class CGTCountry: public CMessage
{
public:
    CGTCountry(char *msg_data = 0x0000);
    ~CGTCountry() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCountryKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);   
#endif    

public:
};

class CGTMoney: public CMessage
{
public:
    CGTMoney(char *msg_data = 0x0000);
    ~CGTMoney() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMoneyKey *key_ptr);
    virtual int  Init     (void);   
    
public:
};

class CGTFixedIncomeMarkets: public CMessage
{
public:
    CGTFixedIncomeMarkets(char *msg_data = 0x0000);
    ~CGTFixedIncomeMarkets() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey *idx_ptr);
    virtual void Init_Key (CFixedIncomeMarketsKey *key_ptr);
public:
};

class CGTRates: public CMessage
{
public:
    CGTRates(char *msg_data = 0x0000);
    ~CGTRates() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey *idx_ptr);
public:
};

#endif

//CConfigCredit
class CConfigCredit : public CMessage
{
public:
    CConfigCredit(char *msg_data = 0x0000);
    ~CConfigCredit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CConfigCreditKey *key_ptr);    
#ifdef DMEAPP
    virtual int Check_Pers_Data (int from_pers_file_load);
#endif
public:
};

//CMsgOneCfgCredit
class CMsgOneCfgCredit : public CMessage
{
public:
    CMsgOneCfgCredit(char *msg_data = 0x0000);
    ~CMsgOneCfgCredit() {}
public:
};

class CGTOrder : public CMessage
{
public:
    CGTOrder(char *msg_data = 0x0000);
    ~CGTOrder() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey              *idx_ptr);
    virtual void Init_Key (COrderKey            *key_ptr);
    virtual void Init_Key (CInstrumentFeedKey   *key_ptr);
    virtual void Init_Key (COrderKeyLoginName   *key_ptr);

#if DGIAPP || GDIAPP
    virtual void Init_Key (CETCPageKey          *key_ptr);
    virtual void Init_Key (CInstrumentKey       *key_ptr);
#endif

#ifdef GDIAPP
    virtual void Init_Key (CMXInstrumentKey     *key_ptr, int sort_type);
    virtual void Init_Key (CSummaryRFKey        *key_ptr);
    virtual void Init_Key (CSummaryFIKey        *key_ptr);
    virtual void Init_Key (CSummaryCOKey        *key_ptr);    
#endif
    virtual void Init_Key (CPosturaKey          *key_ptr);
    //virtual void Init_Key (CBranchesRfqsKey     *key_ptr);
    virtual void Init_Key (CBranchesKey         *key_ptr);
    virtual void Init_Key (CUserKey             *key_ptr);
    virtual void Init_Key (CMarketUserKey       *key_ptr);
    // Price key used by dgi.
    virtual void Init_Key (CSummaryKey          *key_ptr);
    virtual void Init_Key (CTermKey             *key_ptr);
    virtual void Init_Key (CVoiceSummaryKey     *key_ptr);
    virtual void Init_Key (CPriceOrderKey       *key_ptr);
    virtual void Init_Key (CPriceG7MarketKey    *key_ptr);
    virtual void Init_Key (CPriceVoicemarketKey *key_ptr);

#if DGIAPP || DMEAPP
    virtual void Init_Key (CCurrencyKey         *key_ptr);
#endif
    virtual void Init_Key (CPriceOrderKey       *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey    *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey *key_ptr, const char *bid_offer, int fill_char);

    virtual void Add_Services_Map (CMapService *services_map);

#ifdef DMEAPP
    virtual const char *Get_Market_Ptr     (void );
    virtual int         Get_Order_Market_Respect_Amounts (int market_respect_amounts);
#endif
public:
};

class CGTVoiceSummary: public CMessage
{
public:
    CGTVoiceSummary(char *msg_data = 0x0000);
    ~CGTVoiceSummary() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey          *idx_ptr);
    virtual void Init_Key (CVoiceSummaryKey *key_ptr);
#ifdef GDIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
    virtual void Init_Key (CMXInstrumentKey *key_ptr, int sort_type);
#endif

#ifdef DMEAPP
    virtual void  *Get_TableList  (void);
#endif
public:
};

class CGTSuperGroup: public CMessage
{
public:
    CGTSuperGroup(char *msg_data = 0x0000);
    ~CGTSuperGroup() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSuperGroupKey *key_ptr);
#ifdef DAPI        
    virtual int  Init     (void);
#endif    
public:
};

#ifdef DMEAPP

class CGTBridge: public CMessage
{
public:
    CGTBridge(char *msg_data = 0x0000);
    ~CGTBridge() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CBridgeKey     * key_ptr);
    virtual void Init_Key (CBridgeSymKey  * key_ptr);
    virtual int  Check_Pers_Data (int from_pers_file_load);
public:
};

class CGTCounterBridge: public CMessage
{
public:
    CGTCounterBridge(char *msg_data = 0x0000);
    ~CGTCounterBridge() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCounterBridgeKey     * key_ptr);
    virtual int  Check_Pers_Data (int from_pers_file_load);
public:
};

class CMsgIndicatives: public CMessage
{
public:
    CMsgIndicatives(char *msg_data = 0x0000);
    ~CMsgIndicatives() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIndicativesKey *key_ptr);
public:
};

class CGTIndicatives: public CMessage
{
public:
    CGTIndicatives(char *msg_data = 0x0000);
    ~CGTIndicatives() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIndicativesKey *key_ptr);

    virtual void *Get_TableList    (void);
    virtual int   Check_Pers_Data  (int from_pers_file_load);

public:
};

class CMsgSpot: public CMessage
{
public:
    CMsgSpot (char *msg_data = 0x0000);
    ~CMsgSpot() {}
};

class CGTBridgeData: public CMessage
{
public:
    CGTBridgeData(char *msg_data = 0x0000);
    ~CGTBridgeData() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPrTermKey *key_ptr);
};

class CMsgBridge: public CMessage
{
public:
    CMsgBridge(char *msg_data = 0x0000);
    ~CMsgBridge() {}
    USING_CMESSAGE_INIT_KEY
     virtual void Init_Key (CBridgeKey *key_ptr);
};

class CGTHistoryTransDayMem: public CMessage
{
public:
    CGTHistoryTransDayMem(char *msg_data = 0x0000);
    ~CGTHistoryTransDayMem() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMarketsKey *key_ptr);

public:
};

#endif

#if DMEAPP || DMDIAPP
class CMsgRBridge: public CMessage
{
public:
    CMsgRBridge(char *msg_data = 0x0000);
    ~CMsgRBridge() {}
public:
};

class CMsgCmdBridge: public CMessage
{
public:
    CMsgCmdBridge(char *msg_data = 0x0000);
    ~CMsgCmdBridge() {}
public:
};
#endif

#ifdef DMDIAPP
class CMsgUserLoginBridge: public CMessage
{
public:
    CMsgUserLoginBridge(char *msg_data = 0x0000);
    ~CMsgUserLoginBridge() {}
public:
};
#endif

#if (!DAPI && !DMDIAPP)
class CMsgFinalPending : public CMessage
{
public:
    CMsgFinalPending(char *msg_data = 0x0000);
    ~CMsgFinalPending() {}

public:
};

class CMsgTradeNeg : public CMessage
{
public:
    CMsgTradeNeg(char *msg_data = 0x0000);
    ~CMsgTradeNeg() {}

    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CDatatecTradeKey *key_ptr, int change_c_v_flag);
};

class CMsgTextTradeNeg : public CMessage
{
public:
    CMsgTextTradeNeg(char *msg_data = 0x0000);
    ~CMsgTextTradeNeg() {}

    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CDatatecTradeKey *key_ptr, int change_c_v_flag);
};

class CGTTerm: public CMessage
{
public:
    CGTTerm(char *msg_data = 0x0000);
    ~CGTTerm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CTermKey               *key_ptr);
    virtual void Init_Key (CTermAltKey            *key_ptr);
    virtual void Init_Key (CTermMaturityKey       *key_ptr);
    virtual void Init_Key (CSummaryKey            *key_ptr);
    virtual void Init_Key (CSummaryKey            *key_ptr, int fill_char);
    virtual void Init_Key (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char);

#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
#endif

public:
};

#ifdef GDIAPP

class CGTHistoryFI: public CMessage
{
public:
    CGTHistoryFI(char *msg_data = 0x0000);
    ~CGTHistoryFI() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CHistoryFIKey  *key_ptr);     
    virtual void Init_Key (CSummaryCOKey  *key_ptr);  
    virtual void Init_Key (CETCPageKey    *key_ptr);     
};

class CGTSummaryFI: public CMessage
{
public:
    CGTSummaryFI(char *msg_data = 0x0000);
    ~CGTSummaryFI() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSummaryFIKey  *key_ptr);
    virtual void Init_Key (CETCPageKey    *key_ptr);
    virtual void Init_Key (CInstrumentKey *key_ptr);
    virtual void Init_Key (CSummaryCOKey  *key_ptr);
};

class CGTSummaryRF: public CMessage
{
public:
    CGTSummaryRF(char *msg_data = 0x0000);
    ~CGTSummaryRF() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSummaryRFKey *key_ptr);    
};
#endif

class CGTSummary: public CMessage
{
public:
    CGTSummary(char *msg_data = 0x0000);
    ~CGTSummary() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSummaryKey          *key_ptr);
    virtual void Init_Key (CTermKey             *key_ptr);
#if DGIAPP || GDIAPP
    virtual void Init_Key (CETCPageKey          *key_ptr);
    virtual void Init_Key (CInstrumentKey       *key_ptr);
#endif
    virtual void Init_Key (CUserKey             *key_ptr, int is_bid);
    virtual void Init_Key (CPriceOrderKey       *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey    *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey *key_ptr, const char *bid_offer, int fill_char);
#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
    virtual void       *Get_TableList  (void);
#endif

public:
};

#if DGIAPP || DMEAPP
class CMsgHandover: public CMessage
{
public:
    CMsgHandover(char *msg_data = 0x0000);
    ~CMsgHandover() {}
public:
};
#endif

class CGTStatistics: public CMessage
{
public:
    CGTStatistics(char *msg_data = 0x0000);
    ~CGTStatistics() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CStatisticsKey *key_ptr);
    virtual void Init_Key (CSummaryKey *key_ptr, int fill_char);
#ifdef DGIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
#endif
public:
};

class CGTSpot: public CMessage
{
public:
    CGTSpot(char *msg_data = 0x0000);
    ~CGTSpot() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CSpotKey *key_ptr);
#ifdef DGIAPP
    virtual void Init_Key (CETCPageKey    *key_ptr);
    virtual void Init_Key (CInstrumentKey *key_ptr);
#endif
public:
};

class CGTVoiceTerm: public CMessage
{
public:
    CGTVoiceTerm(char *msg_data = 0x0000);
    ~CGTVoiceTerm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CVoiceTermKey          *key_ptr);
    virtual void Init_Key (CVoiceSummaryKey       *key_ptr);
    virtual void Init_Key (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char);
#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
#endif
public:
};

class CMsgVoiceTerm: public CMessage
{
public:
    CMsgVoiceTerm(char *msg_data = 0x0000);
    ~CMsgVoiceTerm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CVoiceTermKey          *key_ptr);
    virtual void Init_Key (CVoiceSummaryKey       *key_ptr);
    virtual void Init_Key (CPriceOrderKey         *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey      *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey   *key_ptr, const char *bid_offer, int fill_char);
#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
#endif
public:
};

class CMsgDelDeferred : public CMessage
{
public:
    CMsgDelDeferred(char *msg_data = 0x0000);
    ~CMsgDelDeferred() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPriceOrderKey       *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey    *key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey *key_ptr, const char *bid_offer, int fill_char);
#ifdef DMEAPP
    virtual const char *Get_Market_Ptr (void );
#endif
public:
};

class CGTCashRate: public CMessage
{
public:
    CGTCashRate(char *msg_data = 0x0000);
    ~CGTCashRate() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCashRateKey *key_ptr);
public:
};

class CGTOvernight: public CMessage
{
public:
    CGTOvernight(char *msg_data = 0x0000);
    ~CGTOvernight() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCashRateKey *key_ptr);
public:
};

class CGTDMEConnected: public CMessage
{
public:
    CGTDMEConnected(char *msg_data = 0x0000);
    ~CGTDMEConnected() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMarketUserKey *key_ptr);
public:
};

class CGTIncrements: public CMessage
{
public:
    CGTIncrements(char *msg_data = 0x0000);
    ~CGTIncrements() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CTermKey *key_ptr);
#ifdef DGIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
#endif
public:
};

class CGTCurrency: public CMessage
{
public:
    CGTCurrency(char *msg_data = 0x0000);
    ~CGTCurrency() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCurrencyKey *key_ptr);
public:
};

//CMsgNameCheck
class CMsgNameCheck : public CMessage
{
public:
    CMsgNameCheck(char *msg_data = 0x0000);
    ~CMsgNameCheck() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CNamecheckKey *key_ptr);    
public:
};

class CGTNamecheck: public CMessage
{
public:
    CGTNamecheck(char *msg_data = 0x0000);
    ~CGTNamecheck() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CNamecheckKey *key_ptr);

public:
};

//CMsgNmChkPlz  (previous CMsgFreeRisk)
class CMsgNmChkPlz : public CMessage
{
public:
    CMsgNmChkPlz(char *msg_data = 0x0000);
    ~CMsgNmChkPlz() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CNmChkPlzKey *key_ptr);    
public:
};

class CGTInstNum: public CMessage
{
public:
    CGTInstNum(char *msg_data = 0x0000);
    ~CGTInstNum() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CInstNumKey *key_ptr);
    virtual void Take_Unic_Number  (unsigned int u_n);
#ifdef DMEAPP
    virtual int  Init (void);
#endif
public:
};

class CGTNmChkPlz: public CMessage
{
public:
    CGTNmChkPlz(char *msg_data = 0x0000);
    ~CGTNmChkPlz() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CNmChkPlzKey *key_ptr);
#ifdef DMEAPP
    virtual void *Get_TableList    (void);
#endif

public:
};

class CGTBranchNum: public CMessage
{
public:
    CGTBranchNum(char *msg_data = 0x0000);
    ~CGTBranchNum() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CBranchesKey *key_ptr);
    virtual void Init_Key (CBranchesNumKey *key_ptr);
    virtual void Take_Unic_Number  (unsigned int u_n);
#ifdef DMEAPP
    virtual int  Init (void);
#endif
public:
};

class CGTPInsDme: public CMessage
{
public:
    CGTPInsDme(char *msg_data = 0x0000);
    ~CGTPInsDme() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPInsDME_BCKey *key_ptr);
public:
};

class CMsgPInsDme: public CMessage
{
public:
    CMsgPInsDme(char *msg_data = 0x0000);
    ~CMsgPInsDme() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPInsDME_BCKey *key_ptr);
public:
};

class CGTCreditAdm: public CMessage
{
public:
    CGTCreditAdm(char *msg_data = 0x0000);
    ~CGTCreditAdm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CGTCreditAdmCodeKey *key_ptr);
#ifdef DMEAPP
    virtual void *Get_TableList    (void);
#endif
public:
};

//CMsgCupo
class CMsgCupoG : public CMessage
{
public:
    CMsgCupoG(char *msg_data = 0x0000);
    ~CMsgCupoG() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoGKey     *key_ptr);    
public:
};

class CGTCuposG: public CMessage
{
public:
    CGTCuposG(char *msg_data = 0x0000);
    ~CGTCuposG() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoGKey     *key_ptr);
#ifdef DMEAPP
    virtual void *Get_TableList    (void);
#endif

public:
};

//CMsgCupoL
class CMsgCupoL : public CMessage
{
public:
    CMsgCupoL(char *msg_data = 0x0000);
    ~CMsgCupoL() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoLKey      *key_ptr);
    virtual void Init_Key (CCupoXPLKey    *key_ptr);
public:
};

//CMsgCupoXPL
class CMsgCupoXPL : public CMessage
{
public:
    CMsgCupoXPL(char *msg_data = 0x0000);
    ~CMsgCupoXPL() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoXPLKey    *key_ptr);    
public:
};

#ifdef DMEAPP

class CGTCuposXPL: public CMessage
{
public:
    CGTCuposXPL(char *msg_data = 0x0000);
    ~CGTCuposXPL() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoXPLKey  *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);

public:
};

class CGTCuposL: public CMessage
{
public:
    CGTCuposL(char *msg_data = 0x0000);
    ~CGTCuposL() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoLKey     *key_ptr);
    virtual void Init_Key (CCupoXPLKey   *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);

public:
};

class CGTCuposXDIA: public CMessage
{
public:
    CGTCuposXDIA(char *msg_data = 0x0000);
    ~CGTCuposXDIA() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCupoXPLKey   *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);
public:
};

#endif

class CGTCobranzas: public CMessage
{
public:
    CGTCobranzas(char *msg_data = 0x0000);
    ~CGTCobranzas() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCobranzasKey  *key_ptr);
public:
};

#ifdef DMEAPP
class CGTPendCobranzas: public CMessage
{
public:
    CGTPendCobranzas(char *msg_data = 0x0000);
    ~CGTPendCobranzas() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPendCobranzasKey  *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);

public:
};
#endif

// CMsgPendCobranzas
class CMsgPendCobranzas : public CMessage
{
public:
    CMsgPendCobranzas(char *msg_data = 0x0000);
    ~CMsgPendCobranzas() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPendCobranzasKey     *key_ptr);    
public:
};

class CMsgFactRiesgo : public CMessage
{
public:
    CMsgFactRiesgo(char *msg_data = 0x0000);
    ~CMsgFactRiesgo() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CFactRiesgoKey     *key_ptr);    
public:
};

#ifdef DMEAPP
class CGTFactRiesgo: public CMessage
{
public:
    CGTFactRiesgo(char *msg_data = 0x0000);
    ~CGTFactRiesgo() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CFactRiesgoKey *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);
public:
};

class CGTCompartCupos: public CMessage
{
public:
    CGTCompartCupos(char *msg_data = 0x0000);
    ~CGTCompartCupos() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCompartCuposKey       * key_ptr);
    virtual void Init_Key (CCompartCuposMasterKey * key_ptr);
    virtual int  Do_We_Maintain_It (void);
#ifdef DMEAPP
    virtual int  Init (void);
#endif
public:
};

class CGTConfCupos: public CMessage
{
public:
    CGTConfCupos(char *msg_data = 0x0000);
    ~CGTConfCupos() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CConfCuposKey *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
    virtual int  Check_Pers_Data (int from_pers_file_load);
public:
};
#endif

//CMsgLimInstit
class CMsgLimInstit : public CMessage
{
public:
    CMsgLimInstit(char *msg_data = 0x0000);
    ~CMsgLimInstit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CLimInstitKey     *key_ptr);    
public:
};

#ifdef DMEAPP
class CGTLimInstit: public CMessage
{
public:
    CGTLimInstit(char *msg_data = 0x0000);
    ~CGTLimInstit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CLimInstitKey  *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
public:
};
#endif

class CGTBuyAverPrice: public CMessage
{
public:
    CGTBuyAverPrice(char *msg_data = 0x0000);
    ~CGTBuyAverPrice() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CBuyAverPriceKey  *key_ptr);
#ifdef DMEAPP
    virtual void *Get_TableList    (void);
#endif

public:
};

#ifdef DMEAPP
class CGTMarketSpotPrices: public CMessage
{
public:
    CGTMarketSpotPrices(char *msg_data = 0x0000);
    ~CGTMarketSpotPrices() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMarketsKey     *key_ptr);
    virtual int  Init (void);
    virtual int  Do_We_Maintain_It (void);
public:
};
#endif

class CGTConfirmPend: public CMessage
{
public:
    CGTConfirmPend(char *msg_data = 0x0000);
    ~CGTConfirmPend() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CConfirmPendKey       *key_ptr);
    virtual void Init_Key (CConfirmPendCodKey    *cod_key_ptr);
    virtual void Init_Key (CConfirmPendMarketKey *cod_key_ptr);

public:
};

class CGTGenData: public CMessage
{
public:
    CGTGenData(char *msg_data = 0x0000);
    ~CGTGenData() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CGenDataKey  *key_ptr);
};

#ifdef DMEAPP
// CMsgFixPrice
class CMsgFixPrice : public CMessage
{
public:
    CMsgFixPrice(char *msg_data = 0x0000);
    ~CMsgFixPrice() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CIdxKey              * idx_ptr);
    virtual void Init_Key (CFixPriceKey         * key_ptr);
    virtual void Init_Key (CPriceOrderKey       * key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceG7MarketKey    * key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CPriceVoicemarketKey * key_ptr, const char *bid_offer, int fill_char);
    virtual void Init_Key (CSummaryKey          * key_ptr);
    virtual const char *  Get_Market_Ptr (void);
public:
};
#endif

//CMsgConfigCredit
class CMsgConfigCredit : public CMessage
{
public:
    CMsgConfigCredit(char *msg_data = 0x0000);
    ~CMsgConfigCredit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CConfigCreditKey *key_ptr);   
public:
};

//CMsgOnePlcCredit
class CMsgOnePlcCredit : public CMessage
{
public:
    CMsgOnePlcCredit(char *msg_data = 0x0000);
    ~CMsgOnePlcCredit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPlcCreditKey *key_ptr);    
#ifdef DGIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
#endif
public:
};


//CMsgPolicyCredit PU
class CMsgPolicyCredit : public CMessage
{
public:
    CMsgPolicyCredit(char *msg_data = 0x0000);
    ~CMsgPolicyCredit() {}
    // virtual void Init_Key (CPolicyCreditKey *key_ptr);    
#ifdef DGIAPP
    virtual void Init_Key (CETCPageKey *key_ptr);
#endif
public:
};

//CPolicyCredit PU
class CPolicyCredit : public CMessage
{
public:
    unsigned int m_branch_num;
  
    CPolicyCredit(char *msg_data = 0x0000);
    ~CPolicyCredit() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CPolicyCreditKey *key_ptr);
#ifdef DMEAPP
    virtual int   Check_Pers_Data (int from_pers_file_load);

    void          Associate_Branch_Number (void);
#endif
public:
};

//CMsgCreditAdm
class CMsgCreditAdm : public CMessage
{
public:
    CMsgCreditAdm(char *msg_data = 0x0000);
    ~CMsgCreditAdm() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CGTCreditAdmCodeKey *key_ptr);    
public:
};

class CGTCcyPair: public CMessage
{
public:
    CGTCcyPair(char *msg_data = 0x0000);
    ~CGTCcyPair() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CCcyPairKey *key_ptr);
public:
};

class CGTIndicativesMark: public CMessage
{
public:
    CGTIndicativesMark(char *msg_data = 0x0000);
    ~CGTIndicativesMark() {}
    USING_CMESSAGE_INIT_KEY
    virtual void Init_Key (CMarketsKey *key_ptr);
public:
};

class CGTConnDmServer: public CMessage
{
public:
    CGTConnDmServer(char *msg_data = 0x0000);
    ~CGTConnDmServer() {}
    virtual void Init_Key (CConnDmServerKey *key_ptr);
public:
};

class CGTHoliday: public CMessage
{
public:
    CGTHoliday(char *msg_data = 0x0000);
    ~CGTHoliday() {}
    virtual void Init_Key (CHolidayKey *key_ptr);
public:
};

class CGTRFQRequest: public CMessage
{
public:
    CGTRFQRequest(char *msg_data = 0x0000);
    ~CGTRFQRequest() {}
    virtual void Init_Key (CRFQRequestKey *key_ptr);
public:
};

class CMsgRFQRequest: public CMessage
{
public:
    CMsgRFQRequest(char *msg_data = 0x0000);
    ~CMsgRFQRequest() {}
public:
};

__int64 Message_Get_Key (const char *);
#endif // DAPI

#endif //__MESSAGE_H__
