///////////////////////////////////////////////////////////////////////////////
// table.h
//
// CTable
///////////////////////////////////////////////////////////////////////////////

#ifndef __TABLE_H__
#define __TABLE_H__

#ifndef USE_SOLARIS 
#  ifndef USE_LINUX
#   pragma warning(disable: 4786)
#   include <crtdbg.h>
#  endif
#endif

extern char (*glb_ds_dest)[4];
#include "common.h"
#include "c_calc.h"
#include "arrm.h"
#include "dscom.h"
#include "restr.h"
#include "tblequiv.h"
#include "persis.h"
#ifndef DAPI
#include "log.h"
#endif

#ifdef DMEAPP
// required to control transactional processes
extern int Glb_TRANS_Activated;
extern void TRANS_Check (CMessage *, char);
extern unsigned int TRANS_services_counters[];
#endif

// This is to use the STLPort implementation of STL
// using namespace _STL;
// namespace _STL is used by STLPort instead of the default std.
// STLPort is used because it has the hash_map class that is not STL standard.
// #include <hash_map>
#include <map>
#include <list>
#include <set>
using namespace std;

#define Map_Key_Table  map <KEY_TYPE,  int>

#define FIRST_KEY    1
#define SECOND_KEY   2
#define THIRD_KEY    3

#define TITLE_TERM_DATE_SORT_KEY 1
#define TERM_TITLE_DATE_SORT_KEY 2
#define TERM_DATE_TITLE_SORT_KEY 3

#define NOT_PUT_ENV_AGAIN        0

class CTableBase;
class CMessage;
class CServiceKey;
class CPosKey;
class CMarket;
class CGTBridge;
class CGTCounterBridge;
class CGTIndicatives;
class CGTCobranzas;
class CGTCompartCupos;
class CGTPendCobranzas;
class CCobranzasKey;
class CLimInstitKey;
class CBuyAverPriceKey;
class CConfirmPendKey;
class CConfirmPendCodKey;
class CConfirmPendMarketKey;
class CConfCuposKey;
class CCompartCuposKey;
class CCompartCuposMasterKey;
class CFactRiesgoKey;
class CPendCobranzasKey;
class CGenDataKey;

typedef  map  <CServiceKey, CPosKey>      Map_Service_Table;
typedef  map  <CPosKey    , CServiceKey>  Map_Pos_Table;
typedef  list <DLONG>                     Env_Crc32_List;
typedef  list <unsigned int>             Idx_List;

extern int            Glb_table_counter;
extern unsigned int  Glb_range_environment;

#ifdef DAPI
#define INITIAL_TABLE_RECORDS   6
#else
#define INITIAL_TABLE_RECORDS   200
#endif

#define mem_add(x,y,z)            \
    memcpy(x, y, z);              \
    x += z;          

#if DGIAPP || GDIAPP

// Used in DGI persistent fields handling to
// have a similar define than mem_add but doing the opposite
// operation. It means the source and target are changed in 
// comparison with mem_add macro.
#define mem_add_source(x,y,z)            \
    memcpy(y, x, z);              \
    x += z;          

#endif

#define TABLE_MAX_TRANSACTIONAL_ELEMENT_LEN     64    // enough initially.

#define COUPON_LEN              18
#define ISIN_CODE_LEN           12
#define TITLE_CODE_LEN          3  
#define ISSUER_CODE_LEN         5
#define REFERENCE_CODE_LEN      10
#define LEVEL_CODE_LEN          2    
#define DUEDATE_CODE_LEN        8
#define CURRENCY_CODE_LEN       3


// Defines for Key classes.
#define ORD_MARKET_LEN        3
#define ORD_PAIR_CODE_LEN     7
#define ORD_O_D_LEN           1
#define ORD_SMARKET_LEN       1
#define ORD_GRPCODE_LEN       4
#define ORD_TITLE_REFERENCE_LEN 12
#define ORD_ARRANGE_LEN       2
#define ORD_VALUEDATE_LEN     2
#define ORD_NEMOTEC_LEN      19
#define ORD_NEMOTEC3C_LEN     3  
#define ORD_FLAG_LEN          1  

#define ORD_PGNUM_LEN          2    
#define ORD_INSNUM_LEN        4

#define ORD_TEXT_TERM_LEN     10
#define ORD_POSTURE_LEN       DFN_ORDER_CODE_LEN
#define DATATEC_TRADE_KEY_LEN DFN_TRADE_CODE_LEN
#define GTN_TRADE_KEY_LEN     20

#define ORD_LEVEL_LEN         1
#define ORD_MATUREDAY_LEN     8
#define ORD_HISTORY_DATE_LEN  8
#define ORD_XDAYS_LEN         2
#define ORD_DAYS_LEN          2
#define ORD_CATEGORY_LEN      1
#define ORD_PRICE_LEN         19
#define ORD_NUMERO_RONDA_LEN  6
// BANDERA_FISICO_COMPENSACION
#define ORD_TYPE_LEN          1
#define ORDERING_REST_LEN     1

#define UNIQUE_CODE_LEN       6
#define PU_CODE_LEN           DFN_PU_CODE_LEN
#define USR_CODE_LEN          DFN_USER_CODE_LEN
#define MARKET_USR_CODE_LEN   ORD_MARKET_LEN + USR_CODE_LEN
#define BRCH_BV1_LEN          5
#define USER_BV1_LEN          5
#define INSTIT_CODE_LEN       4
#define COD_BRANCH_REPLY      8 
#define COD_BRANCH_REQUEST    8

#define CITY_AGENCY_LEN       4
#define BRCH_CODE_LEN         (INSTIT_CODE_LEN + CITY_AGENCY_LEN)
#define BRCHRFQ_CODE_LEN      (ORD_MARKET_LEN  + COD_BRANCH_REPLY + COD_BRANCH_REQUEST)

#ifdef DAPI
#define COUNTRY_CODE          2
#define EMISOR_CODE           3
#define EMISORES_CODE_LEN     (COUNTRY_CODE + EMISOR_CODE)
#define COMITENTE_CODE        15
#define TYPE_COMITENTE_CODE   1
#define COMITENTES_CODE_LEN   (TYPE_COMITENTE_CODE + COMITENTE_CODE)
#define CLASE_CODIGO_SUBCOMITENTE   TYPE_COMITENTE_CODE
#define CODIGO_SUBCOMITENTE         COMITENTE_CODE
#define SUBCOMITENTE_CODE_LEN      (CLASE_CODIGO_SUBCOMITENTE + CODIGO_SUBCOMITENTE + CLASE_CODIGO_SUBCOMITENTE + CODIGO_SUBCOMITENTE)
#define SUBCOMITENTE_SUB_CODE_LEN  (CLASE_CODIGO_SUBCOMITENTE + CODIGO_SUBCOMITENTE)
#define FIDUC_CODE_LEN        3
#define FIDUCIARIOS_CODE_LEN  (INSTIT_CODE_LEN + FIDUC_CODE_LEN)
#define EMPLOYEE_CODE_LEN     4
// mercado esta definido con 4 caracteres.
#define AVGPURCHASEPRICE_CODE_LEN  (INSTIT_CODE_LEN + (DFN_MARKET_LEN+1) + DFN_DATE_LEN)
#define ID_ELEMENT_LEN        2
#define ESQUEMA_CODE_LEN      8
#define ESQUEMA_KEY_LEN      (ESQUEMA_CODE_LEN + ID_ELEMENT_LEN)

#endif

#define LOW_SEQ_CHAR          0x00
#define HIGH_SEQ_CHAR         0xFF

#define ORD_KEY_LEN          ( ORD_MARKET_LEN + ORD_O_D_LEN + ORD_SMARKET_LEN + ORD_TEXT_TERM_LEN + ORD_POSTURE_LEN )
#define NMCK_KEY_LEN         ( ORD_MARKET_LEN + BCLEN)

#define SUPERGROUP_KEY_LEN    4
#define VOICETERM_CODE_LEN    3
#define VOICETERM_KEY_LEN     (ORD_MARKET_LEN + VOICETERM_CODE_LEN)
#define VOICESUMMARY_KEY_LEN  (ORD_MARKET_LEN + ORD_GRPCODE_LEN + ORD_TITLE_REFERENCE_LEN + ORD_TEXT_TERM_LEN + DFN_CURRENCY_LEN + ORD_ARRANGE_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN + ORD_VALUEDATE_LEN + DFN_1C_LEN + DFN_2C_LEN + DFN_1C_LEN)

#define VOICEGROUP_KEY_LEN           4
#define FIXEDINCOMEMARKETS_KEY_LEN   4
#define PAIS_DM_LEN                  1

#define BRANCHES_PAIR_KEY_LEN ( ORD_MARKET_LEN + BRCH_CODE_LEN + BRCH_CODE_LEN)
#define PRICEORD_GEN_KEY_LEN  ( ORD_MARKET_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN   + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN+ORD_SMARKET_LEN+ORD_O_D_LEN)
#define PRICEORD_UNIC_KEY_LEN ( ORD_PRICE_LEN  + ORD_TYPE_LEN  + ORDERING_REST_LEN + UNIQUE_CODE_LEN + ORD_POSTURE_LEN)
#ifdef DGIAPP
#define INSTRUMENT_KEY_LEN    ( ETCPAGE_KEY_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN   + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN)
#endif
#ifdef GDIAPP
#define INSTRUMENT_KEY_LEN    ( ETCPAGE_KEY_LEN + TITLE_CODE_LEN  + ISSUER_CODE_LEN         + REFERENCE_CODE_LEN + CURRENCY_CODE_LEN + DUEDATE_CODE_LEN)
#define MX_INSTRUMENT_KEY_LEN ( ETCPAGE_KEY_LEN + ORD_GRPCODE_LEN + ORD_TITLE_REFERENCE_LEN + ORD_TEXT_TERM_LEN  + DFN_CURRENCY_LEN  + ORD_ARRANGE_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN + ORD_VALUEDATE_LEN + DFN_1C_LEN + DFN_2C_LEN + DFN_1C_LEN)
#define HISTORYFI_KEY_LEN     ( ORD_MARKET_LEN  + ORD_DAYS_LEN    + DFN_HOUR_LEN            + CODIGO_POSTURA_LEN + CODIGO_POSTURA_LEN)   
#endif
#define INSTRUMENT_FEED_KEY_LEN (ORD_MARKET_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN   + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN + ORD_O_D_LEN + USR_CODE_LEN)
#define TERM_KEY_LEN          ( ORD_MARKET_LEN + ORD_TEXT_TERM_LEN )
#define TERM_KEY_ALT_LEN      ( ORD_MARKET_LEN + ORD_TEXT_TERM_LEN + ORD_TEXT_TERM_LEN )
#define TERM_MATURITY_KEY_LEN ( ORD_MARKET_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + DFN_USHORT_LEN + ORD_TEXT_TERM_LEN )
#define SUMMARY_KEY_LEN       ( ORD_MARKET_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN )
#define STATISTICS_KEY_LEN    ( ORD_MARKET_LEN + ORD_HISTORY_DATE_LEN + ORD_SMARKET_LEN + ORD_TEXT_TERM_LEN + ORD_DAYS_LEN)
#define BRANCHES_BV1_KEY_LEN  ( BRCH_BV1_LEN   + BRCH_CODE_LEN )
#define BRANCHES_RFQ_KEY_LEN  ( DFN_1C_LEN     + BRCH_BV1_LEN  + BRCH_CODE_LEN )
#define SPOT_KEY_LEN          ( ORD_PAIR_CODE_LEN)//ORD_MARKET_LEN)
#define CURRENCY_KEY_LEN      ( PAIS_DM_LEN + DFN_CURRENCY_LEN )
#define CASHRATE_KEY_LEN      ( ORD_MARKET_LEN + DFN_CURRENCY_LEN + ORD_TEXT_TERM_LEN)
#define USER_BV1_KEY_LEN      ( USER_BV1_LEN + USR_CODE_LEN)
#define ORD_USER_KEY_LEN      ( ORD_KEY_LEN + PU_CODE_LEN + USR_CODE_LEN)
#define SUMMARYRF_KEY_LEN     ( ORD_MARKET_LEN + ORD_PGNUM_LEN + ORD_INSNUM_LEN)
#define SUMMARYFI_KEY_LEN     ( DFN_MARKET_LEN + TITLE_CODE_LEN + ISSUER_CODE_LEN + REFERENCE_CODE_LEN +  CURRENCY_CODE_LEN + DUEDATE_CODE_LEN )  
#define SUMMARYCO_KEY_LEN     ( ETCPAGE_KEY_LEN + ORD_NEMOTEC_LEN + ORD_VALUEDATE_LEN + ORD_FLAG_LEN + DUEDATE_CODE_LEN + ORD_FLAG_LEN)  



                                                  // LENGTH OF CODIGO_PRIMARYUSER + LENGTH OF USUARIO   (SEE HEADER)

#define ORD_LOGIN_NAME_KEY_LEN ( ORD_MARKET_LEN + DFN_ETC_LOGIN_NAME_LEN + ORD_O_D_LEN + ORD_POSTURE_LEN)


#define ORD_ALL_CHAR         '?'

#define CALC_MNUM_LEN         CALC_MXDIGITS*sizeof(CALC_MNUM)
#define SERVICE_TYPE_LEN      1
#define LONG_VAR_SIZE         4
#define SERVICE_KEY_LEN       (SERVICE_TYPE_LEN + ORD_MARKET_LEN + CALC_MNUM_LEN + LONG_VAR_SIZE)
#define POS_KEY_LEN           (LONG_VAR_SIZE + SERVICE_TYPE_LEN)

// Services defines. Order protection cases.
// 0-15  long element
// 16-31 CALC_MNUM * element
#define SERVICE_INVALID_RECORD                  -1
#define SERVICE_NO_SERVICE                      0
#define SERVICE_LONG_TIMEOUT_PROTECTION         1   // used for Good Till when BANDERA_POSTURA_CON_EXPIRACION is set
#define SERVICE_LONG_FAT_FINISHING_AUT_TIME     2   // used for finishing automatic time in voice markets
#define SERVICE_LONG_TRADING_RIGHTS             3   // used for TRADING RIGHTS, DERECHO DE TANTO, fixed income, voice market, mexico
#define SERVICE_LONG_FIXING_PRICE_STATUS        4   // used for FIXING PRICE status in G7
#define SERVICE_LONG_8                          8   // not used.
#define SERVICE_LONG_UPPER_LIMIT                15

#define SERVICE_CALC_SPOT_LINK_PROTECTION       16 

#define KEY_BID_CHAR 'D'
int     TABLE_KEY_Price_Comp (const char *key_ptr, 
                              const char *x_key_ptr, 
                              int         bid_offer, 
                              int         price_key_sorting_method);
void    TABLE_KEY_Price_Key_Fill (char       *cur_cpr_ptr, 
                                  const char *o_d, 
                                  int         fill_value_char,
                                  int         price_key_sorting_method);

#ifndef DAPI
#define TABLE_SENT_RECORD_LOG_TEXT              "IT-Table record sent."
#else
#define TABLE_SENT_RECORD_LOG_TEXT              NULL
#endif

#define MAX_KEYS   3

#ifdef  GLOBAL_HERE
#define GLOBAL
#else
#define GLOBAL extern
#endif

GLOBAL int                  Glb_range_limit_err_msg;
GLOBAL unsigned int         Glb_range_limit;
GLOBAL char                 Glb_Local_Machine[4];
GLOBAL unsigned int         glb_min_gtable_rec_index;
GLOBAL unsigned int         glb_max_gtable_rec_index;
GLOBAL bool                 Glb_ds_range_received;
GLOBAL int                  glb_dss_counter;
GLOBAL unsigned int         Glb_Unic_Number;

#ifdef DAPI
GLOBAL unsigned int        glb_min_pu_connected_gtable_rec_index;
#endif

void    LOG_Put_Err_Msg    (const char *text_format, ...);

// key used to hold triggering data for services over table records
// services available are:
// goodfor or Good Till,  Useful to withdraw records when time comes
// protection,            useful to withdraw orders when the price is out of range
// 
class CServiceKey
{
public:
  char m_key[SERVICE_KEY_LEN];
  
  CServiceKey()
  {
      memset(m_key, 0, SERVICE_KEY_LEN);
  }
  CServiceKey *Init(char  service_type,
                    char *market,          // used only for price range protection
                    void *service,         // data to compare to  (price, lower or opper), hour, or anything simmilar
                    void *obj_ptr);        // pointer to the related record position 
  CServiceKey *Init(char  service_type,
                    char *market,
                    char  fill_char);
  int          First_Part_Comp(const CServiceKey &X) const;
  bool operator <( const CServiceKey &X) const;
};


class CServiceKeyIterator
{
public:
  CServiceKey           m_map_first_key;
  CServiceKey           m_map_last_key;
  CServiceKey           m_map_prev_key; // used to ensure not infinite loop, when inside errors occur
};

class CPosKey
{
public:
  DWORD64 m_key_position;
  char    m_key_service_type;

  CPosKey()
  {
      m_key_position = 0;
      m_key_service_type = 0;
  }
  CPosKey *Init(DWORD64 obj_ptr,
                char service_type);
  bool operator <(const CPosKey &X) const;
};

#define CSK_DELETED   0
#define CSK_INSERTED  1

#include "vector.h"
// Base key class.
class CBaseKey
{
public:
  char *m_key;
  int   m_key_len;

public:
         CBaseKey (const CBaseKey &X);
         CBaseKey (void);
         CBaseKey (int len);
virtual ~CBaseKey ();

public:
  void      Init          ( const char *new_key_value);
  virtual int       Fit_Pattern   ( CBaseKey *key_pattern);
  virtual char     *Get_Str       ( void) const;
  virtual const char *Get_Key_Ptr ( void) const;
  virtual void        Set_Price_Key_Sorting_Method(int price_key_sorting_method) {} //For G7 and voice keys

  void operator  =(CBaseKey &X);
  bool operator  <(const CBaseKey &X) const;
  bool operator ==(const CBaseKey &X) const;
  bool operator !=(const CBaseKey &X) const;
};

class CNamecheckKey : public CBaseKey
{
public:
  CNamecheckKey();
  CNamecheckKey *Init (const char *market, const char *branch_code);
  const char *Get_Market() {return m_key;}
  const char *Get_Branch_Code() {return (m_key + ORD_MARKET_LEN);}
};

// this is the class for the index of a record
// this new idx replaces a previoux idx value for a record with only the unsigned int rec_idx
class CIdxKey
{
public:
  unsigned short data_code;
  unsigned int  rec_idx;

  CIdxKey();
  CIdxKey *Init (unsigned short data_code_v, unsigned int rec_idx_v);
  char    *Get_Str (void) const;
  bool operator <   (const CIdxKey &X) const;
  bool operator ==  (const CIdxKey &X) const;
};

typedef map< CIdxKey,
             int//,
             //hash<int>,
             //equal_to<unsigned int>
             > Map_Idx_Table;

#define ONE_CHAR_FLAG_LEN   1
#define C_V_FLAG_LEN     1
#define CITY_3C_LEN      3
#define MONEY_CODE_LEN   3
#define INSPERM_CODE_LEN 5
#define AGENCY_LEN       1
#define RANGE_FLAG       2
#define RISK_F_TERM_LEN  ( ORD_MARKET_LEN + BCLEN + BCLEN + C_V_FLAG_LEN + DFN_TEXT_TERM_LEN + DFN_TEXT_TERM_LEN)
#define PLCCREDIT_KEY_LEN       0  //TBD
#define CUPOG_KEY_LEN           ( ORD_MARKET_LEN + BRCH_CODE_LEN + BRCH_CODE_LEN + C_V_FLAG_LEN )
#define CUPO_BRCH_KEY_LEN       ( ORD_MARKET_LEN + BRCH_CODE_LEN + INSTIT_CODE_LEN + C_V_FLAG_LEN )
#define TRANSACTION_DATE_LEN    8
#define TRANSACTION_HOUR_LEN    6
#define CUPOL_KEY_LEN           ( ORD_MARKET_LEN + INSTIT_CODE_LEN + INSTIT_CODE_LEN + C_V_FLAG_LEN + CITY_AGENCY_LEN)
#define CUPOXDIA_KEY_LEN        ( ORD_MARKET_LEN + INSTIT_CODE_LEN + INSTIT_CODE_LEN + C_V_FLAG_LEN + CITY_AGENCY_LEN)
#define INF_SUP_RATE_LEN        5
#define CUPOXPL_KEY_LEN         ( ORD_MARKET_LEN + INSTIT_CODE_LEN + INSTIT_CODE_LEN + C_V_FLAG_LEN + CITY_AGENCY_LEN + INF_SUP_RATE_LEN + INF_SUP_RATE_LEN)
#define COBRANZAS_KEY_LEN       ( INSTIT_CODE_LEN + INSTIT_CODE_LEN)
#define CODIGO_POSTURA_LEN      10
#define PENDCOBRANZAS_KEY_LEN   ( ORD_MARKET_LEN + INSTIT_CODE_LEN +INSTIT_CODE_LEN + C_V_FLAG_LEN + CODIGO_POSTURA_LEN + CODIGO_POSTURA_LEN + DFN_DATE_LEN + DFN_HOUR_LEN + INF_SUP_RATE_LEN + INF_SUP_RATE_LEN)
#define FACTRIESGO_KEY_LEN      ( ORD_MARKET_LEN + DFN_BRANCH_CODE_LEN + INSTIT_CODE_LEN + ONE_CHAR_FLAG_LEN + RANGE_FLAG + RANGE_FLAG)  
#define CONFCUPOS_KEY_LEN       ( ORD_MARKET_LEN + INSTIT_CODE_LEN )
#define COMPARTCUPOS_KEY_LEN    ( ORD_MARKET_LEN + INSTIT_CODE_LEN )
#define COMPARTCUPOS_MAST_KEY_LEN ( ORD_MARKET_LEN + INSTIT_CODE_LEN + ORD_MARKET_LEN)
#define SUCURS_CODE_LEN     8
#define LIMINSTIT_KEY_LEN       ( ORD_MARKET_LEN + INSTIT_CODE_LEN + SUCURS_CODE_LEN)
#define BUYAVERPRICE_KEY_LEN    ( ORD_MARKET_LEN + INSTIT_CODE_LEN + TRANSACTION_DATE_LEN)
#define CONFIRMPEND_KEY_LEN     ( CODIGO_POSTURA_LEN + CODIGO_POSTURA_LEN )
#define CONFIRMPEND_MARKET_KEY_LEN   ( ORD_MARKET_LEN + CODIGO_POSTURA_LEN + CODIGO_POSTURA_LEN )
#define CONFIGCREDIT_KEY_LEN    ( ORD_MARKET_LEN + INSTIT_CODE_LEN + ONE_CHAR_FLAG_LEN)
#define POLICYCREDIT_KEY_LEN    ( ORD_MARKET_LEN + BRCH_CODE_LEN + ONE_CHAR_FLAG_LEN + ONE_CHAR_FLAG_LEN)
#define NUM_BRANCH_CODE_KEY_LEN 8
#define NUM_BRANCH_NUM_KEY_LEN  2
#define INSCRIBE_DME_PU_CODE_KEY_LEN      4
#define INSCRIBE_DME_BRANCH_CODE_KEY_LEN  (ORD_MARKET_LEN + BRCH_CODE_LEN + USR_CODE_LEN)
#define CREDITADM_KEY_LEN                 (ORD_MARKET_LEN + INSTIT_CODE_LEN)
#define CCYPAIR_KEY_LEN                   (DFN_FLAG_LEN + DFN_CURRENCY_LEN + DFN_CURRENCY_LEN)
#define CONNDMSERVER_KEY_LEN              (DFN_PU_CODE_LEN)
#define CCYPAIR_TERM_KEY_LEN              (DFN_CURRENCY_LEN + DFN_CURRENCY_LEN + DFN_TEXT_TERM_LEN + DFN_DATE_LEN + DFN_DATE_LEN)
#define HOLIDAY_KEY_LEN                   (PAIS_DM_LEN + DFN_COUNTRY_CODE_LEN + DFN_DATE_LEN)
#define RFQREQUEST_KEY_LEN                (ORD_MARKET_LEN + DFN_DATE_LEN + USR_CODE_LEN + UNIQUE_CODE_LEN)
#define RFQREQDATATEC_KEY_LEN             (ORD_MARKET_LEN + USR_CODE_LEN + UNIQUE_CODE_LEN)
#define RFQREQETC_KEY_LEN                 (DFN_4C_LEN + DFN_4C_LEN + UNIQUE_CODE_LEN)
#define RFQENTRYDATATEC_KEY_LEN           (RFQREQDATATEC_KEY_LEN + DFN_BRANCH_CODE_LEN)
#define GENDATA_KEY_LEN                   12
#define FIXPRICE_KEY_LEN                  (ORD_MARKET_LEN+DFN_LEVEL_LEN+DFN_DATE_LEN+DFN_XDIAS_LEN+DFN_DIAS_LEN+DFN_TEXTO_LEN)

#ifdef GDIAPP
#define ETCPAGE_KEY_LEN                   (ORD_MARKET_LEN + sizeof (unsigned int))
#else
#define ETCPAGE_KEY_LEN                   (ORD_MARKET_LEN + DFN_CURRENCY_LEN + DFN_CURRENCY_LEN)
#endif

class CNmChkPlzKey : public CBaseKey
{
public:
  CNmChkPlzKey();
  char *Init (const char *market, 
              const char *my_branch_code, 
              const char *othr_branch_code, 
              const char *c_v_flag,
              const char *initial_text_term,
              const char *final_text_term);
};

// this is the class for the index of a record
// this new idx replaces a previoux idx value for a record with only the unsigned int rec_idx
class CEnvCrc32Key : public CBaseKey
{
public:
  CEnvCrc32Key();
  CEnvCrc32Key *Init (DLONG val_env_crc32);
};

class CPlcCreditKey : public CBaseKey
{
public:
  CPlcCreditKey();
  //To be defined G.CH
  CPlcCreditKey *Init ();
};

class CInstNumKey : public CBaseKey
{
public:
  CInstNumKey();
};

class CPInsDME_BCKey : public CBaseKey
{
public:
  CPInsDME_BCKey();
  CPInsDME_BCKey *Init ( const char *market,
                         const char *branch_code,
                         const char *pu_code);

  CPInsDME_BCKey *Init ( const char *market,
                         const char *inst_code,
                               int  fill_char);

};

class CGTCreditAdmCodeKey : public CBaseKey
{
public:
  CGTCreditAdmCodeKey *Init  (const char *market, 
                              const char *instit_code);

  CGTCreditAdmCodeKey();
};

class CCcyPairKey : public CBaseKey
{
public:
  CCcyPairKey();
  CCcyPairKey *Init  (const char *bandera_ndf_forward, 
                      const char *codigo_moneda,
                      const char *codigo_moneda_contraparte);

};

class CCupoGKey : public CBaseKey
{
public:
  CCupoGKey *Init  (const char *market, 
                   const char *my_branch,   // <cod_instit><city><agency>
                   const char *other_branch,// if credit configuration by Instit, city and agency has '-' character.
                   const char *c_v);

  CCupoGKey *Init  (const char *market, 
                   const char *my_branch,    // <cod_instit><city><agency>
                   const char *other_instit, // if credit configuration by Instit, city and agency has '-' character.
                   int  fill_char);

  CCupoGKey *Init  (const char *market, 
                   const char *my_instit,    // <inst_code>
                   int  fill_char);
  CCupoGKey();
};

class CCupoLKey : public CBaseKey
{
public:
  CCupoLKey *Init (const char *market, 
                   const char *my_instit,        // <cod_instit>
                   const char *other_instit,     // <cod_instit>
                   const char *c_v,
                   const char *city_subsidiary); // <city><subsidiary>

  CCupoLKey *Init (const char *market, 
                   const char *my_branch,    // <cod_instit><city><agency>
                   const char *other_instit, // <cod_instit>
                   const char *c_v);
  
  CCupoLKey *Init (const char *market, 
                   const char *my_instit,
                   int  fill_char); 
  CCupoLKey();
};

class CCupoXPLKey : public CBaseKey
{
public:

  CCupoXPLKey *Init (const char *market, 
                     const char *my_instit,         // <cod_instit>
                     const char *other_instit,      // <cod_instit>
                     const char *c_v,
                     const char *city_subsidiary,   // <city><subsidiary>))
                     const char *inf_rate,
                     const char *sup_rate);
                     
  CCupoXPLKey *Init (const char *market, 
                     const char *my_branch,         // <cod_instit><city><agency>
                     const char *other_instit,      // <cod_instit>
                     const char *c_v,
                     const char *inf_range,   
                     const char *sup_range);
  
  CCupoXPLKey *Init (const char *market, 
                     const char *my_branch,         // <cod_instit><city><agency>
                     const char *other_instit,      // <cod_instit>
                     const char *c_v);
  
  CCupoXPLKey *Init (const char *market, 
                     const char *my_branch,         //<cod_instit><city><agency>
                     const char *other_instit,      // <cod_instit>
                     const char *c_v,
                     int         fill_char); 

  CCupoXPLKey *Init (const char *market, 
                     const char *my_instit,
                     int  fill_char); 
  CCupoXPLKey();
};

class CCobranzasKey : public CBaseKey
{
public:
  CCobranzasKey *Init (const char *codigo_institucion_nuestra, 
                       const char *codigo_institucion_cobranza);
  CCobranzasKey *Init (const char *market, int  fill_char);
  CCobranzasKey();
};

class CPendCobranzasKey : public CBaseKey
{
public:
  CPendCobranzasKey *Init (const char *market,
                           const char *mine_cod_instit,
                           const char *other_cod_instit,
                           const char *c_v,
                           const char *cod_postura,
                           const char *cod_postura_1,
                           const char *transaction_date,
                           const char *transaction_hour);

  CPendCobranzasKey *Init (const char *market,
                           const char *mine_cod_instit,
                           const char *other_cod_instit,
                           const char *c_v,
                           int  fill_char);
  CPendCobranzasKey();
};

class CFactRiesgoKey : public CBaseKey
{
public:
  CFactRiesgoKey *Init (const char *market, 
                        const char *mine_code_branch,
                        const char *other_code_instit, 
                        const char *c_v_flag,
                        const char *range_inf,
                        const char *range_sup);
  CFactRiesgoKey();
};

class CCompartCuposKey : public CBaseKey
{
public:
  CCompartCuposKey *Init (const char *market,
                          const char *mine_code_instit);
  CCompartCuposKey();
};

class CCompartCuposMasterKey : public CBaseKey
{
public:
  CCompartCuposMasterKey *Init (const char *master_market,
                                const char *mine_code_instit,
                                const char *market);

  CCompartCuposMasterKey();
};

class CConfCuposKey : public CBaseKey
{
public:
  //CConfCuposKey *Init (const char *code);
  CConfCuposKey *Init ( const char *market, 
                        const char *cod_instit);
  CConfCuposKey();
};

class CGenDataKey : public CBaseKey
{
public:
  CGenDataKey();
  CGenDataKey *Init (const char *id_code); 
};

class CLimInstitKey : public CBaseKey
{
public:
  CLimInstitKey *Init (const char *mercado, 
                       const char *cod_garantizadora,  
                       const char *sucursal_limit);
  CLimInstitKey();
};

class CBuyAverPriceKey : public CBaseKey
{
public:
  CBuyAverPriceKey *Init (const char *mercado, 
                          const char *cod_instit_sucur,  
                          const char *fecha_registro);
  CBuyAverPriceKey();
};

class CConfirmPendKey : public CBaseKey
{
public:
  unsigned int end_curtime;

  CConfirmPendKey();
  CConfirmPendKey *Init (unsigned int  end_cur_time_v,
                         const char    *code_order,
                         const char    *code_order_1);
  CConfirmPendKey *Init (unsigned int  end_cur_time_v,
                         int            fill_char);

  bool operator <   (const CConfirmPendKey &X) const;
};

class CConfirmPendCodKey : public CBaseKey
{
public:

  CConfirmPendCodKey();
  CConfirmPendCodKey *Init (const char  *code_order,
                            const char  *code_order_1);

};

class CConfirmPendMarketKey : public CBaseKey
{
public:

  CConfirmPendMarketKey();
  CConfirmPendMarketKey *Init (const char  *market,   
                               const char  *cod_post,
                               const char  *cod_post_1);
};

class CFixPriceKey : public CBaseKey
{
public:

  CFixPriceKey();
  CFixPriceKey *Init (const char  *market,   
                      const char  *level,
                      const char  *date_maturity,
                      const char  *xdays,
                      const char  *days,
                      const char  *text);
};

class CConnDmServerKey : public CBaseKey
{
public:
  CConnDmServerKey();
};

class CCcyPairTermKey : public CBaseKey
{

public:
  CCcyPairTermKey();
  CCcyPairTermKey *Init (const char *currency1,
                         const char *currency2,
                         const char *text_term,
                         const char *start_date,
                         const char *end_date);
};

class CHolidayKey : public CBaseKey
{
public:
  CHolidayKey();
  CHolidayKey *Init  (const char *country_dm_code,
                      const char *code); // codigo_pais + fecha_feriado
};

class CRFQRequestKey : public CBaseKey
{
public:
  CRFQRequestKey();
  CRFQRequestKey   *Init  (const char *mercado, 
                           const char *fecha_ronda,
                           const char *usuario_ronda,
                           const char *numero_ronda);
};

///////////////////////////////////////////////////////////////////////////////
// CRFQReqDatatecKey, CRFQReqETCKey, CRFQEntryDatatecKey
// Keys used in RFQ process.
///////////////////////////////////////////////////////////////////////////////
class CRFQReqDatatecKey : public CBaseKey
{
public:
    CRFQReqDatatecKey();

    CRFQReqDatatecKey   *Init   (const char *market, 
                                 const char *user_code, 
                                 const char *rfq_code);

    const char          *Get_User_RFQ_Code() {return (m_key + ORD_MARKET_LEN);}
};

class CRFQReqETCKey : public CBaseKey
{
public:
    CRFQReqETCKey();

    CRFQReqETCKey       *Init   (const char *firmid, 
                                 const char *instrument,
                                 const char *rfq_code);
};

class CRFQEntryDatatecKey : public CBaseKey
{
public:
    CRFQEntryDatatecKey();

    CRFQEntryDatatecKey *Init   (const char *rfq_entry_datatec_key);
    CRFQEntryDatatecKey *Init   (const char *market, 
                                 const char *user_code, 
                                 const char *rfq_code,
                                 const char *branch_code);

    CRFQEntryDatatecKey *Init   (const char *rfqreq_datatec_key, 
                                 const char *branch_code);

    const char          *Get_Market()           {return (m_key);}
    const char          *Get_User_RFQ_Code()    {return (m_key + ORD_MARKET_LEN);}
    const char          *Get_RFQ_Request_Code() {return (m_key + ORD_MARKET_LEN + USR_CODE_LEN);}
};

class CConfigCreditKey: public CBaseKey
{
public:
  CConfigCreditKey();
  virtual void operator  =(CConfigCreditKey &X);
  CConfigCreditKey *Init (const char *market,
                          const char *instit_code,
                          const char *mine_other_flag);
  void              Get  (char **market,
                          char **instit_code,
                          char **mine_other_flag);
};

class CPolicyCreditKey: public CBaseKey
{
public:
  CPolicyCreditKey();
  CPolicyCreditKey *Init (const char *market,
                          const char *branch_code,
                          const char *m_s_flag,
                          const char *c_v_flag);

  CPolicyCreditKey *Init (const char *market,
                          int         fill_char);


};

// BV1 Key for User Table
class CUserBv1Key : public CBaseKey
{

public:
  CUserBv1Key();
  CUserBv1Key *Init(const char *bv1_code, const char *user_code);
};

// Key for User Table
class CUserKey : public CBaseKey
{

public:
  CUserKey();  
};

class CBranchUserKey : public CBaseKey
{

public:
  CBranchUserKey();
  CBranchUserKey *Init(const char *branch_code, const char *user_code);
  CBranchUserKey *Init(const char *branch_code, int fill_value_char);
};

#ifdef DMEAPP
class CBridgeKey : public CBaseKey
{
public:
  CBridgeKey();
  CBridgeKey *Init( const char *market_code, 
                    const char *currency, 
                    const char *two_currency, 
                    const char *send_recv_flag,
                    const char *bandera_tipo,
                    const char *sequence);
  void        Init( const char *key_code);
};

class CCounterBridgeKey : public CBaseKey
{
public:
  CCounterBridgeKey();
  CCounterBridgeKey *Init( const char *market, 
                           const char *currency, 
                           const char *two_currency, 
                           const char *flag_info_type,
                           const char *delim,
                           const char *sequence);
};

class CBridgeSymKey : public CBaseKey
{
public:
  CBridgeSymKey();
  CBridgeSymKey *Init(const char *symbol_str, const char *producer_name, const char *clave_str);
};

class CIndicativesKey : public CBaseKey
{
public:
  CIndicativesKey();
  CIndicativesKey *Init(const char *mercado,
                        const char *texto,
                        const char *nivel, 
                        const char *fecha_vencimiento, 
                        const char *xdias_num, 
                        const char *dias_num);
  CIndicativesKey *Init(const char *mercado, int fill_char);
  CIndicativesKey *Init(const char *mercado, const char *texto, int fill_char);
};

class CPrTermKey : public CBaseKey
{
public:
  CPrTermKey();  
  CPrTermKey *Init (const char *productor, const char *mercado, const char *texto_term, const char *simbolo);
};

#endif

class CMarketsKey : public CBaseKey
{
public:
  CMarketsKey();  
};


class CSpotMarketsKey: public CBaseKey
{
public:
  CSpotMarketsKey();  
  CSpotMarketsKey *Init(const char *spot_key, const char *market_code);
  CSpotMarketsKey *Init(const char *spot_key, int         fill_char);
};


class CMarketUserKey : public CBaseKey
{
public:
  CMarketUserKey();
  CMarketUserKey * Init(const char *market, const char *user_code);
  CMarketUserKey * Init(const char *market, const CUserKey *user_key);
};

// Key for Spot Price Table
class CSpotKey : public CBaseKey
{

public:
  CSpotKey();  
  void      Init(const char *pait_code);
  CSpotKey *Init(const char *currrency1, const char *currency2);
};

// Key for Tables with key market and text_term.
class CCurrencyKey : public CBaseKey
{

public:
  CCurrencyKey();  
  CCurrencyKey *Init(const char *pais_dm, const char *codigo);
};

// Key for Tables with key market and text_term.
class CCashRateKey : public CBaseKey
{

public:
  CCashRateKey();  
  CCashRateKey *Init(const char *currency, const char *text_term);
};

#if DGIAPP || GDIAPP

// Key1 for Branch Table
class CBranchesBv1Key : public CBaseKey
{

public:
  CBranchesBv1Key();
  CBranchesBv1Key *Init(const char *bv1_code,
                        const char *branch_code);
};

// Key1 for Branch Table
class CBranchesRFQKey : public CBaseKey
{

public:
  CBranchesRFQKey();
  CBranchesRFQKey *Init(const char *rfq_flag,
                        const char *bv1_code,
                        const char *branch_code);
};

#endif

#ifdef DAPI
class CEmisoresKey : public CBaseKey
{

public:
  CEmisoresKey();  
  CEmisoresKey *Init(const char *code);
};

class CComitentesKey : public CBaseKey
{

public:
  CComitentesKey();  
  CComitentesKey *Init(const char *type_code,
                       const char *code);
};
          
class CSubComitenteKey : public CBaseKey
{

public:
  CSubComitenteKey();  
  CSubComitenteKey *Init(const char *class_code_father,
                         const char *code_father_subcom,
                         const char *class_code_subcom,
                         const char *code_subcom);
};

class CSubComitenteSubKey : public CBaseKey
{

public:
  CSubComitenteSubKey();  
  CSubComitenteSubKey *Init(const char *class_code,
                            const char *code_subcom);
};      

class CFiduciariosKey : public CBaseKey
{

public:
  CFiduciariosKey();  
  CFiduciariosKey *Init(const char *cod_instit_nuestra,
                        const char *cod_fiduciario);
};

class CEmployeeKey : public CBaseKey
{

public:
  CEmployeeKey();  
  CEmployeeKey *Init(const char *employee_code);
};

class CAvgPurchasePriceKey : public CBaseKey
{

public:
  CAvgPurchasePriceKey();  
  CAvgPurchasePriceKey *Init(const char *instit_branch,
                             const char *market,
                             const char *date);
};

class CEsquemaKey : public CBaseKey
{
public:
  CEsquemaKey();  
  CEsquemaKey  *Init(const char     *esquema_code, 
                     unsigned int   id_element);
};

#endif

// Key2 for Branch Table
class CBranchesRfqsKey : public CBaseKey
{

public:
  CBranchesRfqsKey();  
  void              Init(const char *branch_code);
  CBranchesRfqsKey *Init(const char *market, 
                         const char *cod_reply_branch,
                         const char *cod_request_branch);
};


// Key2 for Branch Table
class CBranchesKey : public CBaseKey
{

public:
  CBranchesKey();  
  void          Init(const char *branch_code);
  CBranchesKey *Init(const char *instit_code, int fill_value_char);
};

class CBranchesNumKey : public CBaseKey
{
  unsigned int branch_num;
public:
  CBranchesNumKey();  
  void          Init(unsigned int b_num);
  char         *Get_Str (void) const;
  bool operator < (const CBranchesNumKey &X) const;
  bool operator ==(const CBranchesNumKey &X) const;
};

// key used for credit policies restriction.
class CBranchesPairKey : public CBaseKey
{
public:
  CBranchesPairKey ();
  CBranchesPairKey *Init(const char *market,
                         const char *branch1,
                         const char *branch2);
  void              Get_Elements(const char **market,
                                 const char **branch1,
                                 const char **branch2);
};

// Class used for algorithms like find_if that needs a predicate.
// Used with credit policies restrictions multimap 
class IsEqualBranchesPairKey
{
public:
  const CBranchesPairKey *branchespair_key_;
  // constructor of function object
  IsEqualBranchesPairKey(const CBranchesPairKey *branchespair_key) : branchespair_key_(branchespair_key) {}
  long operator()(pair < long const, CBranchesPairKey > &X)
          {return (X.second == *branchespair_key_);}
  long operator()(pair < long const, CBranchesPairKey > *X)
          {return (X->second == *branchespair_key_);}
};

// Key for Summary Table
class CStatisticsKey : public CBaseKey
{

public:
  CStatisticsKey();
  CStatisticsKey *Init(const char *market,
                       const char *history_date,
                       const char *sub_mercado_statistics_ptr,
                       const char *texto_term_statistics_ptr,
                       const char *dias_aux_ptr);
};

// Key for Summary Table
class CSummaryKey : public CBaseKey
{

public:
  CSummaryKey();
  CSummaryKey *Init(const char *market,
                    const char *xdays, 
                    const char *days,
                    const char *texto_term,
                    const char *level,
                    const char *maturity_date);
  CSummaryKey *Init(const char *market,
                    const char *xdays, 
                    const char *days,
                    const char *texto_term,
                    int         fill_char);
  CSummaryKey *Init(const char *market, int fill_char);

  inline const char *Get_Market        ()  {return m_key;}
  inline const char *Get_Xdays         ()  {return (m_key + ORD_MARKET_LEN);}
  inline const char *Get_Days          ()  {return (m_key + ORD_MARKET_LEN + ORD_XDAYS_LEN);}
  inline const char *Get_Text_Term     ()  {return (m_key + ORD_MARKET_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN);}
  inline const char *Get_Level         ()  {return (m_key + ORD_MARKET_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN);}
  inline const char *Get_Maturity_Date ()  {return (m_key + ORD_MARKET_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN + ORD_TEXT_TERM_LEN + ORD_LEVEL_LEN);}
};

// Key for Term Table
class CTermKey : public CBaseKey
{

public:
  CTermKey();
  CTermKey *Init(const char *market,
                 const char *text_term);
};

// Key for Term Table
class CTermAltKey : public CBaseKey
{

public:
  CTermAltKey();
  CTermAltKey *Init(const char *market,
                    const char *text_term_alt,
                    const char *text_term);
};

// Key for Term Table that includes the maturity date.
class CTermMaturityKey : public CBaseKey
{

public:
  CTermMaturityKey();
  CTermMaturityKey *Init(const char *market,
                         const char *level,
                         const char *maturity_date,
                         const char *sequence,
                         const char *text_term);
};

#ifdef DGIAPP

class CETCPageKey : public CBaseKey
{
public:
  CETCPageKey();
  CETCPageKey *Init        (const char *datatec_market, 
                            const char *currency1,
                            const char *currency2);

  const char     *Get_Market()     {return (m_key);}
  const char     *Get_Currency1()  {return (m_key + ORD_MARKET_LEN);}
  const char     *Get_Currency2()  {return (m_key + ORD_MARKET_LEN + DFN_CURRENCY_LEN);}
};

#endif

#ifdef DGIAPP
// Key with the parameters that defines an instrument.
class CInstrumentKey : public CBaseKey
{

public:
  CInstrumentKey();

  CInstrumentKey *Init(const char *etc_page_key,
                       const char *level,
                       const char *maturity_date,
                       const char *xdays, 
                       const char *days,
                       const char *text_term);

  CInstrumentKey *Init(const char *etc_page_key,
                       int         fill_char);

  void            Build_ETCPageKey (CETCPageKey *etc_page_key) {etc_page_key->Init(Get_Market(),Get_Currency1(),Get_Currency2());}

  const char     *Get_Market()    {return (m_key);}

  const char     *Get_Currency1() {return (m_key + ORD_MARKET_LEN);}
  const char     *Get_Currency2() {return (m_key + ORD_MARKET_LEN + DFN_CURRENCY_LEN);}
  const char     *Get_Text_Term() {return (m_key + ORD_MARKET_LEN + DFN_CURRENCY_LEN + DFN_CURRENCY_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN);}
};

#endif

// Key with the parameters that defines an instrument per side and user.
// Used to get the order codes of feed orders.
class CInstrumentFeedKey : public CBaseKey
{

public:
  CInstrumentFeedKey();
  CInstrumentFeedKey *Init(const char *market,
                           const char *level,
                           const char *maturity_date,
                           const char *xdays, 
                           const char *days,
                           const char *text_term,
                           const char *o_d,
                           const char *user_code);
  CInstrumentFeedKey *Init(CSummaryKey *summary_key_ptr,
                           const char  *o_d,
                           const char  *user_code);
};

#define G7_TO_O_D_KEY_LEN       (ORD_MARKET_LEN+ORD_LEVEL_LEN+ORD_MATUREDAY_LEN+ORD_XDAYS_LEN+ORD_DAYS_LEN+ORD_TEXT_TERM_LEN+ORD_SMARKET_LEN)
#define G7_TO_PRICE_KEY_LEN     (G7_TO_O_D_KEY_LEN+ORD_O_D_LEN+ORD_CATEGORY_LEN)
#define PRICE_G7_MARKET_KEY_LEN (G7_TO_PRICE_KEY_LEN+ORD_PRICE_LEN+ORD_TYPE_LEN+DFN_DATE_LEN+UNIQUE_CODE_LEN+ORD_POSTURE_LEN)

#define VOICE_PART_KEY_LEN          (ORD_GRPCODE_LEN + ORD_TITLE_REFERENCE_LEN + DFN_CURRENCY_LEN + ORD_ARRANGE_LEN + ORD_VALUEDATE_LEN + DFN_1C_LEN + DFN_2C_LEN + DFN_1C_LEN)
#define VOICE_TO_O_D_KEY_LEN        (ORD_MARKET_LEN+ORD_XDAYS_LEN+ORD_DAYS_LEN+VOICE_PART_KEY_LEN)
#define VOICE_TO_PRICE_KEY_LEN      (VOICE_TO_O_D_KEY_LEN+ORD_O_D_LEN+ORD_CATEGORY_LEN)
#define PRICE_VOICE_MARKET_KEY_LEN  (VOICE_TO_PRICE_KEY_LEN+ORD_TYPE_LEN+ORD_PRICE_LEN+ORDERING_REST_LEN+ORD_NUMERO_RONDA_LEN+UNIQUE_CODE_LEN+ORD_POSTURE_LEN)

class CPriceG7MarketKey : public CBaseKey
{
public:
  int  m_price_key_sorting_method;

  bool operator <( const CPriceG7MarketKey &X) const;
  void operator =( const CPriceG7MarketKey &X);
  //void operator ==( const CPriceG7MarketKey &X) const; ???
  //void operator !=( const CPriceG7MarketKey &X) const; ???

  CPriceG7MarketKey (void);
  virtual void Set_Price_Key_Sorting_Method (int price_key_sorting_method)
                                            {m_price_key_sorting_method = price_key_sorting_method;}
  virtual CPriceG7MarketKey *Init ( const char * market,
                                    const char * level,
                                    const char * maturity_date,
                                    const char * xdays,
                                    const char * days,
                                    const char * text_term,
                                    const char * sub_market,
                                    const char * o_d,
                                    const char * category_code,
                                    const char * price,
                                    const char * order_type,
                                    const char * order_date, //fecha_postura
                                    const char * operation_hour,
                                    const char * order_code);
                                    

  virtual CPriceG7MarketKey *Init ( const char * market,
                                    const char * level,
                                    const char * maturity_date,
                                    const char * xdays,
                                    const char * days,
                                    const char * text_term,
                                    const char * sub_market,  
                                    const char * o_d,
                                    int          fill_value_char);
  virtual CBaseKey *Init          ( const CBaseKey *lower_key, 
                                    int             fill_value_char);
  CPriceG7MarketKey *Init  ( const char *omarket,
                             int         fill_value_char);
};

class CPriceVoicemarketKey : public CBaseKey
{
public:
  int  m_price_key_sorting_method;

  bool operator <( const CPriceVoicemarketKey &X) const;
  void operator =( const CPriceVoicemarketKey &X);
  //void operator ==( const CPriceVoicemarketKey &X) const; ???
  //void operator !=( const CPriceVoicemarketKey &X) const; ???

  CPriceVoicemarketKey  ();
  virtual void Set_Price_Key_Sorting_Method (int price_key_sorting_method)
                                            {m_price_key_sorting_method = price_key_sorting_method;}
  virtual CPriceVoicemarketKey *Init (  const char  * market,
                                        const char  * xdays,
                                        const char  * days,
                                        const char  * group_code,
                                        const char  * title_reference,
                                        const char  * currency,
                                        const char  * arrange,
                                        const char  * value_date,
                                        const char  * cama_de_monto,   // no proper translation
                                        const char  * tipo_operacion,  // no proper translation
                                        const char  * linea_pico,      // no proper translation
                                        const char  * o_d,
                                        const char  * category_code,
                                        const char  * order_type,
                                        const char  * price,
                                        const char  * ordering_rest,
                                        const char  * order_number,
                                        const char  * operation_hour,
                                        const char  * order_code);

  virtual CPriceVoicemarketKey *Init  ( const char  * market,
                                        const char  * xdays,
                                        const char  * days,
                                        const char  * group_code,
                                        const char  * title_reference,
                                        const char  * currency,
                                        const char  * arrange,
                                        const char  * value_date,
                                        const char  * cama_de_monto,   // no proper translation
                                        const char  * tipo_operacion,  // no proper translation
                                        const char  * linea_pico,      // no proper translation
                                        const char  * o_d,
                                        int           fill_value_char);
  virtual CBaseKey             *Init  ( const CBaseKey    * lower_key, 
                                        int              fill_value_char);
  CPriceVoicemarketKey *Init  ( const char *omarket,
                                int         fill_value_char);
};

#define CPRICE_DEFAULT 0
#define CPRICE_G7MODE  1
#define CPRICE_VOICE   2

#define CPRICE_BID_DESC_OFFER_ASC_SORTING   0 // Default for prices markets (i.e. G7 iForwards)
#define CPRICE_BID_ASC_OFFER_DESC_SORTING   1 // For Markets where rate is used for sorting (i.e. MX Corro).
#define CPRICE_DEFAULT_SORTING_METHOD       CPRICE_BID_DESC_OFFER_ASC_SORTING

// Key for order by price.
// this class results to be very variable, by now it depends on the market
#define PRICE_TYPE_NORMAL            0
#define PRICE_TYPE_LINKED            1
#define PRICE_TYPE_INVLINKED         2

class CPriceOrderKey : public CBaseKey
{
public:
  int            m_price_market_type;
  int            m_price_key_sorting_method;
  int            m_class_type;

  // this value depends on the market,,, or anything else
  CBaseKey      *m_internal_price;

  CPriceOrderKey  ( const CPriceOrderKey &X);

  virtual void        Init        (const char *new_key_value);
  virtual int         Fit_Pattern (CBaseKey *key_pattern);
  virtual char       *Get_Str     (void) const;
  virtual const char *Get_Key_Ptr (void) const;

  virtual bool operator < (const CPriceOrderKey &X) const;
  virtual void operator = (CPriceOrderKey &X);
  virtual bool operator ==(const CPriceOrderKey &X) const;
  virtual bool operator !=(const CPriceOrderKey &X) const;
  
  void            Clear_Price  (void);

public:
  CPriceOrderKey  ();
  CPriceOrderKey  ( int class_type);
  virtual ~CPriceOrderKey ();

  CPriceOrderKey *Init  (const CPriceOrderKey *lower_key, 
                         int                    fill_value_char);
 
  CPriceOrderKey *Init   (const char *market,
                          int         fill_value_char);

  CPriceOrderKey *Build     (int price_market_type, int new_price_key_sorting_method, int class_type);
  CPriceOrderKey *Build     (CMessage *ow_message);
  CPriceOrderKey *Build     (CMessage *ow_message, const char *bid_offer, int fill_char);
  CPriceOrderKey *Build_Lnk (CMessage *ow_message, const char *bid_offer, int fill_char);

};

class CDatatecTradeKey : public CBaseKey
{
public:
  CDatatecTradeKey();

  CDatatecTradeKey  *Init(const char *market,
                          const char *hour,
                          const char *order_code,
                          const char *order_code_1,
                          const char *c_v);
};

#if DGIAPP || GDIAPP

class CGTNTradeKey : public CBaseKey
{
public:
  CGTNTradeKey();
};

#endif //#ifdef DGIAPP

class CPriceOrderKeyLink : public CPriceOrderKey
{
public:
  CPriceOrderKeyLink() : CPriceOrderKey(PRICE_TYPE_LINKED) {}
};

class CPriceOrderKeyInvLink : public CPriceOrderKey
{
public:
  CPriceOrderKeyInvLink() : CPriceOrderKey(PRICE_TYPE_INVLINKED) {}
};

class CPriceG7MarketKeyLink : public CPriceG7MarketKey
{
};

class CPriceG7MarketKeyInvLink : public CPriceG7MarketKey
{
};

class CPriceVoicemarketKeyLink : public CPriceVoicemarketKey
{
};

class CPriceVoicemarketKeyInvLink : public CPriceVoicemarketKey
{
};

class CPosturaKey : public CBaseKey
{
public:
  CPosturaKey();
};

class CAceptacionKey : public CBaseKey
{
public:
  CAceptacionKey();
};

class COrderKey: public CBaseKey
{

public:
  COrderKey ();
  COrderKey *Init  (const char *market, const char *order_code, const char *o_d, const char *sub_market, const char *text_term);
  COrderKey *Init  (const char *market, const char *order_code, int fill_char, bool use_order_code_len = false);
  void       Init  (const char *new_key_value);
  COrderKey *Init  (int fill_char);
  int        Fit_Pattern (CBaseKey *key_pattern);
  char      *Get_Str (void) const;
  const char *Get_Order_Code_Ptr (void) {return (m_key + ORD_MARKET_LEN);}
  const char *Get_Market() {return m_key;}
  const char *Get_TextTerm() {return (m_key + ORD_MARKET_LEN + ORD_POSTURE_LEN + ORD_O_D_LEN + ORD_SMARKET_LEN);}
  bool operator <(const CBaseKey &X) const;
  bool operator ==( const CBaseKey &X) const;
};

class COrderKeyUser: public CBaseKey
{
public:
  COrderKeyUser ();
  COrderKeyUser *Init  (const char *primaryuser_code,
                        const char *user,
                        const char *market,
                        const char *order_code,
                        const char *o_d,
                        const char *sub_market,
                        const char *text_term);

  COrderKeyUser *Init  (const char *primaryuser_code,
                        const char *user,
                        int         fill_char);
};

class COrderKeyLoginName: public CBaseKey
{
public:
  COrderKeyLoginName ();
  COrderKeyLoginName *Init  (const char *market,
                             const char *login_name,
                             const char *o_d,
                             const char *order_code);
  COrderKeyLoginName *Init  (const char *market,
                             const char *login,
                             int         fill_char);
};

class CSuperGroupKey: public CBaseKey
{
public:
  CSuperGroupKey ();
};

class CVoiceTermKey: public CBaseKey
{
public:
  CVoiceTermKey ();
  CVoiceTermKey  *Init(const char *market, const char *voiceterm_code);
};

class CVoiceSummaryKey: public CBaseKey
{
public:
  CVoiceSummaryKey ();
  CVoiceSummaryKey *Init(const char *market,
                         const char *group_code,
                         const char *title_reference,
                         const char *text_term,
                         const char *currency,
                         const char *arrange,
                         const char *xdays,
                         const char *days,
                         const char *value_date,
                         const char *cama_de_monto,   // no proper translation
                         const char *tipo_operacion,  // no proper translation
                         const char *linea_pico);     // no proper translation

};

class CTableList
{
  int          m_max_tables;
  int          m_cur_table;
  CvArray <int> m_tables;

public:
  int          m_num_tables;
  int          m_restricted_dest;
       CTableList ();
       CTableList (int unic_table);
  void Clean_Tables (void); 

  void Add_Table (int new_table); 
  int  First_Table (void);
  int  Next_Table  (void);
  void operator =(CTableList &X);
  int  Copy_Tables (void *ptr, int max_length);
};

template <class KEY_TYPE, class TLIST_TYPE >
class CTRules
{
  int  m_max_rules;
  int  m_num_rules;

  // each key pattern hast a table num associated
  CVector <KEY_TYPE> m_key_patterns;
  CvArray  <int>      m_num_tables;

public:
       CTRules    (int max_rules = INITIAL_MAXIMUM_RULES);
      ~CTRules    (void);
  void Add_Rule   (KEY_TYPE *key, int table);
  void Init_TList (KEY_TYPE *key, TLIST_TYPE *table_list, int fit_first_only);
};

// Key for InstitTable
class CInstitKey : public CBaseKey
{
public:
  CInstitKey();  
  CInstitKey *Init(const char *cod_instit);
  CInstitKey *Init(const char *cod_instit, int fill_char);
};

#if DAPI || GDIAPP || DMEAPP
class CVoiceGroupKey: public CBaseKey
{
public:
  CVoiceGroupKey ();
};
#endif


#if DAPI || DMEAPP
class CPInsPermKey : public CBaseKey
{
public:

  CPInsPermKey();
};
#endif

#ifdef DAPI
// Key sequence number
class CSeqKey : public CBaseKey
{
public:
  CSeqKey ();
  CSeqKey *Init (const char *data_server, const char *date, unsigned int seq_number);
  bool operator  <(const CSeqKey &X) const;
};

class CPuKey : public CBaseKey
{
public:

  CPuKey();
};

class CCityKey : public CBaseKey
{
public:

  CCityKey();
};

class CCountryKey : public CBaseKey
{
public:

  CCountryKey();
};

class CMoneyKey : public CBaseKey
{
public:

  CMoneyKey();
};

class CFixedIncomeMarketsKey: public CBaseKey
{
public:
  CFixedIncomeMarketsKey ();
};

#endif

class CMapKeyN  
{
public:
  int                   m_key_n;
  virtual ~CMapKeyN(){}
};

// iterator to loop over a group of records of the table
// or over all the table
template <class KEY_TYPE, class REC_TYPE>
class CMapKeyIterator : public CMapKeyN
{
public:
  KEY_TYPE              m_map_first_key;
  KEY_TYPE              m_map_last_key;
  KEY_TYPE              m_map_prev_key; // used to ensure not infinite loop, when inside errors occur
  int                   m_range_to_end;
};

#define MKITERATOR CMapKeyIterator<KEY_TYPE,REC_TYPE>

template <class KEY_TYPE, class REC_TYPE>
class CMapKeyTable
{
public:
  Map_Key_Table         m_key_table;
  KEY_TYPE              m_key;
  KEY_TYPE              m_key_old;        // for updating keys
  KEY_TYPE              m_map_first_key;
  KEY_TYPE              m_map_last_key;
  int                   m_range_to_end;
  typename Map_Key_Table::iterator  m_key_it;

public:
  CMapKeyTable();

  void  Initialize              (void);
  void  Add_Local_Record        (int position, REC_TYPE *record_to_insert);
  int   Search_Key_Record       (CMessage *record_to_search);
  int   Search_Key_Record       (void *key1);
  void  Delete_Record           (REC_TYPE *record_to_delete);
  void  Begin_Update_Keys       (REC_TYPE *old_rec);
  void  End_Update_Keys         (int update_pos, REC_TYPE *new_rec);
  int   Find_Key_Range_It       (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Next_in_Key_Range_It    (CMapKeyN *mkit);
  int   Find_Key_Last_Range_It  (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Prev_Key_Last_Range_It  (CMapKeyN *mkit);
  int   Begin                   (void);
  int   Next                    (void);
  int   Get_Num_Records         (void);
  int   Get_Num_Keys            (void);
};

template <class KEY_TYPE, class REC_TYPE>
CMapKeyTable <KEY_TYPE,REC_TYPE>::CMapKeyTable()
{ _SA_(0) 
  m_range_to_end  = false;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Begin (void)
{ _SA_(0) 
  m_key_it = m_key_table.begin();
  if (m_key_it == m_key_table.end ())
    return -1;
  return m_key_it->second;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Next (void)
{ _SA_(0) 
  if (m_key_it == m_key_table.end ())
    return -1;
  ++m_key_it;
  if (m_key_it == m_key_table.end ())
    return -1;
  return m_key_it->second;
}

template <class KEY_TYPE, class REC_TYPE>
void  CMapKeyTable <KEY_TYPE,REC_TYPE>::Initialize(void)
{ _SA_(0)   
  m_key_table.clear();
  m_range_to_end  = false;
}

template <class KEY_TYPE, class REC_TYPE>
void  CMapKeyTable <KEY_TYPE,REC_TYPE>::Add_Local_Record (int position, REC_TYPE *record_to_insert)
{ _SA_(0) 
((CMessage *)record_to_insert)->Init_Key     (&m_key);
  // map the keys to the current record
  m_key_table  [m_key]      = position; 
}

template <class KEY_TYPE, class REC_TYPE>
int   CMapKeyTable <KEY_TYPE,REC_TYPE>::Search_Key_Record (CMessage *record_to_search)
{ _SA_(0) 
  typename Map_Key_Table::iterator key_it;

  record_to_search->Init_Key (&m_key);

  key_it = m_key_table.find (m_key);

  if (key_it == m_key_table.end())
  {
    return -1;
  }

  return key_it->second;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Search_Key_Record (void *key1)
{ _SA_(0) 
  typename Map_Key_Table::iterator key_it;
  KEY_TYPE *cur_KEY_TYPE;

  cur_KEY_TYPE = dynamic_cast <KEY_TYPE *> ((CBaseKey *)key1);
  if ( key1 && cur_KEY_TYPE == NULL)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. key given as argument is not the correct class!");
    return -1;
  }

  key_it = m_key_table.find (*cur_KEY_TYPE);

  if (key_it == m_key_table.end())
  {
    return -1;
  }

  return key_it->second;
}

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyTable <KEY_TYPE,REC_TYPE>::Delete_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  typename Map_Key_Table::iterator  key_it;
((CMessage *)record_to_delete)->Init_Key  (&m_key);

  // Delete the map to the first key
  key_it = m_key_table.find (m_key);
  if (key_it == m_key_table.end())
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Delete_Record- The key %s was not found", m_key.Get_Str());
    return;
  }
  m_key_table.erase (key_it);
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Get_Num_Records(void)
{ _SA_(0) 
  return m_key_table.size ();
}

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyTable <KEY_TYPE,REC_TYPE>::Begin_Update_Keys (REC_TYPE *old_rec)
{ _SA_(0) 
  ((CMessage *)old_rec)->Init_Key  (&m_key_old);
}

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyTable <KEY_TYPE,REC_TYPE>::End_Update_Keys (int update_pos, REC_TYPE *new_rec)
{ _SA_(0) 
    typename Map_Key_Table::iterator  key_it;
    ((CMessage *)new_rec)->Init_Key  (&m_key);

    // If the key has changed, update the key map.
    if (!(m_key_old == m_key))
    {
      key_it = m_key_table.find(m_key_old);
      if (key_it != m_key_table.end())
        m_key_table.erase(key_it);
      else
        COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                    "Error. Key changed was not found %s", m_key_old.Get_Str());
      m_key_table[m_key] = update_pos;
    }
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Find_Key_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  typename Map_Key_Table::iterator map_key_it_first;
  int      position;
  MKITERATOR  * Cur_MKITERATOR;
  KEY_TYPE    * Low_KEY_TYPE;
  KEY_TYPE    * High_KEY_TYPE;


  Cur_MKITERATOR = dynamic_cast <MKITERATOR *> ((CMapKeyN *)mkit);
  if ( Cur_MKITERATOR == NULL)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Table Iterator given as argument is not the correct class!");
    return -1;
  }

  if (!key_low)
    Low_KEY_TYPE = (KEY_TYPE *)key_low;
  else
  {
    Low_KEY_TYPE = dynamic_cast <KEY_TYPE *> ((CBaseKey *)key_low);
    if (Low_KEY_TYPE == NULL)
    {
      COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                  "Error. low key given as argument is not the correct class!");
      return -1;
    }
  }

  if (!key_high)
    High_KEY_TYPE = (KEY_TYPE *)key_high;
  else
  {
    High_KEY_TYPE = dynamic_cast <KEY_TYPE *> ((CBaseKey *)key_high);
    if (High_KEY_TYPE == NULL)
    {
      COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                  "Error. high key given as argument is not the correct class!");
      return -1;
    }
  }

  // check lower limit
  if (!Low_KEY_TYPE)
    map_key_it_first  = m_key_table.begin ();    
  else
    map_key_it_first  = m_key_table.lower_bound (*Low_KEY_TYPE);

  Cur_MKITERATOR->m_key_n = key_n;

  // check upper limit
  if (!High_KEY_TYPE)
  {
    Cur_MKITERATOR->m_range_to_end    = true;
  }
  else
  { 
    Cur_MKITERATOR->m_map_last_key    = *High_KEY_TYPE;
    Cur_MKITERATOR->m_range_to_end    = false;
  }

  if (map_key_it_first == m_key_table.end()) 
    return -1;

  // Get the complete Key from the first element of the iterator
  Cur_MKITERATOR->m_map_first_key = (KEY_TYPE &)map_key_it_first ->first;

  // check if we cross the threshold
  if (!Cur_MKITERATOR->m_range_to_end)
    if (Cur_MKITERATOR->m_map_last_key < Cur_MKITERATOR->m_map_first_key)
      return -1; // it's outside the limit

  // Get the corresponding record index
  position  = map_key_it_first ->second;    
  
  return position;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Next_in_Key_Range_It (CMapKeyN *mkit)
{ _SA_(0) 
  typename Map_Key_Table::iterator map_key_it_first;
  int position;
  MKITERATOR *Cur_MKITERATOR;

  Cur_MKITERATOR = dynamic_cast <MKITERATOR *> ((CMapKeyN *)mkit);
  if ( Cur_MKITERATOR == NULL)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Table Iterator given as argument is not the correct class!");
    return -1;
  }

  // Locate the next element taking into the account the first key
  map_key_it_first = m_key_table.upper_bound (Cur_MKITERATOR->m_map_first_key);

  // Make some comparisions before returning a link value
  if (map_key_it_first == m_key_table.end())
    return -1;

  Cur_MKITERATOR->m_map_prev_key  = Cur_MKITERATOR->m_map_first_key;

  // Get the complete Key from the first element of next iterator
  Cur_MKITERATOR->m_map_first_key = (KEY_TYPE &)map_key_it_first ->first;

  // if we are going back.. stop this. som error is behind ther scene 
  if (Cur_MKITERATOR->m_map_first_key < Cur_MKITERATOR->m_map_prev_key)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Going through a range got BACK in the mapping!");
    return -1;
  }

  // Get the last element of the range
  if (!Cur_MKITERATOR->m_range_to_end)
    if (Cur_MKITERATOR->m_map_last_key < Cur_MKITERATOR->m_map_first_key)
      return -1;

  position = map_key_it_first ->second;

  return position;
}


template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Find_Key_Last_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  typename Map_Key_Table::iterator map_key_it_last;
  int      position;
  MKITERATOR  * Cur_MKITERATOR;
  KEY_TYPE    * Low_KEY_TYPE;
  KEY_TYPE    * High_KEY_TYPE;

  Cur_MKITERATOR = dynamic_cast <MKITERATOR *> ((CMapKeyN *)mkit);
  if (Cur_MKITERATOR == NULL)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Table Iterator given as argument is not the correct class!");
    return -1;
  }

  if (!key_low)
    Low_KEY_TYPE = (KEY_TYPE *)key_low;
  else
  {
    Low_KEY_TYPE = dynamic_cast <KEY_TYPE *> ((CBaseKey *)key_low);
    if (Low_KEY_TYPE == NULL)
    {
      COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                  "Error. low key given as argument is not the correct class!");
      return -1;
    }
  }

  if (!key_high)
    High_KEY_TYPE = (KEY_TYPE *)key_high;
  else
  {
    High_KEY_TYPE = dynamic_cast <KEY_TYPE *> ((CBaseKey *)key_high);
    if (High_KEY_TYPE == NULL)
    {
      COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                  "Error. high key given as argument is not the correct class!");
      return -1;
    }
  }

  // If the mapping does not have elements, there is no valid element to return.
  // Avoid infinite loop in decrement operator.
  if (m_key_table.size() <= 0)
    return -1;

  // check upper limit
  if (!Low_KEY_TYPE)
  {
    Cur_MKITERATOR->m_range_to_end    = true;
  }
  else
  { 
    Cur_MKITERATOR->m_map_first_key   = *Low_KEY_TYPE;
    Cur_MKITERATOR->m_range_to_end    = false;
  }

  // check lower limit
  if (!High_KEY_TYPE)
    map_key_it_last  = m_key_table.end ();    
  else
    map_key_it_last  = m_key_table.upper_bound (*High_KEY_TYPE);

  Cur_MKITERATOR->m_key_n = key_n;

  if (--map_key_it_last == m_key_table.end()) 
    return -1;

  // Get the complete Key from the first element of the iterator
  Cur_MKITERATOR->m_map_last_key = (KEY_TYPE &)map_key_it_last ->first;

  // check if we cross the threshold
  if (!Cur_MKITERATOR->m_range_to_end)
    if (Cur_MKITERATOR->m_map_last_key < Cur_MKITERATOR->m_map_first_key)
      return -1; // it's outside the limit

  // Get the corresponding record index
  position  = map_key_it_last ->second;    
  
  return position;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyTable <KEY_TYPE,REC_TYPE>::Prev_Key_Last_Range_It (CMapKeyN *mkit)
{ _SA_(0) 
  typename Map_Key_Table::iterator map_key_it_last;
  typename Map_Key_Table::iterator map_key_it_last_bak;
  int position;
  MKITERATOR *Cur_MKITERATOR;

  Cur_MKITERATOR = dynamic_cast <MKITERATOR *> ((CMapKeyN *)mkit);
  if ( Cur_MKITERATOR == NULL)
  {
    COMMON_Put_Syslog_Message ( ERROR_LOG_MSG, 
                                "Error. Table Iterator given as argument is not the correct class!");
    return -1;
  }

  // If the mapping does not have elements, there is no valid element to return.
  // Avoid infinite loop in decrement operator.
  if (m_key_table.size() <= 0)
    return -1;

  // locate the current if any.
  map_key_it_last = m_key_table.lower_bound (Cur_MKITERATOR->m_map_last_key);

  // Backup the current iterator because decrement operator returns the same
  // iterator when there is only one element in the mapping.
  map_key_it_last_bak = map_key_it_last;

  // now, go for the previous
  --map_key_it_last;
  // If the decremented iterator is the same than the returned by lower_bound() (one element mapping),
  // or the iterator is invalid, return that there is no previous element.
  if ((map_key_it_last_bak == map_key_it_last) || map_key_it_last == m_key_table.end())
    return -1;

  // Get the complete Key from this last element of the iterator
  Cur_MKITERATOR->m_map_last_key = (KEY_TYPE &)map_key_it_last ->first;

  // check if we cross the threshold
  if (!Cur_MKITERATOR->m_range_to_end)
    if (Cur_MKITERATOR->m_map_last_key < Cur_MKITERATOR->m_map_first_key)  // if we have excedeed the limit,
      return -1;

  position = map_key_it_last ->second;

  return position;
}

///////////////////////////////////////////////////////////////////////////////
// CIdxOrder
///////////////////////////////////////////////////////////////////////////////

template <class REC_TYPE>
class CIdxOrder
{
  unsigned int         m_last_idx;
public:
  Map_Idx_Table         m_idx_table;
  CIdxKey               m_idx_key;
  
public:
  CIdxOrder() { m_last_idx = 0;}
  void  Initialize              (void);
  void  Add_Local_Record        (int position, REC_TYPE *record_to_insert);
  void  Add_Record              (int position, REC_TYPE *record_to_insert, int put_environment = 1);
  int   Search_Idx_Record       (REC_TYPE *record_to_search);
  int   Search_By_Idx           (unsigned int idx);
  int   Delete_By_Idx_Record    (REC_TYPE *record_to_delete);
  void  Begin_Update_Keys       (REC_TYPE *old_rec);
  void  End_Update_Keys         (int update_pos, REC_TYPE *new_rec);
  void  Add_Idx                 (int position, unsigned int rec_idx, REC_TYPE *record_to_insert);
  int   Get_Num_Records         (void);
};

template <class REC_TYPE>
void CIdxOrder<REC_TYPE>::Initialize (void)
{ _SA_(0) 
  m_idx_table.clear();
}

template <class REC_TYPE>
void CIdxOrder<REC_TYPE>::Add_Local_Record (int position, REC_TYPE *record_to_insert)
{ _SA_(0) 
  ((CMessage *)record_to_insert)->Init_Key (&m_idx_key);
  // map the idx to the current record
  m_idx_table [m_idx_key] = position;
}

template <class REC_TYPE>
int CIdxOrder<REC_TYPE>::Get_Num_Records (void)
{ _SA_(0) 
  return (int)m_idx_table.size ();
}
template <class REC_TYPE>
void CIdxOrder<REC_TYPE>::Add_Record(int position, REC_TYPE *record_to_insert, int put_environment)
{ _SA_(0) 
  unsigned int rec_idx;
  int           cycle;

  // generate a new idx
  rec_idx = ++m_last_idx;
  if (rec_idx < glb_min_gtable_rec_index || rec_idx >= glb_max_gtable_rec_index)
    rec_idx = glb_min_gtable_rec_index;

  m_idx_key.Init (0, rec_idx);
  for (cycle=0;;)
  {
    // in this case a zero data code is used
    if (m_idx_table.find (m_idx_key) == m_idx_table.end())
    {
      m_last_idx = rec_idx;
      break;
    }

    if (++rec_idx >= glb_max_gtable_rec_index)
    {
      if (++cycle >= 2)
      {
        if (!Glb_range_limit_err_msg)
        {
          // El Rango limite para los registros de la tabla global ha sido alcanzado
          COMMON_Put_Syslog_Message(ERROR_LOG_MSG, 
                                    "Range limit for global table records (%u) has been reached. !!!",
                                    (glb_max_gtable_rec_index - glb_min_gtable_rec_index + 1));
          Glb_range_limit_err_msg = TRUE;
        }
        break;
      }
      rec_idx = glb_min_gtable_rec_index;
    }
    // use this hext idx
    m_idx_key.rec_idx = rec_idx;
  }

  // assign the idx to the internal field
  (*record_to_insert)[REC_IDX]      = rec_idx;

  if (put_environment)
    (*record_to_insert)[ENVIRONMENT] = Glb_range_environment;

  // map the idx to the current record
  m_idx_table  [m_idx_key]  = position;
}

template <class REC_TYPE>
void CIdxOrder<REC_TYPE>::Add_Idx (int position, unsigned int rec_idx, REC_TYPE *record_to_insert)
{ _SA_(0) 
  m_idx_key.Init (0, rec_idx);

  // Check collitions
  if (m_idx_table.find (m_idx_key) != m_idx_table.end())
  {
    if (!Glb_range_limit_err_msg)
    {
      COMMON_Put_Syslog_Message(ERROR_LOG_MSG, 
                                "Idx value (%u) for global table already exists. !!!",
                                rec_idx);
      Glb_range_limit_err_msg = TRUE;
    }
  }

  // assign the idx to the internal field
  (*record_to_insert)[REC_IDX]  = rec_idx;

  // map the idx to the current record
  m_idx_table  [m_idx_key]  = position;
}

template <class REC_TYPE>
int CIdxOrder<REC_TYPE>::Search_Idx_Record (REC_TYPE *record_to_search)
{ _SA_(0) 
  Map_Idx_Table::iterator idx_it;
  ((CMessage *)record_to_search)->Init_Key (&m_idx_key);

  // find the record1
  idx_it = m_idx_table.find (m_idx_key);

  if (idx_it == m_idx_table.end())
  {
    // LOG_Put_Err_Msg("Error. Search_Idx_Record- The idx %ld was not found", m_idx_key.rec_idx);
    return -1;
  }

  return idx_it->second;
}

template <class REC_TYPE>
int CIdxOrder<REC_TYPE>::Search_By_Idx  (unsigned int idx)
{ _SA_(0) 
  Map_Idx_Table::iterator idx_it;

  m_idx_key.Init (0, idx);

  // find the record1
  idx_it = m_idx_table.find (m_idx_key);

  if (idx_it == m_idx_table.end())
  {
    // LOG_Put_Err_Msg("Error. Search_By_Idx- The idx %ld was not found", m_idx_key.rec_idx);
    return -1;
  }

  return idx_it->second;
}

template <class REC_TYPE>
int CIdxOrder<REC_TYPE>::Delete_By_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int       delete_pos;
  Map_Idx_Table::iterator idx_it;

  ((CMessage *)record_to_delete)->Init_Key (&m_idx_key);
  // find the record
  idx_it = m_idx_table.find (m_idx_key);
  if (idx_it == m_idx_table.end())
  {
    // LOG_Put_Err_Msg("Error. Delete_Local_Record- The idx %ld was not found", idx);
    return -1;
  }
  delete_pos = idx_it->second;
  // erase the map items
  m_idx_table.erase (idx_it);

  return delete_pos;
}


///////////////////////////////////////////////////////////////////////////////
// CMap3KeyOrder
///////////////////////////////////////////////////////////////////////////////
template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
class CMap3KeyOrder
{
public:
  CIdxOrder   <REC_TYPE>            m_key_idx;
  CMapKeyTable<KEY_TYPE,  REC_TYPE> m_key1_mkt;
  CMapKeyTable<KEY2_TYPE, REC_TYPE> m_key2_mkt;
  CMapKeyTable<KEY3_TYPE, REC_TYPE> m_key3_mkt;
  int          m_num_key;

public:
  void  Initialize              (void);
  void  Add_Local_Record        (int position, REC_TYPE *record_to_insert);
  void  Add_Record              (int position, REC_TYPE *record_to_insert, int put_environment = 1);
  int   Search_Idx_Record       (REC_TYPE *record_to_search);
  int   Search_By_Idx           (unsigned int idx);
  int   Search_NKey_Record      (CMessage *record_to_search, int key_n);
  int   Search_NKey_Record      (void *key2, int key_n);
  int   Delete_By_Idx_Record    (REC_TYPE *record_to_delete);
  int   Delete_But_Not_Idx_Record    (REC_TYPE *record_to_delete);

  void  Begin_Update_Keys       (REC_TYPE *old_rec);
  void  End_Update_Keys         (int update_pos, REC_TYPE *new_rec);

  int   Search_Key_Record       (CMessage *record_to_search, int key_n);
  int   Search_Key_Record       (void *key, int key_n);
  int   Find_Key_Range_It       (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Next_in_Key_Range_It    (CMapKeyN *mkit, int key_n);
  int   Find_Key_Last_Range_It  (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Prev_Key_Last_Range_It  (CMapKeyN *mkit, int key_n);
  int   Get_Num_Records         (void);
  int   Get_Num_Keys            (void);
};

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
void CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Initialize (void)
{ _SA_(0) 
  m_key_idx.Initialize();
  m_key1_mkt.Initialize();
  m_key2_mkt.Initialize();
  m_key3_mkt.Initialize();
  m_num_key = THIRD_KEY;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
void CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Add_Local_Record (int position, REC_TYPE *record_to_insert)
{ _SA_(0) 
  m_key_idx.Add_Local_Record  (position, record_to_insert);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
  m_key2_mkt.Add_Local_Record (position, record_to_insert);
  m_key3_mkt.Add_Local_Record (position, record_to_insert);
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
void CMap3KeyOrder<KEY_TYPE, KEY2_TYPE, KEY3_TYPE, REC_TYPE>::Add_Record(int position, REC_TYPE *record_to_insert, int put_environment)
{ _SA_(0) 
  m_key_idx.Add_Record        (position, record_to_insert, put_environment);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
  m_key2_mkt.Add_Local_Record (position, record_to_insert);
  m_key3_mkt.Add_Local_Record (position, record_to_insert);

}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int  CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Get_Num_Keys (void)
{ _SA_(0) 
  return m_num_key;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int  CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Get_Num_Records (void)
{ _SA_(0) 
  return m_key_idx.Get_Num_Records();
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
void  CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Begin_Update_Keys   (REC_TYPE *old_rec)
{ _SA_(0) 
  m_key1_mkt.Begin_Update_Keys   (old_rec);
  m_key2_mkt.Begin_Update_Keys   (old_rec);
  m_key3_mkt.Begin_Update_Keys   (old_rec);
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
void  CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::End_Update_Keys     (int update_pos, REC_TYPE *new_rec)
{ _SA_(0) 
  m_key1_mkt.End_Update_Keys   (update_pos, new_rec);
  m_key2_mkt.End_Update_Keys   (update_pos, new_rec);
  m_key3_mkt.End_Update_Keys   (update_pos, new_rec);
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Search_Idx_Record (REC_TYPE *record_to_search)
{ _SA_(0) 
  return m_key_idx.Search_Idx_Record (record_to_search);
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Search_By_Idx (unsigned int idx)
{ _SA_(0) 
  return m_key_idx.Search_By_Idx (idx);
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Search_Key_Record (CMessage *record_to_search, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (record_to_search);
  case 2:
    return m_key2_mkt.Search_Key_Record (record_to_search);
  case 3:
    return m_key3_mkt.Search_Key_Record (record_to_search);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Search_Key_Record (void *key, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (key);
  case 2:
    return m_key2_mkt.Search_Key_Record (key);
  case 3:
    return m_key3_mkt.Search_Key_Record (key);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Delete_By_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int delete_pos;

  delete_pos = m_key_idx.Delete_By_Idx_Record (record_to_delete);
  m_key1_mkt.Delete_Record (record_to_delete);
  m_key2_mkt.Delete_Record (record_to_delete);
  m_key3_mkt.Delete_Record (record_to_delete);

  return delete_pos;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Delete_But_Not_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  m_key1_mkt.Delete_Record (record_to_delete);
  m_key2_mkt.Delete_Record (record_to_delete);
  m_key3_mkt.Delete_Record (record_to_delete);

  return 0;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Find_Key_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Range_It (key_low, key_high, mkit, key_n);
  case 2:
    return m_key2_mkt.Find_Key_Range_It (key_low, key_high, mkit, key_n);
  case 3:
    return m_key3_mkt.Find_Key_Range_It (key_low, key_high, mkit, key_n);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Next_in_Key_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Next_in_Key_Range_It (mkit);
  case 2:
    return m_key2_mkt.Next_in_Key_Range_It (mkit);
  case 3:
    return m_key3_mkt.Next_in_Key_Range_It (mkit);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Find_Key_Last_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Last_Range_It (key_low, key_high, mkit, key_n);
  case 2:
    return m_key2_mkt.Find_Key_Last_Range_It (key_low, key_high, mkit, key_n);
  case 3:
    return m_key3_mkt.Find_Key_Last_Range_It (key_low, key_high, mkit, key_n);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class KEY3_TYPE, class REC_TYPE>
int CMap3KeyOrder<KEY_TYPE,KEY2_TYPE,KEY3_TYPE,REC_TYPE>::Prev_Key_Last_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Prev_Key_Last_Range_It (mkit);
  case 2:
    return m_key2_mkt.Prev_Key_Last_Range_It (mkit);
  case 3:
    return m_key3_mkt.Prev_Key_Last_Range_It (mkit);
  }
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
// CMap2KeyOrder
///////////////////////////////////////////////////////////////////////////////
template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
class CMap2KeyOrder
{
public:
  CIdxOrder   <REC_TYPE>            m_key_idx;
  CMapKeyTable<KEY_TYPE,  REC_TYPE> m_key1_mkt;
  CMapKeyTable<KEY2_TYPE, REC_TYPE> m_key2_mkt;
  int          m_num_key;

public:
  void  Initialize              (void);
  void  Add_Local_Record        (int position, REC_TYPE *record_to_insert);
  void  Add_Record              (int position, REC_TYPE *record_to_insert, int put_environment = 1);
  int   Search_Idx_Record       (REC_TYPE *record_to_search);
  int   Search_By_Idx           (unsigned int idx);
  int   Search_Key_Record       (CMessage *record_to_search, int key_n);
  int   Search_Key_Record       (void *key2, int key_n);
  int   Delete_By_Idx_Record    (REC_TYPE *record_to_delete);
  int   Delete_But_Not_Idx_Record (REC_TYPE *record_to_delete);

  void  Begin_Update_Keys       (REC_TYPE *old_rec);
  void  End_Update_Keys         (int update_pos, REC_TYPE *new_rec);

  int   Find_Key_Range_It       (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Next_in_Key_Range_It    (CMapKeyN *mkit, int key_n);
  int   Find_Key_Last_Range_It  (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Prev_Key_Last_Range_It  (CMapKeyN *mkit, int key_n);
  int   Get_Num_Records         (void);
  int   Get_Num_Keys            (void);
};

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
void CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Initialize (void)
{ _SA_(0) 
  m_key_idx.Initialize();
  m_key1_mkt.Initialize();
  m_key2_mkt.Initialize();
  m_num_key = SECOND_KEY;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
void CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Add_Local_Record (int position, REC_TYPE *record_to_insert)
{ _SA_(0) 
  m_key_idx.Add_Local_Record (position, record_to_insert);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
  m_key2_mkt.Add_Local_Record (position, record_to_insert);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
void CMap2KeyOrder<KEY_TYPE, KEY2_TYPE, REC_TYPE>::Add_Record(int position, REC_TYPE *record_to_insert, int put_environment)
{ _SA_(0) 
  m_key_idx.Add_Record        (position, record_to_insert, put_environment);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
  m_key2_mkt.Add_Local_Record (position, record_to_insert);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int  CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Get_Num_Records (void)
{ _SA_(0) 
  return m_key_idx.Get_Num_Records();
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int  CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Get_Num_Keys (void)
{ _SA_(0) 
  return m_num_key;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
void  CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Begin_Update_Keys   (REC_TYPE *old_rec)
{ _SA_(0) 
  m_key1_mkt.Begin_Update_Keys   (old_rec);
  m_key2_mkt.Begin_Update_Keys   (old_rec);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
void  CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::End_Update_Keys     (int update_pos, REC_TYPE *new_rec)
{ _SA_(0) 
  m_key1_mkt.End_Update_Keys   (update_pos, new_rec);
  m_key2_mkt.End_Update_Keys   (update_pos, new_rec);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Search_Idx_Record (REC_TYPE *record_to_search)
{ _SA_(0) 
  return m_key_idx.Search_Idx_Record (record_to_search);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Search_By_Idx (unsigned int idx)
{ _SA_(0) 
  return m_key_idx.Search_By_Idx (idx);
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Search_Key_Record (CMessage *record_to_search, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (record_to_search);
  case 2:
    return m_key2_mkt.Search_Key_Record (record_to_search);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Search_Key_Record (void *key, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (key);
  case 2:
    return m_key2_mkt.Search_Key_Record (key);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Delete_By_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int delete_pos;

  delete_pos = m_key_idx.Delete_By_Idx_Record (record_to_delete);
  m_key1_mkt.Delete_Record (record_to_delete);
  m_key2_mkt.Delete_Record (record_to_delete);

  return delete_pos;
}
template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Delete_But_Not_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  m_key1_mkt.Delete_Record (record_to_delete);
  m_key2_mkt.Delete_Record (record_to_delete);

  return 0;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Find_Key_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Range_It (key_low, key_high, mkit, key_n);
  case 2:
    return m_key2_mkt.Find_Key_Range_It (key_low, key_high, mkit, key_n);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Next_in_Key_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Next_in_Key_Range_It (mkit);
  case 2:
    return m_key2_mkt.Next_in_Key_Range_It (mkit);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Find_Key_Last_Range_It (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Last_Range_It (key_low, key_high, mkit, key_n);
  case 2:
    return m_key2_mkt.Find_Key_Last_Range_It (key_low, key_high, mkit, key_n);
  }
  return -1;
}

template <class KEY_TYPE, class KEY2_TYPE, class REC_TYPE>
int CMap2KeyOrder<KEY_TYPE,KEY2_TYPE,REC_TYPE>::Prev_Key_Last_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Prev_Key_Last_Range_It (mkit);
  case 2:
    return m_key2_mkt.Prev_Key_Last_Range_It (mkit);
  }
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
// CMapKeyOrder
///////////////////////////////////////////////////////////////////////////////
template <class KEY_TYPE, class REC_TYPE>
class CMapKeyOrder
{
public:
  CIdxOrder     <REC_TYPE>            m_key_idx;
  CMapKeyTable  <KEY_TYPE,  REC_TYPE> m_key1_mkt;
  int            m_num_key;
  
public:
  void  Initialize              (void);
  void  Add_Local_Record        (int position, REC_TYPE *record_to_insert);
  void  Add_Record              (int position, REC_TYPE *record_to_insert, int put_environment = 1);
  int   Search_Idx_Record       (REC_TYPE *record_to_search);
  int   Search_By_Idx           (unsigned int idx);
  int   Search_Key_Record       (CMessage *record_to_search, int key_n);
  int   Search_Key_Record       (void *key2, int key_n);
  int   Delete_By_Idx_Record    (REC_TYPE *record_to_delete);
  int   Delete_But_Not_Idx_Record (REC_TYPE *record_to_delete);
  int   Get_Num_Records           (void);
  int   Get_Num_Keys              (void);
  
  void  Begin_Update_Keys       (REC_TYPE *old_rec);
  void  End_Update_Keys         (int update_pos, REC_TYPE *new_rec);

  int   Find_Key_Range_It       (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Next_in_Key_Range_It    (CMapKeyN *mkit, int key_n);
  int   Find_Key_Last_Range_It  (CBaseKey *key_low, CBaseKey *key_high, CMapKeyN *mkit, int key_n);
  int   Prev_Key_Last_Range_It  (CMapKeyN *mkit, int key_n);
};

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyOrder<KEY_TYPE,REC_TYPE>::Initialize (void)
{ _SA_(0) 
  m_key_idx.Initialize();
  m_key1_mkt.Initialize();
  m_num_key = FIRST_KEY;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Get_Num_Records (void)
{ _SA_(0) 
  return m_key_idx.Get_Num_Records();
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Get_Num_Keys (void)
{ _SA_(0) 
  return m_num_key;
}

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyOrder<KEY_TYPE,REC_TYPE>::Add_Local_Record (int position, REC_TYPE *record_to_insert)
{ _SA_(0) 
  m_key_idx.Add_Local_Record  (position, record_to_insert);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
}

template <class KEY_TYPE, class REC_TYPE>
void CMapKeyOrder<KEY_TYPE, REC_TYPE>::Add_Record(int position, REC_TYPE *record_to_insert, int put_environment)
{ _SA_(0) 
  m_key_idx.Add_Record        (position, record_to_insert, put_environment);
  m_key1_mkt.Add_Local_Record (position, record_to_insert);
}

template <class KEY_TYPE, class REC_TYPE>
void  CMapKeyOrder<KEY_TYPE,REC_TYPE>::Begin_Update_Keys  (REC_TYPE *old_rec)
{ _SA_(0) 
  m_key1_mkt.Begin_Update_Keys   (old_rec);
}

template <class KEY_TYPE, class REC_TYPE>
void  CMapKeyOrder<KEY_TYPE,REC_TYPE>::End_Update_Keys    (int update_pos, REC_TYPE *new_rec)
{ _SA_(0) 
  m_key1_mkt.End_Update_Keys   (update_pos, new_rec);
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Search_Idx_Record (REC_TYPE *record_to_search)
{ _SA_(0) 
  return m_key_idx.Search_Idx_Record (record_to_search);
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Search_By_Idx  (unsigned int idx)
{ _SA_(0) 
  return m_key_idx.Search_By_Idx  (idx);
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Search_Key_Record (CMessage *record_to_search, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (record_to_search);
  }
  return -1;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Search_Key_Record (void *key, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Search_Key_Record (key);
  }
  return -1;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Delete_By_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int delete_pos;

  delete_pos = m_key_idx.Delete_By_Idx_Record (record_to_delete);
  m_key1_mkt.Delete_Record (record_to_delete);

  return delete_pos;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Delete_But_Not_Idx_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  m_key1_mkt.Delete_Record (record_to_delete);

  return 0;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Find_Key_Range_It (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Range_It (key1, key2, mkit, key_n);
  }

  return -1;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Next_in_Key_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Next_in_Key_Range_It (mkit);
  }
  return -1;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Find_Key_Last_Range_It (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Find_Key_Last_Range_It (key1, key2, mkit, key_n);
  }

  return -1;
}

template <class KEY_TYPE, class REC_TYPE>
int CMapKeyOrder<KEY_TYPE,REC_TYPE>::Prev_Key_Last_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  switch (key_n)
  {
  case 1:
    return m_key1_mkt.Prev_Key_Last_Range_It (mkit);
  }
  return -1;
}

#ifndef DAPI
///////////////////////////////////////////////////////////////////////////////
// CMapService
///////////////////////////////////////////////////////////////////////////////
class CMapService
{
  Map_Service_Table     m_service_table;
  Map_Pos_Table         m_pos_table;

#ifdef DMEAPP
  int                   m_trans_counter;
  CVector <CServiceKey> m_trans_CServiceKey;
  CVector <CPosKey>     m_trans_CPosKey;
  CVector <int>         m_trans_op_CVector;
  CVector <int>         m_trans_pos_CVector;  // positions for every transaction start point in previous vectors
#endif

  Map_Service_Table::iterator m_service_key_it_first;
  Map_Service_Table::iterator m_service_key_it_last;
  Map_Service_Table::iterator m_service_key_it_end;

public:
  CMapService();
  ~CMapService();
  void          Initialize              (void);

#ifdef DMEAPP
  void          Begin_Trans             (void);
  void          Commit_Trans            (void);
  void          Cancel_Trans            (void);
#endif

  void          Add_Service_Record      (CServiceKey *service_key, CPosKey *pos_key);
  CServiceKey * Find_Service_Record     (CPosKey *pos_key);
  void          Update_Service_Record   (CPosKey *pos_key, CServiceKey *service_key);
#define SRVF_IS_DELETING_RECORD    0
#define SRVF_IS_UPDATING_RECORD    1
  void          Delete_Service_Record   (DWORD64 pos, int is_updating_flag=SRVF_IS_DELETING_RECORD);
  int           Delete_Service_Node     (CPosKey *pos_key);

  DWORD64       Find_ServiceKey_Range   (CServiceKeyIterator *cskit, CServiceKey *service_key_low, CServiceKey *service_key_high);
  DWORD64       Next_In_ServiceKey_Range(CServiceKeyIterator *cskit);
  DWORD64       Find_High_ServiceKey    (CServiceKey *service_key);
  DWORD64       Find_Low_ServiceKey     (CServiceKey *service_key, const CServiceKey **found_service_key);
};
#endif

// indicates the last operation made in a table
#define OPER_UNDEF      0
#define OPER_INSERT     1
#define OPER_DELETE     2
#define OPER_UPDATE     3

/////////////////
// CTableBase  //
/////////////////

typedef void  (*table_action_func)(void *, const char *);

class CTableBase
{
protected:
  int                     m_use_crc32_computing;
  int                     m_crc32_records_marked; // usado para saber si hubo registros marcados y proceder al final del ajuste con crc32
  CTableBase()          {m_use_crc32_computing = false;m_crc32_records_marked = false;}
  virtual ~CTableBase() {}
public:
  virtual int           Init_Transaction_In_Element   (int pos_in_table, char action, char *element_data) {return 0;}
  virtual int           Cancel_Transaction_In_Element (int pos_in_table, char action, const char *element_data, const char *content_to_use) {return 0;}
  virtual int           Commit_Transaction_In_Element (int pos_in_table, char action, const char *element_data) {return 0;}
  virtual int           Used_Num_Records  (void) {return 0;}
#if DMEAPP || GDIAPP || DAPI
  virtual unsigned int Get_Crc32         (void) {return 0;}
#endif
  virtual void          Initialize_Missing_Crc32s (void) {}
  virtual void          Set_Idx_For_Record_With_Env_Crc32 ( const struct IDX_ENV_CRC32_PAIR * cur_pair,
                                                            DLONG env_crc32, 
                                                            unsigned int idx) {}
  virtual CMessage    * Get_Record_By_Env_Crc32 (DLONG env_crc32) { return NULL; }

#define IS_NOT_LAST_REQUIREMENT false
#define IS_LAST_REQUIREMENT     true
  virtual void          Check_Crc32_Updating ( void    (*call_back_del_ptr) (unsigned int idx),
                                               void    (*call_back_fun_ptr) (int is_last_requirement, int num_of_records, DLONG *env_crc32_ptr)) {}
  virtual void          Finalize_Crc32_Updating ( void ) {}
  virtual int Do_We_Maintain_It     (void)
                                    {
                                      return false;
                                    }
};

#define Find_Range(a,b,c,x)    Find_Range_It(b, c, x, a)
#define Next_Range(x)          Next_Range_It(x, (x)->m_key_n)

#define Find_Last_Range(a,b,c,x)    Find_Last_Range_It (b, c, x, a)
#define Prev_Range(x)               Prev_Range_It      (x, (x)->m_key_n)

////////////
// CTable //
////////////

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
class CTable : public CTableBase
{
// Define here the function pointer because its argument depends on one template argument.
typedef int  (*table_process_func)(REC_TYPE *);

  int                   m_table_num;
  int                   m_table_flag;

#define CONFIRM_NOT_PENDING                 0
#define CONFIRM_PENDING_TO_PUBLISH          1
// flags for records
#define CONFIRM_INSERTED_FLAG               1
#define CONFIRM_UPDATED_FLAG                2

  int                   m_confirm_status; // when insergin do as local until confirmed (for cobranzas)
  Env_Crc32_List      * m_missing_Env_Crc32_List;
  Idx_List            * m_missing_Idx_List;
  // methods
  CVectorAt<REC_TYPE>   m_recs;
  CVector  <TLIST_TYPE> m_tlists;
  TLIST_TYPE          * m_tn_list; // default table list for all records, when defined

  MAP_KEY_TYPE          m_key_maps;
  int                   m_service_support;
#ifndef DAPI
  CMapService          *m_map_service;
#endif
  CPersFile            *m_Pers_File;
  char                 *m_Pers_File_last_write_date;

  table_action_func     m_action_func_ptr;

  // find the next available index for the table

public:
  int                   m_last_operation;
// usados para el protocolo de ajuste de tabla con hash y crc32
#define START_PROTOCOL_CORRECTION_NOT_ACTIVATED 0
#define START_PROTOCOL_CORRECTION_ACTIVATED     1
  int                   m_start_protocol_correction_status;

              CTable        (int max_records = INITIAL_TABLE_RECORDS, char *type = NULL);
             ~CTable        ();
  int         Total_Num_Records (void);
  int         Used_Num_Records  (void);
  int         Get_Record_Element_Size   (void);
  int         Get_Num_Records           (void);
#if DMEAPP || GDIAPP || DAPI
  unsigned int Get_Crc32         (void);
#endif
  virtual CMessage *
              Get_Record_By_Env_Crc32   (DLONG env_crc32);
  void        Initialize_Missing_Crc32s (void);
  void        Set_Idx_For_Record_With_Env_Crc32 ( const struct IDX_ENV_CRC32_PAIR * cur_pair, 
                                                  DLONG env_crc32, 
                                                  unsigned int idx);
  void        Check_Crc32_Updating ( void    (*call_back_del_ptr) (unsigned int idx),
                                     void    (*call_back_fun_ptr) (int is_last_requirement, int num_of_records, DLONG *env_crc32_ptr));
  void        Finalize_Crc32_Updating ( void );
  void        Send_All_Records           (int destination);
  void        Send_Record_Messages       (REC_TYPE *record,
                                          TLIST_TYPE *list_table,
                                          const char *table_action,
                                          int         restr_action_required,
                                          int         destination);
  REC_TYPE   *Insert_Local_Record (REC_TYPE *record_to_insert,   TLIST_TYPE *list_table);
  REC_TYPE   *Insert_Record       (const char *record_to_insert, TLIST_TYPE *list = NULL);
             
  REC_TYPE   *Update_Local_Record (const char *record_to_update, CMessage *cur_msg);
  REC_TYPE   *Update_Local_Record (REC_TYPE *record_to_update);
  REC_TYPE   *Update_Record       (const char *record_to_update, CMessage *cur_msg);
  REC_TYPE   *Update_Record       (REC_TYPE *record_to_update);

  void        Delete_All_Records  (void);
  REC_TYPE   *Delete_Local_Record (REC_TYPE *record_to_delete);
  TLIST_TYPE *Delete_Record       (CMessage *cur_msg);
  TLIST_TYPE *Delete_Record       (REC_TYPE *record_to_delete);
  REC_TYPE   *Confirm_Record      (REC_TYPE *record_to_confirm);

  REC_TYPE   *Find_Idx_Key_Record (REC_TYPE *record_to_find);
  REC_TYPE   *Find_By_Idx         (unsigned int idx);

  REC_TYPE   *Find_NKey_Record    (CMessage *cur_msg, int key_n=1);
  REC_TYPE   *Find_NKey_Record    (void *key1, int key_n=1);

  REC_TYPE   *Get_Rec             (int pos);

  REC_TYPE   *Find_Range_It       (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n);
  REC_TYPE   *Next_Range_It       (CMapKeyN *mkit, int key_n);

  REC_TYPE   *Find_Last_Range_It  (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n);
  REC_TYPE   *Prev_Range_It       (CMapKeyN *mkit, int key_n);

  void          Begin_Service_Trans         (void);
  void          Commit_Service_Trans        (void);
  void          Cancel_Service_Trans        (void);

  void          Insert_Service              (CServiceKey *service_key, CPosKey *pos_key);
  CServiceKey * Find_Service                (CPosKey *pos_key);
  int           Delete_Service              (CPosKey *pos_key);
  void          Update_Service              (CPosKey *pos_key, CServiceKey *service_key);

  REC_TYPE    * Find_ServiceKey_Range       (CServiceKeyIterator *cskit, CServiceKey *service_key_low, CServiceKey *service_key_high);
  REC_TYPE    * Next_In_ServiceKey_Range    (CServiceKeyIterator *cskit);
  REC_TYPE    * Find_High_ServiceKey_Record (CServiceKey *service_key);
  REC_TYPE    * Find_Low_ServiceKey_Record  (CServiceKey *service_key, const CServiceKey **found_service_key=NULL);
    
  int         Get_Table_Flag              (void)  {return m_table_flag;}
  int         Set_Table_Flag              (int table_flag);
  int         Confirm_Later               (int confirm_status);
  void        Set_Table_Num               (int table_num);
  int         Set_Table_Type              (int table_type_id);  // Using the table type, the flag and num can be set.

  void        Set_Service_Support_Flag    (int flag_value);
#define DO_NOT_CHECK_IDX  0
#define CHECK_IDX         1
  void        Check_Delete_Existing_Records   (REC_TYPE   *record_to_check, 
                                               int         check_idx,
                                               int         is_local_action);

#define IS_GLOBAL_ACTION  0
#define IS_LOCAL_ACTION   1
#define IS_DS_CORRECTION  2
#define SAVE_IF_PERSISTENT          0
#define DO_NOT_SAVE_IF_PERSISTENT   1
  void        Check_Record_Actions            (int         record_pos,
                                               TLIST_TYPE *list_table,
                                               const char *table_action,
                                               int         restr_action_required,
                                               int         is_local_action,
                                               int         save_persistent_flag=SAVE_IF_PERSISTENT);

  int         Search_Next_Free_Pos  (void);
  void        Send_Fake_Record_Change  (REC_TYPE *record, const char *table_action, int restr_action_required);
  virtual int Do_We_Maintain_It     (void)
                                    {
                                      return m_table_flag & TABLE_PUBLISHED_TO_DS;
                                    }

#if DMEAPP || GDIAPP
  void        Set_Action_Function_Pointer  (table_action_func  action_func_ptr);
#endif

#if DMEAPP || DAPI
  int         Assoc_To_Pers_File    (unsigned int file_code,
                                     const char   *file_name,
                                     unsigned int rec_size,
                                     int           keep_last_write_date);
  void        Enable_Pers_File_Last_Write_Date    (int enable);                                     
  void        Load_Pers_File        (void);
#endif

#ifdef DMEAPP
  
  unsigned int Get_Max_Rec_Number  (void);
  void        Check_Set_Pers_File_Last_Write_Date (void);
  int         Get_Pers_File_Last_Write_Date       (char *last_write_date);

  unsigned int Safe_Process_All_Records (table_process_func process_func_ptr);

  int           Init_Transaction_In_Element   (int pos_in_table, char action, char *element_data);
  int           Cancel_Transaction_In_Element (int pos_in_table, char action, const char *element_data, const char *content_to_use);
  int           Commit_Transaction_In_Element (int pos_in_table, char action, const char *element_data);
  REC_TYPE     *Insert_Transactional_Record   (int pos_in_table, const char *content_to_use, int last_free_to_use);
  REC_TYPE     *Update_Transactional_Record   (int pos_in_table, const char *content_to_use);
  REC_TYPE     *Delete_Transactional_Record   (int pos_in_table);

#endif
};

/////////////
// CVector //
/////////////

template <class T>
CVector<T>::CVector (int max)
{ _SA_(0) 
  T_ptr      = NULL;
  max_number = 0;
  Set_New_Limit (max);
}

template <class T>
CVector<T>:: ~CVector (void)
{ _SA_(0) 
  int i;

  for (i=0; i< max_number; ++i)
    delete T_ptr[i];

  free (T_ptr);
}
  
template <class T>
void CVector<T>::Set_New_Limit (int new_max_number)
{ _SA_(0) 
  int i;

  // check if the new limit is under the maximum value
  if (new_max_number <= max_number)
    return;

  T_ptr = (T **) realloc (T_ptr, new_max_number * sizeof(T *));
  for (i=max_number; i< new_max_number; ++i)
    T_ptr[i] = (T *) new T();

  max_number = new_max_number;
}

////////////
// CRules //
////////////

template <class KEY_TYPE, class TLIST_TYPE >
CTRules<KEY_TYPE,TLIST_TYPE>::CTRules    (int max_rules) :
    m_key_patterns (max_rules), m_num_tables (max_rules)
{ _SA_(0) 
  m_max_rules = max_rules;
  m_num_rules = 0;
}

template <class KEY_TYPE, class TLIST_TYPE >
CTRules<KEY_TYPE,TLIST_TYPE>::~CTRules    (void)
{

}

template <class KEY_TYPE, class TLIST_TYPE >
void CTRules<KEY_TYPE,TLIST_TYPE>::Add_Rule   (KEY_TYPE *key, int table)
{ _SA_(0) 
  if (m_num_rules + 1 > m_max_rules)
  {
    m_max_rules *= 2;
    m_key_patterns.Set_New_Limit (m_max_rules);
    m_num_tables.Set_New_Limit   (m_max_rules);
  }
  m_key_patterns [m_num_rules] = *key;
  m_num_tables   [m_num_rules] = table;
  ++m_num_rules;
}

template <class KEY_TYPE, class TLIST_TYPE >
void CTRules<KEY_TYPE,TLIST_TYPE>::Init_TList (KEY_TYPE *key, TLIST_TYPE *table_list, int fit_first_only)
{ _SA_(0) 
  int i;

  // initializes the table_list
  table_list->Clean_Tables ();

  // check for all the patterns
  for (i= 0; i< m_num_rules; ++i)
    if (key->Fit_Pattern (&m_key_patterns[i]))
    {
      // fits well the pattern
      table_list->Add_Table (m_num_tables[i]);
      if (fit_first_only)
        return;
    }
}

/////////////
// CTable  //
/////////////

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::CTable        (int max_records, char *type) :
    m_recs (max_records, type), 
    m_tlists(max_records)
{ _SA_(0) 
  m_table_num           = ++Glb_table_counter;
  // By default do not redirect messages when they are received. (i.e. it applies for DGI tables).
  m_table_flag          = TABLE_NOT_PUBLISHED;
  m_service_support     = 0;
  m_service_support     = 0;
  m_Pers_File           = NULL;
  m_Pers_File_last_write_date = NULL;
  m_action_func_ptr     = NULL;
  m_tn_list             = NULL;
  m_missing_Env_Crc32_List = NULL;
  m_missing_Idx_List       = NULL;
  m_confirm_status         = CONFIRM_NOT_PENDING;
  m_key_maps.Initialize();
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::~CTable  (void)
{ _SA_(0) 
  if (m_missing_Env_Crc32_List)
    delete m_missing_Env_Crc32_List;

  if (m_missing_Idx_List)
    delete m_missing_Idx_List;

#ifndef DAPI
  if (m_service_support)
    delete m_map_service;
#endif

  if (m_Pers_File)
  {
#ifdef DMEAPP
    // Disable Persistent file last write date.
    Enable_Pers_File_Last_Write_Date (FALSE);
#endif

    m_Pers_File->Close();
    delete m_Pers_File;
    m_Pers_File = NULL;
  }

  if (m_tn_list)
  {
    delete m_tn_list;
    m_tn_list = NULL;
  }
}

// sends messages following the given record 
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Send_Fake_Record_Change  ( REC_TYPE   *cur_rec, 
                                                                          const char *table_action,
                                                                          int         restr_action_required)
{ _SA_(0) 
#ifndef DAPI
  int         rec_pos;

  // find the position of the record
  rec_pos = m_key_maps.Search_Idx_Record (cur_rec);
  if (rec_pos < 0)
    return;

  // Just send the message because. Don't check the actions because it is a fake action.
  Send_Record_Messages (&m_recs[rec_pos], &m_tlists [rec_pos], table_action, restr_action_required, NORMAL_DEST);
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Send_Record_Messages(REC_TYPE *record,
                                                                    TLIST_TYPE *list_table,
                                                                    const char *table_action,
                                                                    int         restr_action_required,
                                                                    int         destination)
{ _SA_(0) 
  CDestinationsList *destinations_list_ptr;
  int table_num;
  int prev_value;
  int rec_len;

  // Get the target DS list.
  COM_Get_Target_DS_List (destination, &destinations_list_ptr);

  // Set the table action.
  (*record)[I_OP] = table_action;

  // when deleting only the header size of the record is used
  rec_len = record->m_data_len;
  if (*table_action == DATA_DELETE_CHAR)
    record->m_data_len = GTABLE_MSG_HEADER_SIZE;

  if (list_table->m_num_tables)
  {
    prev_value = (*record)[DATA_CODE];
    for (table_num = list_table->First_Table(); table_num ; table_num = list_table->Next_Table())
    {
      (*record)[DATA_CODE] = table_num;      
#ifdef DMEAPP
      if (destination != DGI_DEST)
        RESTR_Send   (record, table_num, restr_action_required, destination);
      else
#endif
        COM_Send_Pkt (record->m_data, 
                      record->m_data_len, 
                      destinations_list_ptr,
                      PRD_TABLAS,
                      TABLE_SENT_RECORD_LOG_TEXT);
    }
    (*record)[DATA_CODE] = prev_value;
  }
  else
  {
    // Send only to dgi if this table is not normally updated
    // by the dme.
    COM_Send_Pkt (record->m_data, 
                  record->m_data_len, 
                  destinations_list_ptr,
                  PRD_TABLAS,
                  TABLE_SENT_RECORD_LOG_TEXT);
  }

  // restore the length of the message
  if (*table_action == DATA_DELETE_CHAR)
    record->m_data_len = rec_len;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Send_All_Records (int destination)
{ _SA_(0) 
  REC_TYPE     *cur_rec;
  TLIST_TYPE   *list_table;
  int           i, limit;

#ifndef DAPI
  if (m_table_flag & TABLE_NOT_PUBLISHED)
    return;

  limit      = m_recs.max_number;
  cur_rec    = NULL;
  list_table = NULL;
  for (i = 0; i < limit; ++i)
  {
    // If the record is valid, send it.
    if (m_recs.Is_Used(i))
    {
      cur_rec    = &m_recs[i];
      list_table = &m_tlists[i];

#if defined (DEBUG) || defined (_DEBUG)
      if ((*cur_rec)[MSG_NUMBER_FIELD] != MSS_TABLA_GLOBAL)
      {
        COMMON_Put_Syslog_Message(PRIOR_LOG_MSG, 
                                  "Warning!!! Messages with type %d (Not global table records) are being sent in a table.!!!",
                                  (int)(*cur_rec)[MSG_NUMBER_FIELD]);
      }
#endif
      Send_Record_Messages(cur_rec, list_table, DATA_INSERT, RESTR_SKIP, destination); // send only to "destination".
    }
  }

  // Send the last record twice. The second time as an update.
  // Do this to emulate the DS behaviour when sending global table records.
  if (cur_rec)
    Send_Record_Messages(cur_rec, list_table, DATA_UPDATE_COPY, RESTR_SKIP, destination);

#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Used_Num_Records(void)
{ _SA_(THA_THREAD_SAFE) 
  return m_recs.cur_used;
}

#pragma pack (1)
struct IDX_ENV_CRC32_PAIR
{
  unsigned int idx;
  unsigned int crc32;
  unsigned int env;
};
#pragma pack ()

#if DMEAPP || GDIAPP || DAPI
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
unsigned int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Crc32(void)
{ _SA_(0) 
  unsigned int           table_crc32 = 0xFFFFFFFF;
  Map_Idx_Table::iterator idx_it;
  struct IDX_ENV_CRC32_PAIR
                          idx_env_crc32_pair;
  int                     position;
  REC_TYPE              * cur_rec;
  DLONG                   env_crc32;
  unsigned int           tmp_val;

  m_crc32_records_marked = false;

  // go for all the records
  for (idx_it = m_key_maps.m_key_idx.m_idx_table.begin(); idx_it != m_key_maps.m_key_idx.m_idx_table.end(); ++idx_it)
  {
    idx_env_crc32_pair.idx = To_IntLittle(idx_it->first.rec_idx);
    position            = idx_it->second;
    cur_rec             = &m_recs[position];

    // Here, get the internal value for crc32 and environment because Dataserver
    // also uses the internal value for this calculation (little endian).
    env_crc32 = cur_rec->m_env_crc32;
    tmp_val = (unsigned int) (env_crc32 >> 32);
    idx_env_crc32_pair.env   = To_IntLittle(tmp_val);
    tmp_val = (unsigned int) env_crc32;
    idx_env_crc32_pair.crc32 = To_IntLittle(tmp_val);

    // add the idx and crc to the computing value
    table_crc32 = COMMON_Get_Crc32 ((const char *) &idx_env_crc32_pair, sizeof (idx_env_crc32_pair), table_crc32);
  }
  return table_crc32;
}
#endif

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
CMessage *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Record_By_Env_Crc32 (DLONG env_crc32)
{ _SA_(0) 
  int    num_keys, 
         key = SECOND_KEY;
  extern CEnvCrc32Key Glb_CEnvCrc32Key;

  Glb_CEnvCrc32Key.Init (env_crc32);

  num_keys = m_key_maps.Get_Num_Keys();
       
  if (num_keys > 2)
    key = num_keys; 

  return Find_NKey_Record (&Glb_CEnvCrc32Key, key);
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Initialize_Missing_Crc32s (void)
{ _SA_(0) 
  Map_Idx_Table::iterator idx_it;
  int                     position;
  REC_TYPE              * cur_rec;

  m_use_crc32_computing = true;
  // clear all mising records
  if (!m_missing_Env_Crc32_List)
    m_missing_Env_Crc32_List = new Env_Crc32_List;
  else
    m_missing_Env_Crc32_List->clear ();

  // clear all mising records by idx
  if (!m_missing_Idx_List)
    m_missing_Idx_List = new Idx_List;
  else
    m_missing_Idx_List->clear ();

  // uncheck all records
  // go for all the records
  for (idx_it = m_key_maps.m_key_idx.m_idx_table.begin(); idx_it != m_key_maps.m_key_idx.m_idx_table.end(); ++idx_it)
  {
    position              = idx_it->second;
    cur_rec               = &m_recs[position];
    cur_rec->m_is_checked = false;
  }
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Idx_For_Record_With_Env_Crc32 (const struct IDX_ENV_CRC32_PAIR * cur_pair,
                                                                                  DLONG env_crc32, 
                                                                                  unsigned int idx)
{ _SA_(0) 
  CMessage *cur_rec;
  DLONG     env_crc32_lit_endian;

  // Search the record. The received crc32 is stored in Little Endian Order
  // so in order to search it is necessary to invert the byte order 
  cur_rec = Get_Record_By_Env_Crc32 (env_crc32);

  // When found, signal this record, and update the index value
  if (cur_rec)
  {
    cur_rec ->m_is_checked   = true;
    // The updated idx should be stored in Little Endian Order
    cur_rec ->m_assigned_idx = idx;
  }
  else
  {
    if (m_table_flag & TABLE_NOT_PUBLISHED)
    {
      // we do not maintain, save crc32
      // to ask again for those we don't have, we'll need the crc32
      env_crc32_lit_endian = Take_DLong (&cur_pair ->crc32);
      m_missing_Env_Crc32_List->push_back (env_crc32_lit_endian);
    }
    else
    {
      // add to a list of records we do not have
      // to erase, we will need the idx
      m_missing_Idx_List->push_back (idx);
    }
  }
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Check_Crc32_Updating ( void    (*call_back_del_ptr) (unsigned int idx),
                                                                      void    (*call_back_fun_ptr) (int      is_last_requirement, 
                                                                                                    int      num_of_records, 
                                                                                                    DLONG   *env_crc32_ptr))
{ _SA_(0) 
  int                   position;
  Env_Crc32_List::iterator  Env_Crc32_List_it;
  Idx_List::iterator        Idx_List_it;
#define MAX_CRC32_NODES   320
  DLONG                 env_crc32_tab[MAX_CRC32_NODES]; // ??? Max nodes is the same if every item is DLONG ???
  int                   counter, limit;
  REC_TYPE            * cur_rec;
  int                   we_maintain_it;
  int                   ds_dme_maintained;

  // NOTE: whe both ds-dme published, at first it is processed as not maintained
  // for to have the records and do process as required
  we_maintain_it = ! (m_table_flag & TABLE_NOT_PUBLISHED);

  ds_dme_maintained = ((m_table_flag & TABLE_PUBLISHED_TO_DS_AND_FROM_DS) == TABLE_PUBLISHED_TO_DS_AND_FROM_DS);

  // indicar que hubo marcado de registros para publicarlos al final del protocolo de ajuste
  // si llega directamente fin ('z') el ds tiene todos los registros, no hay nada marcado
  // y no hay nada que publicar
  m_crc32_records_marked = true;

  // erase all idx table
  m_key_maps.m_key_idx.Initialize();

  // 1.- CHECKED RECORDS, we have those and they too have those
  // 1.1 go for all records
  limit             = m_recs.max_number;

  for (position = 0; position < limit; ++position)
  {
    // take only used records.
    if (!m_recs.Is_Used(position))
      continue;

    cur_rec = &m_recs[position];

    // 1.1.1- if its checked update idx
    if (cur_rec->m_is_checked)
    {
      // map the idx to the current record
      m_key_maps.m_key_idx.Add_Idx (position, cur_rec->m_assigned_idx, cur_rec);
    }
  }

  // 2.- MISSING RECORDS, we do not have
  // 2.1.- if we maintain
  if (we_maintain_it)
  {
    // send a final ask for no records
    (*call_back_fun_ptr) (IS_LAST_REQUIREMENT, 0, env_crc32_tab);

    // 2.2.- if we do maintain
    // go for all missing records
    for ( Idx_List_it = m_missing_Idx_List ->begin(); 
          Idx_List_it != m_missing_Idx_List ->end(); 
          ++Idx_List_it)
    {
      // send a delete message
      call_back_del_ptr (*Idx_List_it);
    }
  }
  else
  {
    // 2.2.- if we do not maintain
    counter = 0;
    // go for all missing records
    for ( Env_Crc32_List_it = m_missing_Env_Crc32_List ->begin(); 
          Env_Crc32_List_it != m_missing_Env_Crc32_List ->end(); 
          ++Env_Crc32_List_it)
    {
      // Here, get the internal value (little endian) for crc32 and environment
      // because this array will be send in a message to the dataserver.
      env_crc32_tab[counter] = *Env_Crc32_List_it;
      ++counter;

      if (counter >= MAX_CRC32_NODES)
      {
        // send an ask for all missing records
        (*call_back_fun_ptr)  (IS_NOT_LAST_REQUIREMENT, counter, env_crc32_tab);
        counter = 0;
      }
    }
    // final ask for all missing records
    (*call_back_fun_ptr) (IS_LAST_REQUIREMENT, counter, env_crc32_tab);
  }
  // clear missing list
  m_missing_Env_Crc32_List->clear ();
  m_missing_Idx_List->clear ();

  // 3.- extra records, they do not have
  //
  for (position = 0; position < limit; ++position)
  {
    // take only used records.
    if (!m_recs.Is_Used(position))
      continue;

    cur_rec = &m_recs[position];

    // 1.1.2- if it's not checked
    if (!cur_rec->m_is_checked)
    {
      // if maintained for both DME and DS, ask the record for who is maintaining it
      if (ds_dme_maintained)
        we_maintain_it = cur_rec ->Do_We_Maintain_It ();

      if (we_maintain_it)
      {
        // if we maintain it
        // take a new idx, and insert it in the idx table
        m_key_maps.m_key_idx.Add_Record (position, cur_rec);
      }
      else
      {
        // if we do not maintain
        // delete this record, 
        m_key_maps.Delete_But_Not_Idx_Record (cur_rec);

        // set the record as available
        m_recs.Release (position);

        // send only to dgi.
        Check_Record_Actions (position, &m_tlists[position], DATA_DELETE, RESTR_SKIP, IS_LOCAL_ACTION);
      }
    }
  }

}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Finalize_Crc32_Updating ( void )
{ _SA_(0) 
  int                       position;
  int                       limit;
  REC_TYPE                * cur_rec;
  int                       we_maintain_it;
  int                       ds_dme_maintained;

  // NOTE: whe both ds-dme published, at first it is processed as not maintained
  // for to have the records and do process as required
  we_maintain_it    = ! (m_table_flag & TABLE_NOT_PUBLISHED);

  ds_dme_maintained = ((m_table_flag & TABLE_PUBLISHED_TO_DS_AND_FROM_DS) == TABLE_PUBLISHED_TO_DS_AND_FROM_DS);

  if (!we_maintain_it && !ds_dme_maintained)
    return;

  // si la tabla nunca fue ajustada con hash, no hay registros marcados.. todo estuvo correcto
  if (!m_crc32_records_marked)
    return;

  // 1.- CHECKED RECORDS, we have those and they too have those
  // 1.1 go for all records
  limit             = m_recs.max_number;

  // 3.- extra records, they do not have
  //
  for (position = 0; position < limit; ++position)
  {
    // take only used records.
    if (!m_recs.Is_Used(position))
      continue;

    cur_rec = &m_recs[position];

    // 1.1.2- if it's not checked
    if (!cur_rec->m_is_checked)
    {
      // if maintained for both DME and DS, ask the record for who is maintaining it
      if (ds_dme_maintained)
        we_maintain_it = cur_rec ->Do_We_Maintain_It ();

      if (we_maintain_it)
      {
        // publish this record
        Check_Record_Actions (position, &m_tlists[position], DATA_INSERT, RESTR_SKIP, IS_DS_CORRECTION);
      }
    }
  }
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Total_Num_Records(void)
{ _SA_(0) 
  return m_recs.max_number;
}

// Funcion que devuelve el numero total de registros existentes en la tabla.
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Num_Records (void)
{ _SA_(0) 
  return m_key_maps.Get_Num_Records ();
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Record_Element_Size(void)
{ _SA_(0) 
  return m_recs.Get_Element_Size();
}

// this method returns the used-generated idx
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Insert_Local_Record (REC_TYPE *record_to_insert, TLIST_TYPE *list_table)
{ _SA_(0) 
  int             next_p;
  REC_TYPE       *cur_rec;
  unsigned int   u_n;
  unsigned int   env;

  // search for the next available position
  next_p = Search_Next_Free_Pos ();
  cur_rec = &m_recs [next_p];

  cur_rec->Clean_Data_Buffer();
  
  // check if it's our environment, to maintain the crc32
  if (m_use_crc32_computing)
  {
    env = (unsigned int)(*record_to_insert)[ENVIRONMENT];

    cur_rec ->m_env_crc32_to_use  = ((DLONG)env << 32) + (unsigned int)(*record_to_insert)[CRC32];
  }

  // copy the record, and init it implicitly
  *cur_rec = *record_to_insert;

  // if the table has persistent record, the unic number is not restrained, and
  // the unic number is the same number as the record
  if (m_Pers_File)
  {
    u_n = m_Pers_File->Get_Next_Free();
  }
  else
  {
    u_n = ++Glb_Unic_Number;
  }
  
  cur_rec->Take_Unic_Number (u_n);

  // Some tables (DAPI tables) make the records of the table available to the interface using its own 
  // unique number, so the incoming REC_IDX field is overwritten.
  // Tables received from DS to be just mantained, don't overwrite the REC_IDX field in order
  // that further update/delete actions can be made.
  if (m_table_flag & TABLE_USE_OWN_UNIQUE_IDX)
    (*cur_rec)[REC_IDX] = Glb_Unic_Number;
  else
  {
    // Check for existing records with the same key or keys. If so, delete the existing records.
    Check_Delete_Existing_Records (cur_rec, CHECK_IDX, IS_LOCAL_ACTION);
  }

  // copy the table list
  m_tlists (next_p) = *list_table;

  m_key_maps.Add_Local_Record (next_p, cur_rec);

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_INSERT_CHAR);
#endif

  // send only to dgi.
  Check_Record_Actions (next_p, &m_tlists[next_p], DATA_INSERT, RESTR_SKIP, IS_LOCAL_ACTION);

  // return the used idx
  return cur_rec;
}

// this method returns the used-generated idx
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Insert_Record (const char *record_to_insert, TLIST_TYPE *list)
{ _SA_(0) 
  int             next_p;
  unsigned int   u_n;

  // search for the next available position
  next_p  = Search_Next_Free_Pos ();
  REC_TYPE *cur_rec = &m_recs [next_p];

  // function to set default data info
  cur_rec->Clean_Data_Buffer();

  // copy partial data to the record
  (*cur_rec)[DATA_INFO] = (char *) record_to_insert;

#ifndef DAPI
  // if the table has persistent record, the unic number is not restrained, and
  // the unic number is the same number as the record
  if (m_Pers_File)
  {
    u_n = m_Pers_File->Get_Next_Free();
  }
  else
#endif
  {
    u_n = ++Glb_Unic_Number;
  }

  cur_rec->Take_Unic_Number (u_n);

  // If the record uses crc32, then its crc32 and environment will be set in this
  // initialization.
  cur_rec->Init();

  // Check for existing records with the same key or keys. If so, delete the existing records.
  // Don't check the idx because it is a record mantained by the application and the idx is set
  // later in the mapping Add_Record() method.
  Check_Delete_Existing_Records (cur_rec, DO_NOT_CHECK_IDX, IS_GLOBAL_ACTION);

#if DMEAPP || GDIAPP
  // the list can be taken from the record
  if (!list)
  {
    if (m_tn_list)
      list = m_tn_list;
    else
      list = (TLIST_TYPE *)cur_rec->Get_TableList ();
  }
  // copy the table list
  m_tlists (next_p) = *list;
#endif

  m_key_maps.Add_Record (next_p, cur_rec);

#ifndef DAPI
  if (m_service_support)
    cur_rec->Add_Services_Map(m_map_service);
#endif

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_INSERT_CHAR);
#endif

  if (m_confirm_status == CONFIRM_PENDING_TO_PUBLISH)
  {
    // signal this record as pending to be published
    // sent function signal to do not publish
    cur_rec->m_confirm_status |= CONFIRM_INSERTED_FLAG;
  }

  Check_Record_Actions (next_p, list, DATA_INSERT, RESTR_INSERT_ACTION, IS_GLOBAL_ACTION);

  return cur_rec;
}

// the record was accepted, now a confirmation is set, to publish the record
// and to save as persistent
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Confirm_Record (REC_TYPE *record_to_confirm)
{ _SA_(0) 
  int         update_pos;
  REC_TYPE   *cur_rec;
  int         was_inserted;

  update_pos = m_key_maps.Search_Idx_Record (record_to_confirm);
  if (update_pos < 0)
    return NULL;

  cur_rec    = &m_recs [update_pos];

  // if the record is not pending, do not proceeed
  if (!cur_rec || !cur_rec->m_confirm_status)
    return cur_rec;

  was_inserted = cur_rec->m_confirm_status & CONFIRM_INSERTED_FLAG ? true : false;

  // restore record status and proceed to publish and save in persistent file
  cur_rec->m_confirm_status = CONFIRM_NOT_PENDING;

  Check_Record_Actions (  update_pos,     
                          &m_tlists [update_pos], 
                          was_inserted ? DATA_INSERT          : DATA_UPDATE, 
                          was_inserted ? RESTR_INSERT_ACTION  : RESTR_UPDATE_ACTION,
                          IS_GLOBAL_ACTION);

  return cur_rec;
}


template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_Idx_Key_Record (REC_TYPE *record_to_find)
{ _SA_(0) 
  int             pos;
  REC_TYPE       *cur_rec;

  // find the record
  pos = m_key_maps.Search_Idx_Record (record_to_find);

  if (pos < 0)
    return NULL;

  cur_rec = &m_recs [pos];
  return cur_rec;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_By_Idx (unsigned int idx)
{ _SA_(0) 
  int             pos;
  REC_TYPE       *cur_rec;

  // find the record
  pos = m_key_maps.Search_By_Idx (idx);

  if (pos < 0)
    return NULL;

  cur_rec = &m_recs [pos];
  return cur_rec;
}

// this method, updates the record in the table that fits with the CMessage specified
// cur_rec MUST provide a compatible key to search 
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Local_Record (const char *record_to_update, CMessage *cur_msg)
{ _SA_(0) 
  int         update_pos;
  REC_TYPE   *cur_rec;

  // find the record
  update_pos = m_key_maps.Search_Key_Record (cur_msg,1);
  if (update_pos < 0)
    return NULL;

  cur_rec    = &m_recs [update_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_UPDATE_CHAR);
#endif

  m_key_maps.Begin_Update_Keys (cur_rec);

  // function to set default data info
  // copy partial data to the record
  (*cur_rec)[DATA_INFO] = (char *)record_to_update;

  // initializes the record
  cur_rec->Init (); 

  // Check for existing records with the new keys in order to avoid
  // duplicated records after the update.
  Check_Delete_Existing_Records (cur_rec, CHECK_IDX, IS_LOCAL_ACTION);

  m_key_maps.End_Update_Keys   (update_pos, cur_rec);

#ifndef DAPI
  // when updating, update also the services.
  if (m_service_support)
  {
    m_map_service->Delete_Service_Record((DWORD64)cur_rec, SRVF_IS_UPDATING_RECORD);
    cur_rec->Add_Services_Map(m_map_service);
  }
#endif

  m_last_operation = OPER_UPDATE;
  return cur_rec;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Local_Record (REC_TYPE *record_to_update)
{ _SA_(0) 
  int             update_pos;
  REC_TYPE       *cur_rec;
  unsigned int   env;
  int             keys_may_change;  // if the record is the same, use this flag to verify that 
                                      // the keys are not changed

  // find the record
  update_pos = m_key_maps.Search_Idx_Record (record_to_update);
  if (update_pos < 0)
    return NULL;

  cur_rec    = &m_recs [update_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_UPDATE_CHAR);
#endif

  keys_may_change = record_to_update != cur_rec;

  if (keys_may_change)
    m_key_maps.Begin_Update_Keys (cur_rec);
  else
  {
    // verify that the keys are not changed
    if (m_key_maps.Search_Key_Record (cur_rec, 1) < 0)
      LOG_Put_Err_Msg("Error. Record updating has changed the idx=%d, Table=%d, Record=%.20s", (int)(*record_to_update)[REC_IDX], m_table_num, record_to_update->m_data);
  }

  // check if it's our environment, to maintain the crc32
  if (m_use_crc32_computing)
  {
    env = (*record_to_update)[ENVIRONMENT];        

    cur_rec ->m_env_crc32_to_use  = ((DLONG)env << 32) + (unsigned int)(*record_to_update)[CRC32];
  }
  
  cur_rec ->Update_Message_Info(record_to_update->m_data);

  if (keys_may_change)
  {
    // Check for existing records with the new keys in order to avoid
    // duplicated records after the update.
    Check_Delete_Existing_Records (cur_rec, CHECK_IDX, IS_LOCAL_ACTION);

    m_key_maps.End_Update_Keys    (update_pos, cur_rec);
  }

  // send only to dgi.
  Check_Record_Actions (update_pos, &m_tlists[update_pos], DATA_UPDATE, RESTR_SKIP, IS_LOCAL_ACTION);

  return cur_rec;
}

// this method, updates the record in the table that fits with the CMessage specified
// cur_rec MUST provide a compatible key to search 
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Record (const char *record_to_update, CMessage *cur_msg)
{ _SA_(0) 
  int         update_pos;
  REC_TYPE   *cur_rec;

  // find the record
  update_pos = m_key_maps.Search_Key_Record (cur_msg, 1);
  if (update_pos < 0)
    return NULL;

  cur_rec    = &m_recs [update_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_UPDATE_CHAR);
#endif

  m_key_maps.Begin_Update_Keys (cur_rec);

  // function to set default data info
  // copy partial data to the record
  (*cur_rec)[DATA_INFO] = (char *)record_to_update;

  // initializes the record
  cur_rec->Init (); 

  // Check for existing records with the new keys in order to avoid
  // duplicated records after the update.
  Check_Delete_Existing_Records (cur_rec, CHECK_IDX, IS_GLOBAL_ACTION);

  m_key_maps.End_Update_Keys   (update_pos, cur_rec);

#ifndef DAPI
  // when updating, update also the services.
  if (m_service_support)
  {
    m_map_service->Delete_Service_Record((DWORD64)cur_rec, SRVF_IS_UPDATING_RECORD);
    cur_rec->Add_Services_Map(m_map_service);
  }
#endif

  if (m_confirm_status == CONFIRM_PENDING_TO_PUBLISH)
  {
    // signal this record as pending to be published
    // sent function signal to do not publish
    cur_rec->m_confirm_status |= CONFIRM_UPDATED_FLAG;
  }

  Check_Record_Actions (update_pos, &m_tlists [update_pos], DATA_UPDATE, RESTR_UPDATE_ACTION, IS_GLOBAL_ACTION);

  return cur_rec;
}

// the record was changed and, the messages announcing this change is required.
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Record (REC_TYPE *record_to_update)
{ _SA_(0) 
  int         update_pos;
  REC_TYPE   *cur_rec;

  update_pos = m_key_maps.Search_Idx_Record (record_to_update);
  if (update_pos < 0)
    return NULL;

  cur_rec    = &m_recs [update_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_UPDATE_CHAR);
#endif

  // any change implies crc32 computing
  if (cur_rec != record_to_update || m_use_crc32_computing)
  {
    m_key_maps.Begin_Update_Keys (cur_rec);

    (*cur_rec)[DATA_INFO] = (*record_to_update)[DATA_INFO];

    // initializes the record
    cur_rec->Init (); 

    // Check for existing records with the new keys in order to avoid
    // duplicated records after the update.
    Check_Delete_Existing_Records (cur_rec, CHECK_IDX, IS_GLOBAL_ACTION);

    m_key_maps.End_Update_Keys   (update_pos, cur_rec);
  }

#ifndef DAPI
  // when updating, update also the services.
  if (m_service_support)
  {
    m_map_service->Delete_Service_Record((DWORD64)cur_rec, SRVF_IS_UPDATING_RECORD);
    cur_rec->Add_Services_Map(m_map_service);
  }
#endif

  if (m_confirm_status == CONFIRM_PENDING_TO_PUBLISH)
  {
    // signal this record as pending to be published
    // sent function signal to do not publish
    cur_rec->m_confirm_status |= CONFIRM_UPDATED_FLAG;
  }

  Check_Record_Actions (update_pos, &m_tlists [update_pos], DATA_UPDATE, RESTR_UPDATE_ACTION, IS_GLOBAL_ACTION);

  return cur_rec;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_All_Records (void)
{ _SA_(0) 
  int       delete_pos;
  REC_TYPE *cur_rec;

  // using the first map key
  for (delete_pos =   m_key_maps.m_key1_mkt.Begin();
       delete_pos != -1;
       delete_pos =   m_key_maps.m_key1_mkt.Next())
  {
    // find the record
    cur_rec = &m_recs [delete_pos];
    cur_rec ->Do_Clean();

    // set the record as available
    m_recs.Release (delete_pos);

    // send only to dgi.
    Check_Record_Actions (delete_pos, 
                          &m_tlists[delete_pos], 
                          DATA_DELETE, 
                          RESTR_SKIP, 
                          IS_LOCAL_ACTION, 
                          DO_NOT_SAVE_IF_PERSISTENT); // avoid actions over persistent files
  }

  m_last_operation = OPER_DELETE;
  m_key_maps.Initialize();
}


template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_Local_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int       delete_pos;
  REC_TYPE *cur_rec;

  delete_pos = m_key_maps.Search_Idx_Record(record_to_delete);
  if (delete_pos < 0)
    return NULL;

  // find the record
  cur_rec = &m_recs [delete_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_DELETE_CHAR);
#endif

  delete_pos = m_key_maps.Delete_By_Idx_Record (cur_rec);
  if (delete_pos < 0)
    return NULL;

  // set the record as available
  m_recs.Release (delete_pos);

  // send only to dgi.
  Check_Record_Actions (delete_pos, &m_tlists[delete_pos], DATA_DELETE, RESTR_SKIP, IS_LOCAL_ACTION);

  // Return the pointer to the content of the record deleted.
  return cur_rec;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
TLIST_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_Record (CMessage *cur_msg)
{ _SA_(0) 
  int         delete_pos;
  TLIST_TYPE *list;
  REC_TYPE   *cur_rec;

  delete_pos = m_key_maps.Search_Key_Record (cur_msg, 1);
  if (delete_pos < 0)
    return NULL;

  // find the record
  cur_rec = &m_recs [delete_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_DELETE_CHAR);
#endif

#ifndef DAPI
  if (m_service_support)
    m_map_service->Delete_Service_Record((DWORD64)cur_rec);
#endif

  delete_pos = m_key_maps.Delete_By_Idx_Record (cur_rec);
  if (delete_pos < 0)
    return NULL;

  // set the record as available
  m_recs.Release (delete_pos);

  list = &m_tlists [delete_pos];

  Check_Record_Actions (delete_pos, list, DATA_DELETE, RESTR_DELETE_ACTION, IS_GLOBAL_ACTION);

  return list;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
TLIST_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_Record (REC_TYPE *record_to_delete)
{ _SA_(0) 
  int           delete_pos;
  TLIST_TYPE   *list;
  REC_TYPE     *cur_rec;

  delete_pos = m_key_maps.Delete_By_Idx_Record (record_to_delete);
  if (delete_pos < 0)
    return NULL;

  // find the record
  cur_rec = &m_recs [delete_pos];

#ifdef DMEAPP
  if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    TRANS_Check (cur_rec, DATA_DELETE_CHAR);
#endif

#ifndef DAPI
  if (m_service_support)
    m_map_service->Delete_Service_Record((DWORD64)cur_rec);
#endif

  // set the record as available
  m_recs.Release (delete_pos);

  list = &m_tlists [delete_pos];

  // Do not send the global table update messages. Used in test mode alone and
  // when the matching engine is not the principal.
  Check_Record_Actions (delete_pos, list, DATA_DELETE, RESTR_DELETE_ACTION, IS_GLOBAL_ACTION);

  return list;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_NKey_Record (CMessage *cur_msg, int key_n)
{ _SA_(0) 
  int find_pos;

  // find the record
  find_pos = m_key_maps.Search_Key_Record (cur_msg, key_n);
  if (find_pos < 0)
    return NULL;

  return &m_recs [find_pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_NKey_Record (void *key2, int key_n)
{ _SA_(0) 
  int find_pos;

  // find the record
  find_pos = m_key_maps.Search_Key_Record (key2, key_n);
  if (find_pos < 0)
    return NULL;

  return &m_recs [find_pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Rec(int pos)
{ _SA_(0) 
  return &m_recs [pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_Range_It (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  int next_in_range_pos;

  next_in_range_pos = m_key_maps.Find_Key_Range_It (key1, key2, mkit, key_n);
  if (next_in_range_pos < 0)
    return NULL;

  return &m_recs [next_in_range_pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_Last_Range_It (CBaseKey *key1, CBaseKey *key2, CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  int next_in_range_pos;

  next_in_range_pos = m_key_maps.Find_Key_Last_Range_It (key1, key2, mkit, key_n);
  if (next_in_range_pos < 0)
    return NULL;

  return &m_recs [next_in_range_pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Prev_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  int prev_in_range_pos;

  prev_in_range_pos = m_key_maps.Prev_Key_Last_Range_It (mkit, key_n);

  if (prev_in_range_pos  < 0)
    return NULL;

  return &m_recs [prev_in_range_pos ];

}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Next_Range_It (CMapKeyN *mkit, int key_n)
{ _SA_(0) 
  int next_in_range_pos;

  next_in_range_pos = m_key_maps.Next_in_Key_Range_It (mkit, key_n);
  if (next_in_range_pos < 0)
    return NULL;

  return &m_recs [next_in_range_pos];
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Begin_Service_Trans(void)
{ _SA_(0) 
#ifdef DMEAPP
  m_map_service->Begin_Trans();
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Commit_Service_Trans(void)
{ _SA_(0) 
#ifdef DMEAPP
  m_map_service->Commit_Trans();
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Cancel_Service_Trans(void)
{ _SA_(0) 
#ifdef DMEAPP
  m_map_service->Cancel_Trans();
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Insert_Service(CServiceKey *service_key, CPosKey *pos_key)
{ _SA_(0) 
#ifndef DAPI
  m_map_service->Add_Service_Record(service_key, pos_key);
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
CServiceKey *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_Service(CPosKey *pos_key)
{ _SA_(0) 
#ifndef DAPI
  return m_map_service->Find_Service_Record (pos_key);
#else
  return NULL;
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_Service(CPosKey *pos_key)
{ _SA_(0) 
#ifndef DAPI
  return m_map_service->Delete_Service_Node (pos_key);
#else
  return false;
#endif
}

// pos_key !must exist!, and service_key is updated
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Service (CPosKey *pos_key, CServiceKey *service_key)
{ _SA_(0) 
#ifndef DAPI
  m_map_service->Update_Service_Record (pos_key, service_key);
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_ServiceKey_Range(CServiceKeyIterator *cskit, CServiceKey *service_key_low, CServiceKey *service_key_high)
{ _SA_(0) 
  REC_TYPE *cur_rec;

  if (!m_service_support)
    return NULL;

#ifndef DAPI
  cur_rec   = (REC_TYPE *)m_map_service->Find_ServiceKey_Range (cskit, service_key_low, service_key_high);
  return cur_rec;
#else
  return NULL;
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Next_In_ServiceKey_Range(CServiceKeyIterator *cskit)
{ _SA_(0) 
  REC_TYPE *cur_rec;

  if (!m_service_support)
    return NULL;

#ifndef DAPI
  cur_rec = (REC_TYPE*)m_map_service->Next_In_ServiceKey_Range (cskit);
  return cur_rec;
#else
  return NULL;
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE   *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_High_ServiceKey_Record(CServiceKey *service_key)
{ _SA_(0) 
  REC_TYPE *cur_rec;

  if (!m_service_support)
    return NULL;

#ifndef DAPI
  cur_rec = (REC_TYPE *)m_map_service->Find_High_ServiceKey(service_key);
  return cur_rec;
#else
  return NULL;
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE   *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Find_Low_ServiceKey_Record(CServiceKey *service_key, const CServiceKey **found_service_key)
{ _SA_(0) 
  REC_TYPE *cur_rec;

  if (!m_service_support)
    return NULL;
  
#ifndef DAPI
  cur_rec = (REC_TYPE *)(DWORD64)m_map_service->Find_Low_ServiceKey (service_key, found_service_key);
  return cur_rec;
#else
  return NULL;
#endif
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Search_Next_Free_Pos  (void)
{ _SA_(0) 
  int       next_pos;
#if DMEAPP || DAPI
  REC_TYPE *cur_rec;
#endif

  next_pos = m_recs.Use_One();

#if DMEAPP || DAPI
  cur_rec = &m_recs[next_pos];
  // Once used, the record will always belong to this table.
  cur_rec->Set_Owner_Table_Data (this, next_pos);
#endif

  return next_pos;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Table_Flag (int table_flag)
{ _SA_(0) 
  int prev_value;

  prev_value   = m_table_flag;
  m_table_flag = table_flag;
  return prev_value;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Table_Type (int table_type_id)
{ _SA_(0) 
  int cur_table_type_id;

  if (table_type_id < 0 || table_type_id > MAX_CODE_ID)
    cur_table_type_id = ZERO_CODE_ID; // Invalide table type id.
  else
    cur_table_type_id = table_type_id;

  // If there is a unique table code for this table type id that it is the same
  // in all markets, use it.
  if (G_table_id_equivalence[cur_table_type_id].code != TBLNUM_DYNAMIC_CODE)
    Set_Table_Num(G_table_id_equivalence[cur_table_type_id].code);

  // Set the table flag that corresponds to the given table type id.
  Set_Table_Flag(G_table_id_equivalence[cur_table_type_id].publish_flag);

  return 0;
}

// Ensures that all normal insertion is processed as local
// until a conrifmation is received
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Confirm_Later      (int confirm_status)
{ _SA_(0) 
  int prev_confirm_status = m_confirm_status;
  m_confirm_status = confirm_status;
  return prev_confirm_status;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Table_Num (int table_num)
{ _SA_(0) 
  if (m_tn_list)
    delete m_tn_list;
  m_tn_list = new TLIST_TYPE;
  m_tn_list->Add_Table (table_num);
}


template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Service_Support_Flag (int flag_value)
{ _SA_(0) 
#ifndef DAPI
  if (m_service_support == flag_value)
    return;
  if (m_service_support)
    delete m_map_service;
  m_service_support = flag_value;
  if (m_service_support)
    m_map_service = new CMapService();
#endif
}

// Check for any existing record. If there is an existing record, delete it previous the insertion.
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void  CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Check_Delete_Existing_Records (REC_TYPE *record_to_check, 
                                                                               int       check_idx, 
                                                                               int       is_local_action)
{ _SA_(0) 
  REC_TYPE *cur_rec;
  int       skip_log_error;
  int       key_n;
  int       num_keys;

  // Put the error in the log as long as there is not flag indicating the opposite.
  // For example, DAPI tables not available for the interface don't report the error because their
  // records are not deleted in an unexpected disconnection and they are going to receive repeated 
  // records in the reconnection.
  skip_log_error = (m_table_flag & TABLE_NO_LOG_ERRORS_INSERTION_MODIFICATION);

  if (check_idx == CHECK_IDX)
  {
    cur_rec = Find_Idx_Key_Record (record_to_check);
    if (cur_rec && cur_rec != record_to_check)
    {
      if (!skip_log_error)
      {
        LOG_Put_Err_Msg("Error. Record to be inserted/updated has the same idx (%d) of an existing one. Table %d, Record =%.20s", 
                        (int)(*record_to_check)[REC_IDX],
                        (int)(*record_to_check)[DATA_CODE],
                        record_to_check->m_data);
      }
      if (is_local_action == IS_LOCAL_ACTION)
        Delete_Local_Record (cur_rec);
      else
        Delete_Record (cur_rec);
    }
  }

  num_keys = m_key_maps.Get_Num_Keys();

  for (key_n = 1; key_n <= num_keys; ++key_n)
  {
    cur_rec = Find_NKey_Record (record_to_check, key_n);
    if (cur_rec && cur_rec != record_to_check)
    {
      if (!skip_log_error)    
      {
        LOG_Put_Err_Msg("Error. Record received to be inserted/updated has the same n key (%d) of an existing one. Existing idx=%d, Table %d, Record =%.20s", 
                        key_n,
                        (int)(*cur_rec)[REC_IDX],
                        (int)(*record_to_check)[DATA_CODE],
                        record_to_check->m_data);
      }
      if (is_local_action == IS_LOCAL_ACTION)
        Delete_Local_Record (cur_rec);
      else
        Delete_Record (cur_rec);
    }
  }
}

// populates and save record changes
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Check_Record_Actions (int         record_pos,
                                                                     TLIST_TYPE *list_table,
                                                                     const char *table_action,
                                                                     int         restr_action_required,
                                                                     int         is_local_action,
                                                                     int         save_persistent_flag)
{ _SA_(0) 
  
  REC_TYPE  *cur_rec;
  
  cur_rec = &m_recs [record_pos];
  
  if (!cur_rec)
  {
    LOG_Put_Err_Msg("Error. Record could not be retrieved. Record pos: %d.", record_pos);
    return;      
  }
  
  // Set the last table operation.
  switch (*table_action)
  {
  case DATA_INSERT_CHAR:
    m_last_operation = OPER_INSERT;
    break;
  case DATA_UPDATE_CHAR:
    m_last_operation = OPER_UPDATE;
    break;
  case DATA_DELETE_CHAR:
    m_last_operation = OPER_DELETE;
    break;
  default:
    LOG_Put_Err_Msg("Error. Unhandled action when checking record actions. Action %s.", table_action);
    return;
  }

#ifndef DAPI


  int        destination;

  // if the record was marked as pending, stop publishing
  // and saving in the persistent file
  if (cur_rec->m_confirm_status != CONFIRM_NOT_PENDING)
  {
    // if confirmation is disabled, disable flag
    // otherwise when manual transaction is disabled, it will be required 
    // to update limits records
    if (m_confirm_status)
      return;

    cur_rec->m_confirm_status = CONFIRM_NOT_PENDING;
  }

  // Check if the list table has defined a restricted destination
  // if so, use that destination without checking another condition
  if (list_table ->m_restricted_dest != NORMAL_DEST)
  {
    destination = list_table ->m_restricted_dest;
    goto SEND_RECORD_MESSAGE;
  }
  else
    // Check if the record action must be published or not.
    destination = NO_DEST;

  switch (is_local_action)
  {
  case IS_LOCAL_ACTION:
    // If the action was a call to a local action, publish only to DGI if defined.
    if (m_table_flag & TABLE_PUBLISHED_TO_DGI)
      destination = DGI_DEST;
    break;
  case IS_GLOBAL_ACTION:
    // If the action was a call to a global action, publish to the defined targets.
    if ((m_table_flag & TABLE_PUBLISHED_ALL) == TABLE_PUBLISHED_ALL)
      destination = NORMAL_DEST;
    else if (m_table_flag & TABLE_PUBLISHED_TO_DGI)
      destination = DGI_DEST;
    else if (m_table_flag & TABLE_PUBLISHED_TO_DS)
      destination = DS_DEST;
    break;
  case IS_DS_CORRECTION:
    if (m_table_flag & TABLE_PUBLISHED_TO_DS)
      destination = DS_DEST;
    break;
  }

SEND_RECORD_MESSAGE:  

  // If the last operation is an Insertion, then the predefined funtion 
  // should be called before the record insertion, this resolve the case
  // when the summary has to be sent prior to record insert for DGI processing
  // otherwise DGI will reject the insertion if the summary has not been
  // received
  if (m_last_operation == OPER_INSERT)
  {
    // If apply, call a predefined function to manage actions
    if (m_action_func_ptr)
        m_action_func_ptr (cur_rec, table_action);
    
    if (destination != NO_DEST)
        Send_Record_Messages (cur_rec, list_table, table_action, restr_action_required, destination);
  }
  else
  {
    if (destination != NO_DEST)
        Send_Record_Messages (cur_rec, list_table, table_action, restr_action_required, destination);
        
    // If apply, call a predefined function to manage actions
    if (m_action_func_ptr)
        m_action_func_ptr (cur_rec, table_action);
  }
 
    
#endif    

  // If it is a local action, don't update the persistent file if it exists.
  if (!m_Pers_File || save_persistent_flag == DO_NOT_SAVE_IF_PERSISTENT)
    return;

#if DMEAPP
  int update_pers_file_last_write_date = TRUE;
#endif

  switch (*table_action)
  {
  case DATA_INSERT_CHAR:
    m_Pers_File->Save_Record (cur_rec->m_unic_number, 
                              (*cur_rec)[DATA_INFO].Get_Ptr(),
                              (*cur_rec)[ENV_CRC32].Get_Ptr());
    break;
  case DATA_UPDATE_CHAR:
#ifdef DMEAPP
    if (Glb_TRANS_Activated && cur_rec->m_use_transaction)
    {
      update_pers_file_last_write_date = FALSE;
      break;
    }
#endif
      m_Pers_File->Save_Record (cur_rec->m_unic_number, 
                                (*cur_rec)[DATA_INFO].Get_Ptr(),
                                (*cur_rec)[ENV_CRC32].Get_Ptr());
    break;
  case DATA_DELETE_CHAR:
    m_Pers_File->Delete_Record (cur_rec->m_unic_number);
    break;
  default:
#ifdef DMEAPP
    update_pers_file_last_write_date = FALSE;
#endif
    break;
  }

#ifdef DMEAPP
  if (update_pers_file_last_write_date)
    Check_Set_Pers_File_Last_Write_Date ();
#endif
}

#if DMEAPP || GDIAPP

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Set_Action_Function_Pointer  (table_action_func  action_func_ptr)
{ _SA_(0) 
  m_action_func_ptr = action_func_ptr;
}
#endif

#if DMEAPP || DAPI
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Assoc_To_Pers_File (unsigned int  file_code,
                                                                  const    char *file_name,
                                                                  unsigned int  rec_size,
                                                                  int            keep_last_write_date)
{ _SA_(0) 
  int ret_val;

  // create the Persistent File Class
  m_Pers_File = new CPersFile (file_code, file_name, rec_size);

  ret_val = m_Pers_File->Open ();
  
  if (!ret_val)
  {
    delete m_Pers_File;
    m_Pers_File = NULL;
  }

  // If indicated, enable the persis file last write date.
  Enable_Pers_File_Last_Write_Date (keep_last_write_date);

  return ret_val;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Enable_Pers_File_Last_Write_Date (int enable)
{ _SA_(0) 
  // If not persistent file, do nothing.
  if (!m_Pers_File)
    return;

  if (!enable)
  {
    // If already disabled, do nothing.
    if (!m_Pers_File_last_write_date)
      return;

    // Free the memory.
    free (m_Pers_File_last_write_date);
    m_Pers_File_last_write_date = NULL;
    return;
  }

  // If already enabled, do nothing.
  if (m_Pers_File_last_write_date)
    return;

  m_Pers_File_last_write_date = (char *) calloc (DFN_DATE_LEN + 1, 1);
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Load_Pers_File  (void)
{ _SA_(0) 
  // take an area of data to use it
  char record_data [PERSIS_BUFFER_MAX_SIZE];
  unsigned int         rec_num, 
                        environment, 
                        crc32;
  DLONG                 persis_env_crc32;

  REC_TYPE             *cur_rec;
  int                   next_p;
  int                   has_changed, is_Big_Ltl_End;
  int                   has_to_be_updated;
  extern unsigned int  Glb_range_environment;
    
  if (!m_Pers_File)
    return;

#ifdef USE_THALERT
  int                   thalert_state;
  thalert_state = THALERT_Set_State (false);
#endif

  Delete_All_Records ();

  // determine if the file has to be updated from Big_Lit_End
  is_Big_Ltl_End = m_Pers_File->m_header.Get_Big_End();

  // Load the records from the file
  for (rec_num = 1; ; ++rec_num)
  {
    persis_env_crc32 = 0;

    // read the record
    if (!m_Pers_File->Load_Next_Record (&rec_num, record_data, &persis_env_crc32))
    {
      if (m_Pers_File->Check_Forced_Save ())
      {
          unsigned int   records_processed;
          REC_TYPE       *cur_rec;
          int             i;
          int             limit;

          records_processed = 0;
          limit             = m_recs.max_number;

          for (i = 0; i < limit; ++i)
          {
            // Use only currently used records.
            if (!m_recs.Is_Used(i))
              continue;

            cur_rec = &m_recs[i];

            cur_rec->m_unic_number = ++records_processed;

            // Use the callback function sending as argument the current record.
            m_Pers_File->Save_Record (  cur_rec->m_unic_number, 
                                      (*cur_rec)[DATA_INFO].Get_Ptr(),
                                      (*cur_rec)[ENV_CRC32].Get_Ptr());
          }
      }
       // no more records?
      break;; // stop the process
    }

    // search for the next available position
    next_p  = Search_Next_Free_Pos ();
    cur_rec = &m_recs [next_p];

    // copy partial data to the record
    (*cur_rec)[DATA_INFO] = (char *) record_data;

    // save the number of the record in the unic number
    cur_rec ->m_unic_number = rec_num;

    // from now on we are going to access fields
    has_changed = cur_rec ->CMessage::Init();

    // Copy the env and crc, as they come (with little) Intel format
    (*cur_rec)[ENV_CRC32]= (char *)&persis_env_crc32;

    environment = (unsigned int)(*cur_rec)[ENVIRONMENT];
    crc32       = (unsigned int)(*cur_rec)[CRC32];

    // set supposed env_crc32
    persis_env_crc32      =
    cur_rec->m_env_crc32  = (((DLONG)environment) << 32) | crc32;

    // if environment is zero, and crc32 is given, probably this is a record from previous version
    if (!environment)
    {
      if (crc32)
      {
        // check if updating environment do we still have to maintain it
        // this is to check other conditions
        // update environment
        (*cur_rec)[ENVIRONMENT] = Glb_range_environment;

        // check if maintain this record
        if (cur_rec->Do_We_Maintain_It())
        {
          // if we maintain, use the environment as default
          environment = Glb_range_environment;

          // assemble the env-crc32, for this machine
          persis_env_crc32 = (((DLONG)environment) << 32) | crc32;
        }
        else
        {
          // not maintained, but, environment zero!!! 
          // use previous, this record should be erased after synchronized
          (*cur_rec)[ENVIRONMENT] = 0;
        }
      }
      else
      {
        // force this environment .. this not always can be true on international tables!!!.
        // but, those international tables should be persistents
        (*cur_rec)[ENVIRONMENT] = Glb_range_environment;
      }
    }

    if (persis_env_crc32 > 0)
    {
      cur_rec->m_env_crc32_to_use = persis_env_crc32;

      // If the record uses crc32, then its crc32 and environment will be set in this
      // initialization.
      has_changed = cur_rec->Init (); 
    }

    // Check for existing records with the same key or keys. If so, delete the existing records.
    Check_Delete_Existing_Records (cur_rec, DO_NOT_CHECK_IDX, IS_LOCAL_ACTION);

    // make the required conversions-initializations in the record
    // check and change the record if any convertion required
    has_to_be_updated = cur_rec->Check_Pers_Data(DO_PERSISTENT_LOADING_CHANGES);

    // if the record is older, need to be deleted, do it
    if (has_to_be_updated == PERSISTENT_RECORD_SHOULD_BE_DELETED)
    {
      // set the record as available
      m_Pers_File->Delete_Record (rec_num);
      m_recs.Release (next_p);
      continue;
    }

    // if the record needs to be updated
    if (has_changed || (has_to_be_updated == PERSISTENT_RECORD_UPDATED) || is_Big_Ltl_End)
    {
      // this is to recompute crc32 as the record was changed
      cur_rec->Init ();

      m_Pers_File->Save_Record (rec_num, 
                                (*cur_rec)[DATA_INFO].Get_Ptr(),
                                (*cur_rec)[ENV_CRC32].Get_Ptr());
      // Don't update the last write date because this write is result of a
      // checking of the validity of the table data, not of a system functionality.
      // ...
    }

    // copy the table list
#ifdef DAPI    
    if (m_tn_list)
        m_tlists (next_p) = *m_tn_list;
#else
    m_tlists (next_p) = m_tn_list ? *m_tn_list : *(TLIST_TYPE *)cur_rec->Get_TableList ();
#endif

    // el ambiente ya fue seteado antes con el mismo valor que tiene el registro, por tal
    // razn no se require cambiar nuevamente al ambiente por el que est usando el dme
    // as se evita duplicidad de registros
    m_key_maps.Add_Record (next_p, cur_rec, NOT_PUT_ENV_AGAIN);

    // if (m_service_support)
    //   cur_rec->Add_Services_Map(m_map_service);

    // Do not send the global table update messages. Used in test mode alone and
    // ......

  }
#ifdef USE_THALERT
  thalert_state = THALERT_Set_State (thalert_state);
#endif

  // if the file required big to little endiang conversion, 
  // now on reset the conversion flags for header and records.
  if (is_Big_Ltl_End)
  {
    m_Pers_File->m_header.Not_Use_Big_End ();
    m_Pers_File->m_header.Write_Data ();
    m_Pers_File->m_record_header.Not_Use_Big_End ();
    m_Pers_File->m_record_p->Not_Use_Big_End     ();
  }
}

#endif

#ifdef DMEAPP

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
unsigned int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Max_Rec_Number (void)
{ _SA_(0) 
  // if the persisten file is not defined yet return 0
  if (!m_Pers_File)
    return 0;

  // take the maximum number from the file's header.. 
  return  *m_Pers_File->m_max_number_p;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
void CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Check_Set_Pers_File_Last_Write_Date (void)
{ _SA_(0) 
  if (!m_Pers_File || !m_Pers_File_last_write_date)
    return;

  // Update the last write date if the pers file was initiated for it.
  COMMON_Get_Date_Format (m_Pers_File_last_write_date);
  m_Pers_File->Set_Date_Reference (m_Pers_File_last_write_date);
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Get_Pers_File_Last_Write_Date (char *last_write_date)
{ _SA_(0) 
  if (!m_Pers_File || !m_Pers_File_last_write_date)
  {
    memset (last_write_date, 0, DFN_DATE_LEN);
    return FALSE;
  }

  // Get the last write date from the file.
  m_Pers_File->Get_Date_Reference (last_write_date);
  return TRUE;
}

// This is a safe process of the records of a table because it does not access the 
// mappings and  does not use iterators. It must be used for informative purposes 
// and must not change any value of the table. Initially used to get credit policies
// counters.
template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
unsigned int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Safe_Process_All_Records  (table_process_func process_func_ptr)
{ _SA_(0) 
  unsigned int   records_processed;
  REC_TYPE       *cur_rec;
  int             i;
  int             limit;

  records_processed = 0;
  limit             = m_recs.max_number;

  for (i = 0; i < limit; ++i)
  {
    // Use only currently used records.
    if (!m_recs.Is_Used(i))
      continue;

    cur_rec = &m_recs[i];

    // Use the callback function sending as argument the current record.
    if (process_func_ptr (cur_rec))
      records_processed++;
  }
  return records_processed;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Init_Transaction_In_Element   (int pos_in_table, char action, char *element_data)
{ _SA_(0) 
  if (pos_in_table < 0)
    return 1;

  // By now, only when deleting a record, it is necessary to save the table
  // last free element in the record vector.
  if (action == DATA_DELETE_CHAR)
    memcpy (element_data, &m_recs.last_free, sizeof (m_recs.last_free));

  return 0;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Cancel_Transaction_In_Element  (int pos_in_table, char action, const char *element_data, const char *content_to_use)
{ _SA_(0) 
  int last_free;
  int record_updated;

  if (pos_in_table < 0)
    return 0; // record not updated.

  // Assume the record will be updated.
  record_updated = 1;

  switch (action)
  {
  case DATA_INSERT_CHAR:
    // Delete the record with its current content.
    Delete_Transactional_Record (pos_in_table);
    record_updated = 0;
    break;
  case DATA_UPDATE_CHAR:
    Update_Transactional_Record (pos_in_table, content_to_use);
    break;
  case DATA_DELETE_CHAR:
    // If reinserting a deleted record, recover the last free element in the record vector.
    memcpy (&last_free, element_data, sizeof (m_recs.last_free));
    Insert_Transactional_Record (pos_in_table, content_to_use, last_free);
    break;
  default:
    // Error. Record not updated.
    record_updated = 0;
    break;
  }

  // No error.
  return record_updated;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
int CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Commit_Transaction_In_Element (int pos_in_table, char action, const char *element_data)
{ _SA_(0) 
  REC_TYPE *cur_rec;

  if (pos_in_table < 0)
    return 1; // not valid record.

  cur_rec = &m_recs[pos_in_table];

  switch (action)
  {
  case DATA_INSERT_CHAR:
    break;
  case DATA_UPDATE_CHAR:
    if (m_Pers_File)
    {
      m_Pers_File->Save_Record (  cur_rec->m_unic_number, 
                                  (*cur_rec)[DATA_INFO].Get_Ptr(),
                                  (*cur_rec)[ENV_CRC32].Get_Ptr());
      Check_Set_Pers_File_Last_Write_Date ();
    }
    break;
  case DATA_DELETE_CHAR:
    break;
  default:
    break;
  }

  return 0;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Insert_Transactional_Record (int pos_in_table, const char *content_to_use, int last_free_to_use)
{ _SA_(0) 
  REC_TYPE *cur_rec;
  
  cur_rec = &m_recs [pos_in_table];

  // Copy the original data of the record.
  cur_rec->Update_Message_Info (content_to_use);

  m_key_maps.Add_Local_Record (pos_in_table, cur_rec);

  if (m_service_support)
    cur_rec->Add_Services_Map(m_map_service);

  m_recs.Unrelease (pos_in_table, last_free_to_use);

  // if persistent file is used, insert previously deleted record
  if (m_Pers_File)
  {
    m_Pers_File->Save_Record (cur_rec->m_unic_number, 
                              (*cur_rec)[DATA_INFO].Get_Ptr(),
                              (*cur_rec)[ENV_CRC32].Get_Ptr());
  }

  return 0;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Update_Transactional_Record (int pos_in_table, const char *content_to_use)
{ _SA_(0) 
  REC_TYPE   *cur_rec;

  cur_rec = &m_recs [pos_in_table];

  m_key_maps.Begin_Update_Keys (cur_rec);

  // Restore the old record data.
  cur_rec->Update_Message_Info (content_to_use);

  m_key_maps.End_Update_Keys   (pos_in_table, cur_rec);

  // if persistent file is used, Restore previous content
  if (m_Pers_File)
  {
    m_Pers_File->Save_Record (cur_rec->m_unic_number, 
                              (*cur_rec)[DATA_INFO].Get_Ptr(),
                              (*cur_rec)[ENV_CRC32].Get_Ptr());
  }

  m_last_operation = OPER_UPDATE;
  return cur_rec;
}

template <class REC_TYPE, class MAP_KEY_TYPE, class TLIST_TYPE>
REC_TYPE *CTable<REC_TYPE,MAP_KEY_TYPE,TLIST_TYPE>::Delete_Transactional_Record (int pos_in_table)
{ _SA_(0) 
  REC_TYPE *cur_rec;
  int       delete_pos;

  // find the record
  cur_rec = &m_recs [pos_in_table];

  // if persistent file is used, delete from it
  if (m_Pers_File)
  {
    m_Pers_File->Delete_Record (cur_rec->m_unic_number);
  }

  if (m_service_support)
    m_map_service->Delete_Service_Record((DWORD64)cur_rec);

  // set the record as available
  m_recs.Release (pos_in_table);

  delete_pos = m_key_maps.Delete_By_Idx_Record (cur_rec);
  if (delete_pos < 0)
    return NULL;

  // Return the pointer to the content of the record deleted.
  return cur_rec;
}

#endif  // ifdef DMEAPP


//////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//                                                                           //   
//                              G   D  I                                     //
//                                                                           //
//  Gustavo Chafla Ph.D.                                                     //
//  Date:   2008                                                             //
///////////////////////////////////////////////////////////////////////////////
#ifdef GDIAPP

class CETCPageKey : public CBaseKey
{
public:
  CETCPageKey();

  CETCPageKey *Init        (const char    *datatec_market, 
                            unsigned int  page_number);

  const char     *Get_Market()     {return (m_key);}
  const char     *Get_PageNumber() {return (m_key + ORD_MARKET_LEN);}
};

// Key with the parameters that defines an instrument.
class CInstrumentKey : public CBaseKey
{

public:
  CInstrumentKey();
  
  CInstrumentKey *Init(const char *etc_page_key,
                       const char *title_code,   
                       const char *issue,   
                       const char *reference,
                       const char *currency, 
                       const char *due_date);

  CInstrumentKey *Init(const char *etc_page_key,
                       int         fill_char);

  void            Build_ETCPageKey (CETCPageKey *etc_page_key) {etc_page_key ->Init(Get_Market(), Get_PageNumber());}
  
  const char     *Get_Market()    {return (m_key);}
  unsigned int   Get_PageNumber(){return (*(unsigned int*)&m_key[ORD_MARKET_LEN]);}
  const char     *Get_Text_Term() {return (m_key + ORD_MARKET_LEN + DFN_CURRENCY_LEN + DFN_CURRENCY_LEN + ORD_LEVEL_LEN + ORD_MATUREDAY_LEN + ORD_XDAYS_LEN + ORD_DAYS_LEN);}
};


// Key for Summary FI Table
class CSummaryFIKey : public CBaseKey
{
  public:
  CSummaryFIKey();
  CSummaryFIKey *Init(const char *title_code,
                      const char *issuer,
                      const char *reference,
                      const char *currency,
                      const char *due_date);
};

// Key for Summary CO Table
class CSummaryCOKey : public CBaseKey
{
  public:
  CSummaryCOKey();
  CSummaryCOKey *Init(const char *etc_page_key,
                      const char *nemotec,
                      const char *value_date,
                      const char *pcomp_flag,
                      const char *due_date,
                      const char *order_class);

  CSummaryCOKey *Init_Sort (const char *nemotec,
                            const char *due_date,
                            const char *value_date,
                            const char *pcomp_flag,
                            const char *etc_page_key,
                            const char *order_class);

  const char     *Get_Market()    {return (m_key);}    
  void            Build_ETCPageKey (CETCPageKey *etc_page_key) {etc_page_key ->Init(Get_Market(), Get_PageNumber());}
  unsigned int   Get_PageNumber(){return (*(unsigned int*)&m_key[ORD_MARKET_LEN]);}  
};

// Key for Summary RF Table
class CSummaryRFKey : public CBaseKey
{
  public:
  CSummaryRFKey();
  CSummaryRFKey *Init(const char *market,
                      const char *page_number,
                      const char *instrument_number);
    const char     *Get_Market()    {return (m_key);}    
};

// Key with the parameters that defines an instrument.
class CMXInstrumentKey : public CBaseKey
{

private:

   int m_sort_type;

public:
  
  CMXInstrumentKey();
  
  CMXInstrumentKey *Init(const char *etc_page_key,
                         const char *group_code,
                         const char *title_reference,
                         const char *text_term,
                         const char *currency,
                         const char *arrange,
                         const char *xdays,
                         const char *days,
                         const char *value_date,
                         const char *cama_de_monto,   // no proper translation
                         const char *tipo_operacion,  // no proper translation
                         const char *linea_pico,
                         int         sort_type);

  CMXInstrumentKey *Init(const char *etc_page_key,
                         int         fill_char);

  void            Build_ETCPageKey (CETCPageKey *etc_page_key) {etc_page_key ->Init(Get_Market(), Get_PageNumber());}
  
  const char     *Get_Market()    {return (m_key);}
  unsigned int   Get_PageNumber(){return (*(unsigned int*)&m_key[ORD_MARKET_LEN]);}  

  bool            operator <( const CMXInstrumentKey &X) const;
};

class CFITermKey : public CBaseKey
{

public:
  CFITermKey();
  CFITermKey *Init(const char *market,
                   const char *title_code,
                   const char *issue,
                   const char *reference,
                   const char *currency,
                   const char *due_date);

  CFITermKey *Init_IC  (const char *market,
                        const char *nemotec,
                        const char *value_date,
                        const char *pcomp_flag,
                        const char *due_date,
                        const char *ord_class);
};

class CHistoryFIKey : public CBaseKey
{

public:
  CHistoryFIKey();
  CHistoryFIKey *Init(const char *market,
                      const char *days,
                      const char *hour,
                      const char *ordercode,
                      const char *ordercode1);     
};
#endif

#endif //__TABLE_H__
