/////////////////////////////////////////////////////////////////////////////////
//                                                                             //
// Module: thalert.h                                                           //
// the main purpose is to help thread-safety alerting and preventing for       //
// possible collisions and functions not correctly used by multiple threads    //
//                                                                             //
/////////////////////////////////////////////////////////////////////////////////

#ifndef __THALERT_H__
#define __THALERT_H__

// #define  USE_THALERT

#  ifdef USE_THALERT

#define THA_THREAD_SAFE         0xF0000000

void            THALERT_Init        (void);

// used to disable capture..
// for extreemly heavy code, with not meaninful results
int             THALERT_Set_State   (int state);

// called at the end of the program
extern "C" void  THALERT_End    (void);

// Function called by every thread when starting to run
void          * THALERT_Register    (unsigned thread_prg_id, const char *thread_name);

// Functions called by the functions to allow the thread-alert control


// called when a mutex is being locked
void            THALERT_Post_Locking_Mutex      ( const char *mutex_name);

// called when a mutex is being unlocked
void            THALERT_Post_Unlocking_Mutex    ( const char *mutex_name);

// This class is created to have constructor and destructor,
// those functions will be used to manage thread alert calls
class CThal
{
public:
    // used by functions at the beginning and at the end
    // tells a function is starting its execution
    const char  * m_function_name;
    void        * m_aux_ptr;
    CThal (const char *function_name, unsigned allowed_threads);

    // tell a function is finishing its execution
    ~CThal ();
};

// used to set a check point
// start alert in a function
#if USE_SOLARIS || USE_LINUX
#define _SA_(x) CThal _t_CThal (__PRETTY_FUNCTION__ , x);
#else
#define _SA_(x) CThal _t_CThal (__FUNCDNAME__ , x);
#endif

#define _LM_(x) THALERT_Post_Locking_Mutex(x);

#define _UM_(x) THALERT_Post_Unlocking_Mutex(x);

#  else       // USE_THALERT

#define _SA_(x) 
// end alert in a function

#define _LM_(x) 

#define _UM_(x)

#define THALERT_Init()

// called at the end of the program
#define THALERT_End()

// Function called by every thread when starting to run
#define THALERT_Register(x,y) 


#  endif      // USE_THALERT

#endif      // __THALERT_H__

///////////////////////////////////////////////////////////////////////////////////////////////
// end of the module
///////////////////////////////////////////////////////////////////////////////////////////////


